using System;

namespace Evaluant.OPath.Expressions
{
	/// <summary>
	/// Description rsume de Path.
	/// </summary>
	[Serializable]
    public class Value : Expression, ICloneable
	{
		/// <summary>
		/// Creates a new <see cref="Value"/> instance.
		/// </summary>
		public Value() : base()
		{
			text = String.Empty;
			type = ValueEnum.Unknown;
		}

        public Value(string text, ValueEnum type)
        {
            this.text = text;
            this.type = type;
        }

        [System.Diagnostics.DebuggerStepThrough]
        public override void Accept(OPathVisitor visitor)
		{
			visitor.Visit(this);
		}

		private string text;
		/// <summary>
		/// Gets or sets the value.
		/// </summary>
		/// <value></value>
		public string Text
		{
			get { return text; }
			set { text = value; }
		}

		private ValueEnum type;
		/// <summary>
		/// Gets or sets the type.
		/// </summary>
		/// <value></value>
		public ValueEnum Type
		{
			get { return type; }
			set { type = value; }
		}

		/// <summary>
		/// Gets the int representation of the parsed text.
		/// </summary>
		/// <returns>The int instance representing the parsed text</returns>
		public int GetInt()
		{
			return int.Parse(text);
		}

		/// <summary>
		/// Gets the string representation of the parsed text.
		/// </summary>
		/// <returns>The string instance representing the parsed text</returns>
		public string GetString()
		{
			return text;
		}

		/// <summary>
		/// Gets the date time representation of the parsed text.
		/// </summary>
		/// <returns>The DateTime instance representing the parsed text</returns>
		/// <remarks>The default regional settings will be used to recognized the date time format</remarks>
		public DateTime GetDateTime()
		{
			return DateTime.Parse(text);
		}

		/// <summary>
		/// Gets the double representation of the parsed text.
		/// </summary>
		/// <returns>The double instance representing the parsed text</returns>
		public double GetDouble()
		{
			System.Globalization.NumberFormatInfo fi = new System.Globalization.NumberFormatInfo();
			fi.NumberDecimalSeparator = ".";
			return double.Parse(text, fi);
		}

		/// <summary>
		/// Gets the boolean representation of the parsed text.
		/// </summary>
		/// <returns>The boolean instance representing the parsed text</returns>
		public bool GetBoolean()
		{
			return bool.Parse(text);
		}

		/// <summary>
		/// Gets the value represented by the parsed text.
		/// </summary>
		/// <returns>The value represented by the parsed text.</returns>
		public object GetValue()
		{
			switch(type)
			{
				case ValueEnum.Integer :
					return GetInt();
				case ValueEnum.String :
					return GetString();
				case ValueEnum.Float :
					return GetDouble();
				case ValueEnum.Date :
					return GetDateTime();
				case ValueEnum.Boolean :
					return GetBoolean();
			}

			return null;							
		}

        public override void ToString(System.Text.StringBuilder sb)
        {
            switch (type)
            {
                case ValueEnum.Float:
                case ValueEnum.Integer:
                case ValueEnum.Boolean:
                    sb.Append(this.Text);
                    break;
                case ValueEnum.String:
                    sb.Append("'").Append(this.Text).Append("'");
                    break;
                case ValueEnum.Date:
                    sb.Append("#").Append(this.Text).Append("#");
                    break;
            }
        }

		#region Membres de ICloneable

		public override object Clone()
		{
			Value clone = new Value();
			clone.Text = this.Text;
			clone.Type = this.Type;

			return clone;
		}

		#endregion
	}

	public enum ValueEnum
	{
		Unknown = 0,
		Integer = 1,
		String = 3,
		Date = 4,
		Float = 2,
		Boolean = 5
	}
}
