using System;
using System.Collections;
using System.Xml;
using System.Xml.Serialization;
using Evaluant.Uss.MetaData;
using Evaluant.Uss.Common;

namespace Evaluant.Uss.Models
{
	/// <summary>
	/// Description rsume de Model.
	/// </summary>
	/// <remarks>This class is Thread Safe. An instance can be used by several Persistence Engines concurently</remarks>
	[Serializable]
    [XmlRoot("Model")]
	public class Model
    {

        [XmlNamespaceDeclarations]
        public XmlSerializerNamespaces xmlns;
        
        #region Members

        private EntityCollection _Entities;
        private Hashtable _EntitiesIndex;
        private Hashtable _AttributesIndex;
        private Hashtable _ReferencesIndex;

        private Hashtable _FullNameAttributeIndex;
        private Hashtable _FullNameAttributeBranchIndex;

        #endregion

        #region Ctor

        public Model()
		{
			_Entities = new EntityCollection();
            _EntitiesIndex = new Hashtable();
            _AttributesIndex = new Hashtable();
            _ReferencesIndex = new Hashtable();

            _FullNameAttributeIndex = new Hashtable();
            _FullNameAttributeBranchIndex = new Hashtable();

            xmlns = new XmlSerializerNamespaces(new XmlQualifiedName[] { new XmlQualifiedName(String.Empty, "http://euss.evaluant.com/schemas/MetaDataModel.xsd") });
        }

        #endregion

        #region Entities

        [XmlElement("Entity")]
		public EntityCollection Entities
		{
			get { return _Entities; }
			set	{ _Entities = value; }
		}

		
		public Entity GetEntity(string name)
		{
            lock (_EntitiesIndex)
            {
                if (_EntitiesIndex.Contains(name))
				return _EntitiesIndex[name] as Entity;

				foreach(Entity e in _Entities)
					if(e.Type == name)
					{
						_EntitiesIndex.Add(name, e);
						return e;
					}
			}

			return null;
		}

		public Entity GetEntity(string name, bool strict)
		{
			Entity entity = GetEntity(name);

			if(entity == null && strict)
				throw new ModelElementNotFoundException(String.Format("Type [{0}] not found", name));

			return entity;
        }

        #endregion

        #region Inheritance

        public Entity GetParent(Entity entity)
		{
			foreach(Entity e in _Entities)
				if(e.Type == entity.Inherit)
					return e;
			return null;
		}

		public EntityCollection GetTree(string type)
		{
			EntityCollection list = new EntityCollection();
			GetTree(GetEntity(type), list);

			return list;
		}

		public string[] GetTreeAsArray(string type)
		{
			ArrayList types = new ArrayList();
            foreach (Models.Entity e in GetTree(type))
            {
                types.Add(e.Type);
            }

			return (string[])types.ToArray(typeof(string));
		}

		private void GetTree(Entity entity, IList list)
		{
			if(entity == null)
				return;

            if(list.Contains(entity))
            {
                return;
            }

            if (!entity.IsInterface)
            {
                list.Add(entity);
            }

            foreach (Entity sub in _Entities)
            {
                if (sub.Inherit == entity.Type)
                    GetTree(sub, list);
                else
                {
                    if (Array.IndexOf(sub.Interfaces, entity.Type) != -1)
                    {
                        GetTree(sub, list);
                    }
                }
            }
        }

        #endregion

        #region References

        public ReferenceCollection GetInheritedReferences(string type)
		{
			Entity parent = GetEntity(type, true);

		    ReferenceCollection references = new ReferenceCollection();

            while (parent != null)
            {
                foreach (Reference a in parent.References)
                    references.Add(a);

                parent = GetParent(parent);
            }

		    return references;
		}
	
		public Reference GetReference(string type, string name)
		{
			string index = String.Concat(type, ":", name);

            lock (_ReferencesIndex)
            {
                if (_ReferencesIndex.Contains(index))
					return _ReferencesIndex[index] as Reference;

				foreach(Reference e in GetInheritedReferences(type))
					if(e.Name == name)
					{
						_ReferencesIndex.Add(index, e);
						return e;
					}
			}

			return null;
		}

		public Reference GetReference(string type, string name, bool strict)
		{
			Reference reference = GetReference(type, name);

			if(reference == null && strict)
				throw new ModelElementNotFoundException(String.Format("Reference [{0}] not found in Type [{1}]", name, type));

			return reference;
        }

        #endregion

        #region Attributes

        public AttributeCollection GetInheritedAttributes(string type)
        {
            AttributeCollection attributes = new AttributeCollection();
            Entity parent = GetEntity(type, true);

			while (parent != null)
			{
				foreach (Attribute a in parent.Attributes)
					attributes.Add(a);

				parent = GetParent(parent);
			}

            return attributes;
        }

		public Attribute GetAttribute(string type, string name)
		{
            string index = String.Concat(type, ":", name);

            lock (_AttributesIndex)
            {
		        if(_AttributesIndex.Contains(index))
			        return _AttributesIndex[index] as Attribute;

				foreach(Attribute e in GetInheritedAttributes(type))
					if(e.Name == name)
					{
						_AttributesIndex.Add(index, e);
						return e;
					}
			}

			return null;
		}

		public Attribute GetAttribute(string type, string name, bool strict)
		{
			Attribute attribute = GetAttribute(type, name);

			if(attribute == null && strict)
				throw new ModelElementNotFoundException(String.Format("Attribute [{0}] not found in Type [{1}]", name, type));

			return attribute;
        }

        #endregion

        #region Full Attribute Name (Inheritance branch)

        /// <summary>
        /// Return the full attribute name ([Type].[AttName]) with the top most classe type containing this attribute
        /// </summary>
        /// <param name="shortAttributeName"></param>
        /// <param name="typeName"></param>
        /// <returns></returns>
        public string GetFullAttributeName(string shortAttributeName, string typeName)
        {
            string strIndex = String.Concat(typeName, ":", shortAttributeName);
            
            lock (_FullNameAttributeIndex)
            {
                if (_FullNameAttributeIndex.Contains(strIndex))
                return _FullNameAttributeIndex[strIndex].ToString();

				string fullName = String.Empty;

				Evaluant.Uss.Models.Entity parent = this.GetEntity(typeName);
				Evaluant.Uss.Models.EntityCollection parents = new Evaluant.Uss.Models.EntityCollection();

				do
				{
					parents.Add(parent);
					parent = this.GetParent(parent);
				}
				while (parent != null);

				// get attributes of each parent classes, specified class
				foreach (Evaluant.Uss.Models.Entity entity in parents)
					foreach (Evaluant.Uss.Models.Attribute attribute in entity.Attributes)
					{
						if (shortAttributeName == attribute.Name)
						{
							fullName = String.Concat(entity.Type, ".", attribute.Name);
							_FullNameAttributeIndex.Add(strIndex, fullName);
							return fullName;
						}
					}

				// get attributes of each child classes of the specified one
				foreach (Evaluant.Uss.Models.Entity entity in this.GetTree(parents[0].Type))
				{
					// specified class already parsed in the foreach loop above
					if (entity.Type == parents[0].Type) continue;

					foreach (Evaluant.Uss.Models.Attribute attribute in entity.Attributes)
					{
						if (shortAttributeName == attribute.Name)
						{
							fullName = String.Concat(entity.Type, ".", attribute.Name);
							_FullNameAttributeIndex.Add(strIndex, fullName);
							return fullName;
						}
					}
				}
			}

            return shortAttributeName;
        }

        /// <summary>
        /// Return an array of full attributes names ([Type].[AttName]) with the top most classe type, 
        /// for all class in the inheritance branch of the specified class type
        /// </summary>
        /// <param name="typeName"></param>
        /// <returns></returns>
        public string[] GetFullAttributesNames(string typeName)
        {
            lock (_FullNameAttributeBranchIndex)
            {
                if (_FullNameAttributeBranchIndex.Contains(typeName))
                return (string[])_FullNameAttributeBranchIndex[typeName];

				System.Collections.Specialized.StringCollection fullNames = new System.Collections.Specialized.StringCollection();

				Evaluant.Uss.Models.Entity parent = this.GetEntity(typeName);
				Evaluant.Uss.Models.EntityCollection parents = new Evaluant.Uss.Models.EntityCollection();

				do
				{
					parents.Add(parent);
					parent = this.GetParent(parent);
				}
				while (parent != null);

				// get attributes of each parent classes, specified class
				foreach (Evaluant.Uss.Models.Entity entity in parents)
					foreach (Evaluant.Uss.Models.Attribute attribute in entity.Attributes)
						fullNames.Add(String.Concat(entity.Type, ".", attribute.Name));

				// get attributes of each child classes of the specified one
				foreach (Evaluant.Uss.Models.Entity entity in this.GetTree(parents[0].Type))
				{
					// specified class already parsed in the foreach loop above
					if (entity.Type == parents[0].Type) continue;

					foreach (Evaluant.Uss.Models.Attribute attribute in entity.Attributes)
						fullNames.Add(String.Concat(entity.Type, ".", attribute.Name));
				}

				string[] arrayFullNames = new string[fullNames.Count];
				fullNames.CopyTo(arrayFullNames, 0);

				_FullNameAttributeBranchIndex.Add(typeName, arrayFullNames);

				return arrayFullNames;
			}
        }

        #endregion

        /// <summary>
        /// Returns all the base entities in the model
        /// </summary>
        /// <returns>All the base entities in the model</returns>
        public EntityCollection GetBaseEntities()
        {
            EntityCollection baseEntities = new EntityCollection();

            // Searches for all base types in the model
            foreach (Entity entity in this.Entities)
            {
                if (entity.IsInterface)
                {
                    continue;
                }

                // Steps ahead until the base class is reached
                Entity parent = entity;
                while (this.GetParent(parent) != null && !this.GetParent(parent).IsInterface)
                    parent = this.GetParent(parent);

                if (!baseEntities.Contains(parent))
                    baseEntities.Add(parent);
            }

            return baseEntities;
        }

        /// <summary>
        /// Gets all the references in the model
        /// </summary>
        /// <returns>A collection of all the references in the model</returns>
        public ReferenceCollection GetAllReferences()
        {
            ReferenceCollection references = new ReferenceCollection();

            foreach (Entity entity in this.Entities)
                foreach (Reference reference in entity.References)
                    references.Add(reference);

            return references;
        }

        public void RegisterMetadata(IMetaData[] metadata)
        {
            ModelMetaDataVisitor visitor = new ModelMetaDataVisitor(this);
            foreach (IMetaData md in metadata)
                md.Accept(visitor);
        }

        /// <summary>
        /// Clears the attribute cache.
        /// </summary>
        public void ClearAttributeCache()
        {
            _AttributesIndex = new Hashtable();
        }

        /// <summary>
        /// Clears the entity cache.
        /// </summary>
        public void ClearEntityCache()
        {
            _EntitiesIndex = new Hashtable();
        }

        /// <summary>
        /// Clears the reference cache.
        /// </summary>
        public void ClearReferenceCache()
        {
            _ReferencesIndex = new Hashtable();
        }
    }
}
