using System;
using System.Collections;
using System.Reflection;
using System.Threading;

namespace Evaluant.Uss.ObjectContext.Descriptors
{
	/// <summary>
	/// Provides an IPersistentDescriptor implementation based on Reflection returning
	/// all public properties
	/// </summary>
	public class ReflectionDescriptor : IPersistentDescriptor
	{

		private Hashtable _CachedDescriptors = new Hashtable ();
		private Hashtable _CachedIDs = new Hashtable ();

        private static ReaderWriterLock _RWL = new ReaderWriterLock();

        public PropertyDescriptor[] GetPersistentProperties(Type targetType)
        {
            _RWL.AcquireReaderLock(Timeout.Infinite);

            try
            {
                if (_CachedDescriptors.ContainsKey(targetType))
                    return (PropertyDescriptor[])_CachedDescriptors[targetType];
            }
            finally
            {
                _RWL.ReleaseReaderLock();
            }

            _RWL.AcquireWriterLock(Timeout.Infinite);

            ArrayList properties = new ArrayList();

            try
            {
                Type parentType = targetType;

                while (parentType != null && parentType != typeof(object))
                {
                    foreach (PropertyInfo pi in parentType.GetProperties(BindingFlags.Public | BindingFlags.Instance))
                        if (DescriptorHelper.IsValidProperty(pi.Name))
                        {
                            //  Try to get an NotSerialized attribute for this property
                            NotSerializedAttribute[] nsAttrs = pi.GetCustomAttributes(
                                typeof(NotSerializedAttribute), false) as NotSerializedAttribute[];

                            // Skip this property if more than one Attribute
                            if (nsAttrs != null && nsAttrs.Length > 0)
                            {
                                continue;
                            }


                            //  Try to get an attribute for this property
                            PersistentPropertyAttribute[] attrs = pi.GetCustomAttributes(
                                typeof(PersistentPropertyAttribute), false) as PersistentPropertyAttribute[];

                            FieldInfo fi = null;

                            bool serializeAsAttribute = false;
                            bool isComposition = false;
                            bool serializeAsIList = true;

                            // Add this property if more than one Attribute
                            if (attrs != null && attrs.Length > 0)
                            {
                                // Takes the first attribute even if more are set
                                PersistentPropertyAttribute ppa = attrs[0];

                                isComposition = ppa.Composition;
                                serializeAsAttribute = ppa.SerializeAsAttribute;
                                serializeAsIList = ppa.SerializeAsIList;

                                if (ppa.FieldName != null && ppa.FieldName != string.Empty)
                                    fi = parentType.GetField(ppa.FieldName, BindingFlags.DeclaredOnly | BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);
                            }

                            if (fi == null)
                            {
                                // Searches for a Field with the same name but a different case
                                fi = parentType.GetField(pi.Name, BindingFlags.DeclaredOnly | BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);

                                // If no field was found, searches for _{PropertyName}
                                if (fi == null)
                                    fi = parentType.GetField("_" + pi.Name, BindingFlags.DeclaredOnly | BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);

                                // If no field was found, searches for m_{PropertyName}
                                if (fi == null)
                                    fi = parentType.GetField("m_" + pi.Name, BindingFlags.DeclaredOnly | BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);
                            }

                            if (fi != null)
                            {
                                bool isEntity = DescriptorHelper.IsEntity(fi.FieldType);

                                if (serializeAsAttribute)
                                    isEntity = false;

                                bool isIList = DescriptorHelper.IsList(fi.FieldType);

                                if (!serializeAsIList && isIList)
                                    serializeAsIList = false;
                                else
                                    serializeAsIList = isIList;

                                if (!DescriptorHelper.IsList(fi.FieldType) && serializeAsIList)
                                    throw new Exception("invalid parameter: SerializeAsIList (not an IList)");

                                properties.Add(new PropertyDescriptor(fi.Name, pi.Name, isEntity, serializeAsIList
                                    , isComposition, fi.FieldType, fi.IsPrivate));
                            }
                        }
                    parentType = parentType.BaseType;
                }

                // Removes the Id from the list
                PropertyDescriptor idDescriptor = GetIdDescriptor(targetType);
                if (idDescriptor != null)
                {
                    for(int i=0; i<properties.Count;i++)
                    {
                        if (((PropertyDescriptor)properties[i]).PropertyName == idDescriptor.PropertyName)
                        {
                            properties.RemoveAt(i);
                            break;
                        }
                    }
                }

                PropertyDescriptor[] ret = (PropertyDescriptor[])properties.ToArray(typeof(PropertyDescriptor));
                _CachedDescriptors.Add(targetType, ret);

                return ret;
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }
        }

        public PropertyDescriptor GetIdDescriptor(Type targetType)
        {
            _RWL.AcquireReaderLock(Timeout.Infinite);

            try
            {
                if (_CachedIDs.ContainsKey(targetType))
                    return (PropertyDescriptor)_CachedIDs[targetType];

            }
            finally
            {
                _RWL.ReleaseReaderLock();
            }

            _RWL.AcquireWriterLock(Timeout.Infinite);

            try
            {
                foreach (PropertyInfo pi in targetType.GetProperties(BindingFlags.Public | BindingFlags.Instance))
                {
                    FieldInfo fi = null;

                    //  Try to get an attribute for this property
                    PersistentIdAttribute[] attrs = pi.GetCustomAttributes(
                        typeof(PersistentIdAttribute), false) as PersistentIdAttribute[];

                    // Add this property if more than one Attribute
                    if (attrs != null && attrs.Length > 0)
                    {
                        // Takes the first attribute even if more are set
                        PersistentIdAttribute pia = attrs[0];

                        if (pia.FieldName != null && pia.FieldName != string.Empty)
                            fi = targetType.GetField(pia.FieldName, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);
                    }

                    if (fi == null)
                    {
                        string propertyname = String.Concat(targetType.Name, "Id");
                        if (pi.Name != propertyname && pi.Name != "Id")
                            continue;
                        else
                            if (pi.Name == "Id")
                                propertyname = "Id";

                        // Searches for a Field with the default id name but a different case
                        fi = targetType.GetField(propertyname, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);

                        // If no field was found, searches for _{TypeName}Id
                        if (fi == null)
                            fi = targetType.GetField("_" + propertyname, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);

                        // If no field was found, searches for m_{TypeName}Id
                        if (fi == null)
                            fi = targetType.GetField("m_" + propertyname, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);
                    }

                    if (fi != null)
                    {
                        PropertyDescriptor pid = new PropertyDescriptor(fi.Name, pi.Name, DescriptorHelper.IsEntity(fi.FieldType), DescriptorHelper.IsList(fi.FieldType), false, fi.FieldType, fi.IsPrivate);
                        _CachedIDs.Add(targetType, pid);
                        return pid;


                    }
                }
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }

            if (targetType.BaseType != null)
                return GetIdDescriptor(targetType.BaseType);

            //If no Id property / _Id field found
            return null;

        }

        public void SetIdDescriptor(Type target, string propertyName, string fieldName, Type propertyType, bool usePublicProperty)
        {
            _RWL.AcquireWriterLock(Timeout.Infinite);

            try
            {
                _CachedIDs.Add(target, new PropertyDescriptor(fieldName, propertyName, false, false, false, false, propertyType, usePublicProperty));
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }
        }
	}
}
