using System;
using System.Collections;
using System.Globalization;

using Evaluant.Uss;
using Evaluant.Uss.Models;
using Evaluant.Uss.Commands;
using Evaluant.Uss.Collections;
using Evaluant.OPath;

namespace Evaluant.Uss.Common
{
    /// <summary>
    /// Description rsume de PersistenceEngineImplementation.
    /// </summary>
    public abstract class PersistenceEngineImplementation : IPersistenceEngine
    {
        public PersistenceEngineImplementation()
        {
        }

        public EntitySet Load(string opath)
        {
            return Load(opath, new string[0]);
        }

        public EntitySet Load(string opath, string[] attributes)
        {
            return Load(opath, attributes, null, 1, 0);
        }

        public EntitySet Load(string opath, string orderby, int first, int max)
        {
            return Load(opath, new string[0], orderby, first, max);
        }

        public virtual EntitySet Load(string opath, string[] attributes, string orderby, int first, int max)
        {
            Evaluant.OPath.OPathQuery query = new Evaluant.OPath.OPathQuery(opath);
            query.Compile();

            if (query.HasErrors)
                throw new Evaluant.OPath.OPathException(query);

            return Load(query, attributes, orderby, first, max);
        }

        public abstract EntitySet Load(OPathQuery opath, string[] attributes, string orderby, int first, int max);

        public EntitySet LoadWithId(string type, string[] id)
        {
            return LoadWithId(type, id, new string[0]);
        }

        public Entity LoadWithId(string type, string id)
        {
            EntitySet result = LoadWithId(type, new string[] { id });
            return result.Count > 0 ? result[0] : null;
        }

        public abstract EntitySet LoadWithId(string type, string[] id, string[] attributes);

        public void LoadReference(Entity entity)
        {
            LoadReference(new Entity[] { entity });
        }

        public void LoadReference(Entity entity, string[] references)
        {
            LoadReference(new Entity[] { entity }, references);
        }

        public void LoadReference(IEnumerable entities)
        {
            LoadReference(entities, new string[0]);
        }

        public abstract void LoadReference(IEnumerable entities, string[] references);

        public virtual void InitializeRepository()
        {
            Initialize();
        }

        public virtual void Initialize()
        {
        }

        public virtual void BeforeProcessCommands(Transaction tx)
        {
            Hashtable processed = new Hashtable();

            foreach (Command c in tx.PendingCommands)
            {
                if (c.ProcessOrder == 2)
                {
                    DeleteEntityCommand command = c as DeleteEntityCommand;

                    // Checks whether it's really a DeleteEntityCommand
                    if (command == null)
                        continue;

                    processed.Add(GetCommandKey(command.ParentId, command.Type), command);
                }
            }

            CommandCollection newCommands = new CommandCollection();
            foreach (Command c in tx.PendingCommands)
            {
                if (c.ProcessOrder == 2)
                {
                    DeleteEntityCommand command = c as DeleteEntityCommand;

                    // Checks whether it's really a DeleteEntityCommand
                    if (command == null)
                        continue;

                    newCommands.AddRange(ComputeCascadeDeleteCommands(command.ParentId, command.Type, processed));
                }
            }

            tx.PendingCommands.AddRange(newCommands);
        }

        /// <summary>
        /// Execute the commands here
        /// </summary>
        /// <param name="commands"></param>
        public abstract void ProcessCommands(Transaction tx);

        public virtual void AfterProcessCommands(Transaction tx)
        {
        }


        private string GetCommandKey(string id, string type)
        {
            return String.Concat(type, ":", id);
        }

        private CommandCollection ComputeCascadeDeleteCommands(string id, string type, Hashtable processed)
        {
            CommandCollection result = new CommandCollection();

            foreach (Models.Reference reference in Model.GetInheritedReferences(type))
            {
                // Search for composition relationships
                if (!reference.IsComposition)
                    continue;

                Entity e = new Entity(type);
                e.Id = id;

                this.LoadReference(e, new string[] { reference.Name });

                foreach (Entity subEntity in e.GetEntitySet(reference.Name))
                {
                    string subKey = GetCommandKey(subEntity.Id, subEntity.Type);

                    if (!processed.Contains(subKey))
                    {
                        DeleteEntityCommand nec = new DeleteEntityCommand(subEntity.Id, subEntity.Type);
                        result.Add(nec);
                        processed.Add(subKey, nec);

                        // Recursive call
                        result.AddRange(ComputeCascadeDeleteCommands(subEntity.Id, subEntity.Type, processed));
                    }
                }
            }

            return result;
        }


        public virtual object LoadScalar(string opath)
        {
            if (!opath.Trim().StartsWith("eval"))
                opath = string.Format("eval({0})", opath);

            Evaluant.OPath.OPathQuery query = new Evaluant.OPath.OPathQuery(opath, OPathQueryTypeEnum.Expression);
            query.Compile();

            if (query.HasErrors)
                throw new Evaluant.OPath.OPathException(query);

            return LoadScalar(query);
        }

        public abstract object LoadScalar(OPath.OPathQuery opath);

        protected Model _Model = new Model();
        public virtual Model Model
        {
            get { return _Model; }
            set { _Model = value; }
        }

        private CultureInfo _Culture;
        public CultureInfo Culture
        {
            get { return _Culture; }
            set { _Culture = value; }
        }

    }
}
