using System;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using System.Text.RegularExpressions;

using Evaluant.Uss;
using Evaluant.Uss.MetaData;
using Evaluant.Uss.DomainModel.OptionModel;
using Evaluant.Uss.SqlMapper;
using Evaluant.Uss.Models;

using Evaluant.Uss.Xmi;
using Evaluant.Uss.Common;
using System.Reflection;
using System.Collections;

namespace Evaluant.Uss.MappingGenerator
{
	/// <summary>
	/// Description rsume de Class1.
	/// </summary>
	class Program
	{
		/// <summary>
		/// Point d'entre principal de l'application.
		/// </summary>
		[STAThread]
		static int Main(string[] args)
		{
			string model = String.Empty;

			ArgParser parser = new ArgParser(args);
			
			if(parser.Exists("?"))
			{
				Help();
				return -1;
			}

			DomainModel.OptionModel.Option option = new Option();
            option.Strategy = InheritanceStrategy.TablePerHierarchy;

            if (parser.Exists("strategy"))
            {
                try
                {
                    option.Strategy = (InheritanceStrategy)Enum.Parse(typeof(InheritanceStrategy), parser["strategy"].Value);
                }
                catch
                {
                    Console.WriteLine("Incorrect parameters : strategy not recognized");
                    Help();
                    return 2;
                }
            }

            if (!parser.Exists("model") && !parser.Exists("metadata") && !parser.Exists("assembly") && option.Metadata.Count == 0)
            {
                Console.WriteLine("Incorrect parameters : You must provide at least a model, some metadata or an assembly");
                Help();
                return 2;
            }

            if (option.MappingFile == null)
            {
                if (!parser.Exists("out"))
                {
                    Console.WriteLine("Incorrect parameters : You must provide an \"out\" parameter");
                    Help();
                    return 2;
                }
                else
                {
                    option.MappingFile = parser["out"].Value;
                }
            }

            Model domainModel = new Model();

            foreach (VVPair pair in parser)
            {
                switch (pair.Variable)
                {
                    case "model":
                        foreach (string val in pair.Value.Split(';'))
                            domainModel.RegisterMetadata(MetaDataFactory.FromModelFile(val));
                        break;

                    case "metadata":
                        foreach (string val in pair.Value.Split(';'))
                            domainModel.RegisterMetadata(MetaDataFactory.FromMetaDataFile(val));
                        break;

                    case "assembly":

                        foreach (string val in pair.Value.Split(';'))
                        {
                            string filename = new FileInfo(val).FullName;
                            domainModel.RegisterMetadata(MetaDataFactory.FromAssembly(Assembly.LoadFrom(filename)));
                        }
                        break;
                }
            }

            if (parser.Exists("constraints"))
                option.Constraint = true;

			try
			{

				Mapping mapping = MappingGenerator.BuildMapping(option.Strategy, domainModel, option);

                // Updates the generated mapping

                // Uses only native ids
                if (parser.Exists("native"))
                {
                    foreach(EntityMapping em in mapping.Entities)
                    {
                        if (em.Ids[0].Generator.Name != GeneratorMapping.GeneratorType.inherited)
                        {
                            em.Ids[0].Generator.Name = GeneratorMapping.GeneratorType.native;
                            em.Ids[0].Generator.Params = new ParamMapping[] { new ParamMapping("dbType", "Int32") };
                        }
                    }
                }

                
                // Apply maximum identifier length
                if (parser.Exists("truncate"))
                {
                    int length = 0;

                    try
                    {
                        length = int.Parse(parser["truncate"].Value);
                    }
                    catch
                    {
                        Console.WriteLine("Incorrect parameters : invalid length");
                        Help();
                        return 3;
                    }

                    mapping.Normalize(length);
                }

                if (parser.Exists("string"))
                {
                    int size = 0;
                    try
                    {
                        size = int.Parse(parser["string"].Value);
                    }
                    catch
                    {
                        Console.WriteLine("Incorrect parameters : invalid string size");
                        Help();
                        return 3;
                    }

                    foreach (EntityMapping em in mapping.Entities)
                    {

                        foreach (AttributeMapping am in em.Attributes)
                        {
                            if ( (am.DbType == System.Data.DbType.AnsiString
                                ||  am.DbType == System.Data.DbType.AnsiStringFixedLength
                                ||  am.DbType == System.Data.DbType.String
                                ||  am.DbType == System.Data.DbType.StringFixedLength)
                                && am.Size != 0)
                            {
                                am.Size = size;
                            }
                        }
                    }
                }

				XmlSerializer mappingSerializer = new XmlSerializer(typeof(Mapping));

				XmlTextWriter writer = null;
				try
				{
					writer = new XmlTextWriter(option.MappingFile, System.Text.Encoding.UTF8);
					writer.Formatting = Formatting.Indented;

					mappingSerializer.Serialize(writer, mapping);
				}
				catch
				{
					throw;
				}
				finally
				{
					if(writer != null)
						writer.Close();
				}

                return 0;
			}
			catch(Exception e)
			{
                Console.WriteLine(e.Message);
				return 8;
			}
		}

        public static string GetTruncatedTerm(Hashtable truncated, ArrayList names, string type, string term, int length)
        {
            if (term == null || term == String.Empty || term.Length <= length)
                return term;

            if (!truncated.Contains(String.Concat(type, ":", term)))
            {
                string newterm = term.Substring(0, length);
                int i = 1;
                while (names.Contains(String.Concat(newterm, i.ToString()))) i++;
                truncated.Add(String.Concat(type, ":", term), String.Concat(newterm, i.ToString()));
                names.Add(String.Concat(newterm, i.ToString()));
            }

            return (string)truncated[String.Concat(type, ":", term)];

        }

		public static string GetSimpleXML (string xmi)
		{
			XmiVersion ver = Converter.GetVersion (xmi);
			Exporter exp = Converter.GetExporter (xmi);

			return Converter.ConvertToSimple (xmi, ver, exp);
		}

		static public void Help()
		{
			Console.WriteLine(@"

Help for mapping.exe

Generates a mapping file from some models metadata.

mapping [/model:model_filenames] [/assembly:assembly_filenames] [/metadata:metadata_filenames] [/constraints] [/native] [/truncate:length] [/strategy:strategy] [/string:size] /out:mapping_filename 

    mapping_filename
        Where the result should be created

	model_filenames
		The Euss model file locations seperated by ';'

	assembly_filenames
		The .Net assemblies locations seperated by ';'

    metadata_filenames
        The Euss metadata locations separated by ';'

    native
        If specified, uses native identifiers instead of guids

    truncate
        The maximum length of names.

    size
        The default size applied to string fields

    strategy
        The name of the inheritance strategy to use. Allowed values: TablePerHierarchy, TablePerConcreteClass, TablePerSubClass

    
    constraints
        If specified, the constraints will be generated
");

		}
	}
}
