using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using MyNamespace;
using System.IO;
using System.Collections.Generic;

using Evaluant.Uss;
using Evaluant.Uss.MetaData;
using Evaluant.Uss.ObjectContext;


namespace TestPersistenceManager
{
	/// <summary>
	/// Description rsume de Class1.
	/// </summary>
	[TestClass]
	public class TestGenerated
	{
		public ObjectContext pm;
		
		public TestGenerated()
		{
            Configure();
		}

        public virtual void Configure()
        {
            ObjectService factory = new ObjectService(Path.GetDirectoryName(typeof(TestGenerated).Assembly.Location) + "\\engines.config", "Prevalence"); // Xml //SqlMapperTestGenerated //OracleTestGenerated
            factory.RegisterMetaData(MetaDataFactory.FromModelFile(@"..\..\..\NUnit\PersistenceManager\HardCodedAndExplicit\schema.xml"));
            this.pm = factory.CreateObjectContext();
        }

		[TestInitialize]
		public void Init()
		{
			pm.InitializeRepository(); // Vide la base de donnes ou le fichier XML
		}

        [TestMethod]
        public virtual void NonStandardType()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "John";
            p1.Picture = new System.Drawing.Bitmap(200, 100);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            pm.Clear();

            Person p = (Person)pm.Load<Person>(typeof(Person))[0];

            Assert.IsNotNull(p);
            Assert.IsNotNull(p.Picture);
            Assert.IsTrue(p.Picture.GetType().IsSubclassOf(typeof(System.Drawing.Image)));
            Assert.AreEqual(200, p.Picture.Width);
            Assert.AreEqual(100, p.Picture.Height);
        }

		[TestMethod]
		public void TestCompositionDelete()
		{		
			Dog d1 = new Dog();
		
			pm.BeginTransaction();
			pm.Serialize(d1);
			pm.CommitTransaction();
            pm.Clear();

			Dog dog = pm.Load<Dog>(typeof(Dog))[0];
			
			pm.BeginTransaction();
			pm.Delete(dog);
			pm.CommitTransaction();

		}

        [TestMethod]
        public void NullableTypes()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "FirstName";
            p1.Age = 12;

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            p1 = pm.Load<Person>(typeof(Person))[0];
            Assert.AreEqual(12, p1.Age);
            Assert.IsNull(p1.Birthday);
            Assert.IsTrue(pm.IsNull(p1, "Birthday"));
            Assert.IsFalse(pm.IsNull(p1, "Age"));

            p1.Age = null;
            p1.Birthday = DateTime.Now;

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            p1 = pm.Load<Person>(typeof(Person))[0];
            Assert.AreEqual(null, p1.Age);
            Assert.IsNotNull(p1.Birthday);
            Assert.IsFalse(pm.IsNull(p1, "Birthday"));
            Assert.IsTrue(pm.IsNull(p1, "Age"));

        }

		[TestMethod]
		public void Update()
		{
            pm.InitializeRepository();

			Person p1 = new Person();
			p1.FirstName = "FirstName";
			p1.Age = 12;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			p1 = pm.Load<Person>(typeof(Person))[0];
			p1.FirstName = "Modified";
			p1.Age = 13;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			p1 = pm.Load<Person>(typeof(Person))[0];

			Assert.AreEqual("Modified", p1.FirstName);
			Assert.AreEqual(13, p1.Age);
		}


		[TestMethod]
		public void TestEmptyClass()
		{
            Person p1 = new Person();
            p1.FirstName = "FirstName";

            Company c1 = new Company();
            c1.logo = "logo";

            c1.Employees.Add(p1);

            pm.BeginTransaction();
            pm.Serialize(c1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(1, pm.Load<Person>(typeof(Company), ".Employees").Count);
            c1 = pm.Load<Company>(typeof(Company))[0];

            Assert.AreEqual(1, c1.Employees.Count);
            Assert.AreEqual("FirstName", c1.Employees[0].FirstName);
        }
		
		[TestMethod]
		public void TestInterface()
		{		
			Person p1 = new Person();
			p1.FirstName = "lon";
			p1.Code = "the killer";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
            pm.Clear();

			Person p = pm.Load<Person>(typeof(Person), "[FirstName = 'lon']")[0];
			Assert.AreEqual(p.Code, "the killer");
		}


		[TestMethod]
		public void TestEnumeration()
		{
			Person anonyme = new Person();
			anonyme.FirstName ="gerard";
			anonyme.Title = Title.Sir;

			pm.BeginTransaction();
			pm.Serialize(anonyme);
			pm.CommitTransaction();
            pm.Clear();

			Person p = pm.Load<Person>(typeof(Person), "[FirstName = 'gerard']")[0];
			Assert.AreEqual(Title.Sir, p.Title);
		}

        [TestMethod]
        public void TestReferenceCyclique()
        {

            Person p1 = new Person();
            p1.FirstName = "p1";

            Person p2 = new Person();
            p2.FirstName = "p2";

            p1.Friends.Add(p2);
            p2.Friends.Add(p1);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Person sp1 = pm.Load<Person>(typeof(Person), "[FirstName = 'p1']")[0];
            Person sp2 = pm.Load<Person>(typeof(Person), "[FirstName = 'p2']")[0];

            Assert.AreEqual(sp1.Friends[0].FirstName, "p2");
            Assert.AreEqual(sp2.Friends[0].FirstName, "p1");
            Assert.IsTrue(sp1.Friends[0] == sp2);
            Assert.IsTrue(sp2.Friends[0] == sp1);
        }

        [TestMethod]
        public void TestComposition()
        {
            Library l = new Library();
            l.name = "bookshop";
            
            Book book = new Book();
            book.name = "le petit poucet";

            Shelf shelf = new Shelf();
            shelf.name = "big one";

            l.MyBooks.Add(book);
            l.Shelf = shelf;
            
            pm.BeginTransaction();
            pm.Serialize(l);
            pm.CommitTransaction();

            pm.BeginTransaction();
            pm.Delete(l);
            pm.CommitTransaction();
            
            pm.Clear();

            Assert.AreEqual(0, pm.Load<Library>(typeof(Library)).Count);
            Assert.AreEqual(0, pm.Load<Book>(typeof(Book)).Count);
            Assert.AreEqual(0, pm.Load<Shelf>(typeof(Shelf)).Count);
        }

        [TestMethod]
        public void TestComposition2()
        {
            Library l = new Library();
            l.name = "bookshop";

            Book book = new Book();
            book.name = "le petit poucet";

            Shelf shelf = new Shelf();
            shelf.name = "big one";

            l.MyBooks.Add(book);
            l.Shelf = shelf;

            pm.BeginTransaction();
            pm.Serialize(l);
            pm.CommitTransaction();

            pm.BeginTransaction();
            pm.Delete(book);
            pm.CommitTransaction();

            pm.BeginTransaction();
            pm.Delete(shelf);
            pm.CommitTransaction();

            pm.Clear();

            Assert.AreEqual(1, pm.Load<Library>(typeof(Library)).Count);
            Assert.AreEqual(0, pm.Load<Book>(typeof(Book)).Count);
            Assert.AreEqual(0, pm.Load<Shelf>(typeof(Shelf)).Count);
        }

        [TestMethod]
        public void TestCreation()
        {
            Person p1 = new Person();
            //Vrification que l'entit a t cre
            Assert.IsNotNull(((IPersistable)p1).Entity);

            p1.FirstName = "jean";
            Assert.AreEqual("jean", p1.FirstName);

            Company c1 = new Company();
            c1.logo = "evaluant";

            //Vrification que la collection d'employs appartenant  p1 est vide
            Assert.AreEqual(0, (p1.Friends).Count);
        }


        [TestMethod]
        public void TestIdentityMap()
        {
            pm.InitializeRepository();

            Person p = new Person();
            p.FirstName = "Eric";

            pm.BeginTransaction();
            pm.Serialize(p);
            pm.CommitTransaction();

            pm.Clear();

            Person p1 = pm.Load<Person>(typeof(Person))[0];
            Person p2 = pm.Load<Person>(typeof(Person))[0];

            Assert.IsTrue(p1 == p2);
        }

        [TestMethod]
        public void TestCount()
        {
            pm.InitializeRepository();

            Person p = new Person();
            p.FirstName = "Eric";

            pm.BeginTransaction();
            pm.Serialize(p);
            pm.CommitTransaction();

            Assert.AreEqual(1, pm.Load<Person>(typeof(Person)).Count);
        }

        [TestMethod]
        public void GetParents()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "Eric";

            Person p2 = new Person();
            p2.FirstName = "John";

            p1.Friends.Add(p2);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Person john = pm.Load<Person>(typeof(Person), "[FirstName = 'John']")[0];

            Assert.AreEqual(1, pm.LoadParents<Person>(john, "Friends").Count);
            Assert.AreEqual("Eric", pm.LoadParents<Person>(john, "Friends")[0].FirstName);
        }

        [TestMethod]
        public void TestDelete()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "Eric";

            Person p2 = new Person();
            p2.FirstName = "John";

            p1.Friends.Add(p2);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);

            pm.BeginTransaction();
            pm.Delete(p1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(1, pm.Load<Person>(typeof(Person)).Count);
        }

        [TestMethod]
        public void TestDispose()
        {
            Person p1 = new Person();
            Person p2 = new Person();
            p1.FirstName = "sylvain";
            p2.FirstName = "alex";

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.Serialize(p2);
            pm.CommitTransaction();

            IPersistable persist1 = pm.Load<Person>(typeof(Person), "[FirstName = 'sylvain']")[0];

            pm.Clear();

            IPersistable persist2 = pm.Load<Person>(typeof(Person), "[FirstName = 'sylvain']")[0];

            // Vrification que les rfrences sont diffrentes avant et aprs suppression
            Assert.IsTrue(persist1 != persist2);
        }

        [TestMethod]
        public void TestInheritance()
        {
            pm.InitializeRepository();

            Company c = new Company();
            SARL s = new SARL();

            pm.BeginTransaction();
            pm.Serialize(c);
            pm.Serialize(s);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(2, pm.Load<Company>(typeof(Company)).Count);
            Assert.AreEqual(1, pm.Load<SARL>(typeof(SARL)).Count);

            // Make the test again to check the types cache is not buggy

            Assert.AreEqual(2, pm.Load<Company>(typeof(Company)).Count);
            Assert.AreEqual(1, pm.Load<SARL>(typeof(SARL)).Count);

        }

        [TestMethod]
        public void TestCollection()
        {

            PersonCollection people = new PersonCollection();

            Person p = new Person();
            p.FirstName = "toto";

            people.Add(new Person());
            people.Add(new Person());
            people.Add(new Person());
            people.Add(p);

            Assert.AreEqual(4, people.Count);
            Assert.IsNotNull(people[3]);
            Assert.AreEqual("toto", people[3].FirstName);

            pm.BeginTransaction();
            pm.Serialize(people);
            pm.CommitTransaction();

        }

        [TestMethod]
        public void TestReference2()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "John";

            p1.Friends.Add(new Person());

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), ".Friends").Count);

            pm.Clear();

            p1 = pm.Load<Person>(typeof(Person), "[FirstName = 'John']")[0];

            p1.Friends.Add(new Person());

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            pm.Clear();

            p1 = pm.Load<Person>(typeof(Person), "[FirstName = 'John']")[0];

            Assert.AreEqual(2, p1.Friends.Count);
        }

        [TestMethod]
        public void LazyLoading()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "John";

            Person p2 = new Person();
            p2.FirstName = "Eric";
            p2.Age = 24;

            p1.Friends.Add(p2);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            pm.Clear();

            p1 = pm.Load<Person>(typeof(Person), "[FirstName = 'John']")[0];

            Assert.IsNotNull(p1.Friends);
            Assert.AreEqual(1, p1.Friends.Count);
            Assert.AreEqual("Eric", p1.Friends[0].FirstName);
            Assert.AreEqual(24, p1.Friends[0].Age);

        }

        [TestMethod]
        public void EagerLoading()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "John";

            Person p2 = new Person();
            p2.FirstName = "Eric";
            p2.Age = 24;

            Person p3 = new Person();
            p3.FirstName = "Mark";
            p3.Age = 24;

            p1.Friends.Add(p2);
            p1.Deputy = p3;

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            int c1 = p1.Friends.Count;
            int c2 = p2.Friends.Count;
            object obj1 = p1.Deputy;
            object obj2 = p2.Deputy;


            //pm.Clear();

            Query q = new Query(typeof(Person), "[FirstName = 'John']");
            q.InferredRelationships.Add("Friends");
            q.InferredRelationships.Add("Deputy");

            p1 = pm.Load<Person>(q)[0];

            // The code above should not execute a query (loadReference)

            Assert.IsNotNull(p1.Friends);
            Assert.AreEqual(1, p1.Friends.Count);
            Assert.AreEqual("Eric", p1.Friends[0].FirstName);
            Assert.AreEqual(24, p1.Friends[0].Age);
            Assert.AreEqual("Mark", p1.Deputy.FirstName);

            //------------------------------------------
            // test with empty references without clearing the cache (loadReference)

            q = new Query(typeof(Person), "[FirstName = 'Eric']");
            q.InferredRelationships.Add("Friends");
            q.InferredRelationships.Add("Deputy");

            p2 = pm.Load<Person>(q)[0];

            // The code above should not execute a query

            Assert.IsNotNull(p2.Friends);
            Assert.AreEqual(0, p2.Friends.Count);
            Assert.IsNull(p2.Deputy);

            //------------------------------------------
            // test with empty references after clearing the cache
            pm.Clear();

            q = new Query(typeof(Person), "[FirstName = 'Mark']");
            q.InferredRelationships.Add("Friends");
            q.InferredRelationships.Add("Deputy");

            p2 = pm.Load<Person>(q)[0];

            // The code above should not execute a query (loadReference)

            Assert.IsNotNull(p2.Friends);
            Assert.AreEqual(0, p2.Friends.Count);
            Assert.IsNull(p2.Deputy);

            //------------------------------------------
            // test with none empty references after clearing the cache
            pm.Clear();

            q = new Query(typeof(Person), "[FirstName = 'John']");
            q.InferredRelationships.Add("Friends");
            q.InferredRelationships.Add("Deputy");

            p1 = pm.Load<Person>(q)[0];

            // The code above should not execute a query

            Assert.IsNotNull(p1.Friends);
            Assert.AreEqual(1, p1.Friends.Count);
            Assert.AreEqual("Eric", p1.Friends[0].FirstName);
            Assert.AreEqual(24, p1.Friends[0].Age);
            Assert.AreEqual("Mark", p1.Deputy.FirstName);
        }

        [TestMethod]
        public void ReloadEntity()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "John";

            Person p2 = new Person();
            p2.FirstName = "Eric";
            p2.Age = 24;

            Person p3 = new Person();
            p3.FirstName = "Mark";
            p3.Age = 24;

            p1.Friends.Add(p2);
            p1.Deputy = p3;

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            pm.Clear();

            p1 = pm.LoadSingle<Person>(typeof(Person), "[FirstName = 'John']");
            p2 = pm.LoadSingle<Person>(typeof(Person), "[FirstName = 'Eric']");
            Assert.IsNotNull(p2);

            p2.Age = 10;

            // Loads p2 not directly
            Assert.AreEqual(1, p1.Friends.Count);

            // The lazy loading should not have refreshed the properties of p2
            Assert.AreEqual(10, p2.Age);
        }

        [TestMethod]
        public void OneToOne()
        {
            pm.InitializeRepository();

            // Instanciation
            Person p1 = new Person();
            p1.FirstName = "p1";

            Person p2 = new Person();
            p2.FirstName = "p2";

            p1.Deputy = p2;

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1']").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Deputy").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Deputy[FirstName = 'p2']").Count);

            // New Association
            Person p3 = new Person();
            p3.FirstName = "p3";

            p1.Deputy = p3;
            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(3, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1']").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Deputy").Count);
            Assert.AreEqual(0, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Deputy[FirstName = 'p2']").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Deputy[FirstName = 'p3']").Count);

        }

        [TestMethod]
        public void OneToMany()
        {
            pm.InitializeRepository();

            // Instanciation
            Person p1 = new Person();
            p1.FirstName = "p1";

            Person p2 = new Person();
            p2.FirstName = "p2";

            Person p3 = new Person();
            p3.FirstName = "p3";

            p1.Friends.Add(p2);
            p1.Friends.Add(p3);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(3, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1']").Count);
            Assert.AreEqual(2, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']").Count);

            // New Association
            Person p4 = new Person();
            p4.FirstName = "p4";

            p1.Friends[0] = p4;
            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(4, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1']").Count);
            Assert.AreEqual(2, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends").Count);
            Assert.AreEqual(0, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p4']").Count);

            p1.Friends.Clear();
            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(4, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1']").Count);
            Assert.AreEqual(0, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends").Count);
        }

        [TestMethod]
        public void ManyToMany()
        {
            pm.InitializeRepository();

            // Instanciation
            Person p1 = new Person();
            p1.FirstName = "p1";

            Person p2 = new Person();
            p2.FirstName = "p2";

            Person p3 = new Person();
            p3.FirstName = "p3";

            p1.Friends.Add(p2);
            p1.Friends.Add(p3);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(3, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1']").Count);
            Assert.AreEqual(2, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p3']").Count);

            Person p4 = new Person();
            p4.FirstName = "p4";

            Person p5 = new Person();
            p5.FirstName = "p5";

            p1.Friends.Clear();
            p1.Friends.AddRange(new Person[] { p4, p5 });

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();
            pm.Clear();

            Assert.AreEqual(5, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1']").Count);
            Assert.AreEqual(2, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends").Count);
            Assert.AreEqual(0, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']").Count);
            Assert.AreEqual(0, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p3']").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p4']").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p5']").Count);

        }

        [TestMethod]
        public void Reference()
        {
            // Instanciation
            Person p1 = new Person();
            p1.FirstName = "p1";

            Person p2 = new Person();
            p2.FirstName = "p2";

            Person p3 = new Person();
            p3.FirstName = "p3";

            p1.Friends.Add(p2);
            p1.Friends.Add(p3);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            Person x1 = pm.Load<Person>(typeof(Person), "[FirstName = 'p1']")[0] as Person;

            Assert.AreEqual(p1, x1);
            Assert.AreEqual(((IPersistable)p1).Entity, ((IPersistable)x1).Entity);

            Person x2 = pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']")[0] as Person;
            Assert.AreEqual(p2, x2);
            Assert.AreEqual(((IPersistable)p2).Entity, ((IPersistable)x2).Entity);

            if (x1.Friends[0].FirstName == "p2")
            {
                Assert.AreEqual(p2, x1.Friends[0]);
                Assert.AreEqual(((IPersistable)p2).Entity, ((IPersistable)x1.Friends[0]).Entity);
            }

            if (x1.Friends[0].FirstName == "p3")
            {
                Assert.AreEqual(p3, x1.Friends[0]);
                Assert.AreEqual(((IPersistable)p3).Entity, ((IPersistable)x1.Friends[0]).Entity);
            }
        }

        [TestMethod]
        public void DefaultValue()
        {
            Person p = new Person();
            Assert.IsNull(p.FirstName);
            Assert.IsNull(p.Age);
        }

        [TestMethod]
        public void Resource()
        {
            ObjectService ipef;

            try
            {
                ipef = new ObjectService("NUnit.engines.config");
            }
            catch
            {
                ipef = new ObjectService(typeof(TestGenerated).Assembly.GetManifestResourceStream("NUnit.engines.config"));
            }

            Assert.IsNotNull(ipef);
        }

        [TestMethod]
        public void RemoveRelationship()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "p1";

            Person p2 = new Person();
            p2.FirstName = "p2";

            Person p3 = new Person();
            p3.FirstName = "p3";

            p1.Friends.Add(p2);
            p1.Friends.Add(p3);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            pm.Clear();

            Assert.AreEqual(3, pm.Load<Person>(typeof(Person)).Count);

            p1 = pm.LoadSingle<Person>(typeof(Person), "[FirstName = 'p1']");

            Assert.AreEqual(2, p1.Friends.Count);

            p1.Friends.Clear();

            Assert.AreEqual(0, p1.Friends.Count);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            Assert.AreEqual(3, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(0, pm.Load<Person>(typeof(Person), "[FirstName = 'p1'].Friends").Count);

            p1 = pm.LoadSingle<Person>(typeof(Person), "[FirstName = 'p1']");

            Assert.AreEqual(0, p1.Friends.Count);
        }

        [TestMethod]
        public void ForEach()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "p1";

            foreach (Person p in p1.Friends)
                Assert.Fail("Collection should be empty");

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            foreach (Person p in p1.Friends)
                Assert.Fail("Collection should be empty");

            pm.Clear();
            p1 = pm.LoadSingle<Person>(typeof(Person));
            
            foreach (Person p in p1.Friends)
                Assert.Fail("Collection should be empty");

        }

        [TestMethod]
        public void ShouldSerializeSubLevels()
        {
            pm.InitializeRepository();

            Person p1 = new Person();
            p1.FirstName = "p1";

            Person p2 = new Person();
            p2.FirstName = "p2";

            Person p3 = new Person();
            p3.FirstName = "p3";

            Person p4 = new Person();
            p4.FirstName = "p4";

            Person p5 = new Person();
            p5.FirstName = "p5";

            p1.Friends.Add(p2);

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);

            p2.Friends.Add(p3);
            p3.Friends.Add(p4);

            pm.BeginTransaction();
            pm.Serialize(p2);
            pm.CommitTransaction();

            Assert.AreEqual(4, pm.Load<Person>(typeof(Person)).Count);
        }

		[TestCleanup]
		public void Final()
		{
		}
			
	}


	
}
