using System;
using System.Collections;
using System.IO;
using System.Collections.Generic;
using NUnit.PersistenceManager.DynamicProxy;
using NUnit.DynamicProxy;
using System.Linq;

using Evaluant.Uss;
using Evaluant.Uss.ObjectContext;
using Evaluant.Uss.Linq;
using System.Linq.Expressions;
using Microsoft.VisualStudio.TestTools.UnitTesting;


namespace NUnit.Linq.Tests
{
    /// <summary>
    /// Description rsume de Proxy.
    /// </summary>
    [TestClass]
    public class TestFixtures
    {
        public static ObjectService os;

        public Evaluant.Uss.ObjectContext.ObjectContext oc;

        public TestFixtures()
        {
            Configure();
            InitializeData();
        }

        IPersistenceProvider provider = XmlConfigLoader.LoadXmlConfig("engines.config", "Prevalence");

        public virtual void Configure()
        {
            provider.RegisterMetaData(MetaDataFactory.FromAssembly(typeof(Person).Assembly));
            provider.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\..\NUnit\PersistenceEngine\Models\Person.xml"));

            provider.InitializeConfiguration();

            os = new ObjectService(provider);
            oc = os.CreateObjectContext();
        }

        [TestInitialize]
        public void InitializeData()
        {
            oc.InitializeRepository();

            Person bill = new Person();
            bill.Name = "Bill";
            bill.Age = 53;
            bill.Birth = new DateTime(1955, 10, 28);
            bill.Kind = KindEnum.Nice;
            bill.Salary = 10.3f;
            bill.Title = TitleEnum.Sir;
            bill.IsRetired = true;
            bill.Address = new Address("New York");

            Person steve = new Person();
            steve.Name = "Steve";
            steve.Age = 52;
            steve.Birth = new DateTime(1956, 3, 28);
            steve.Kind = KindEnum.Bad;
            steve.Salary = 9.6f;
            steve.Title = TitleEnum.Madam;
            steve.IsRetired = false;
            steve.Address = new Address("Chicago");

            bill.BestFriend = steve;

            bill.Employees = new List<Employee>() {
                    new Employee() { Name = "John O'Shea", Age = 20, Salary=  0.1f,Address=new Address("Boston") },
                    new Employee() { Name = "Employee 2", Age = 30, Salary=  0.2f,Address=new Address("Washington") },
                    new Employee() { Name = "Employee 3", Age = 40, Salary=  0.3f,Address=new Address("Dallas") },
                    new Employee() { Name = "Employee 4", Age = 50, Salary=  0.4f, IsRetired = true,Address=new Address("Los Angeles") },
            };

            steve.Employees = new List<Employee>() {
                    new Employee() { Name = "Employee 5", Age = 25, Salary=  0.5f,Address=new Address("Chicago"),BestFriend=bill },
                    new Employee() { Name = "Employee 6", Age = 55, Salary=  0.6f,Address=new Address("New York") }
            };

            oc.BeginTransaction();
            oc.Serialize(bill);
            oc.Serialize(steve);
            oc.CommitTransaction();

            oc.Clear();
        }
        [TestMethod]
        public void ShouldHandleDateTimes()
        {
            var people = (from Person p in oc
                          orderby p.Birth
                          select p).ToList();

            Assert.AreEqual("Bill", people[0].Name);
            Assert.AreEqual("Steve", people[1].Name);

            people = (from Person p in oc
                      orderby p.Birth descending
                      select p).ToList();

            Assert.AreEqual("Bill", people[7].Name);
            Assert.AreEqual("Steve", people[6].Name);

            people = (from Person p in oc
                      where p.Birth > new DateTime(1955, 1, 1)
                      where p.Birth < new DateTime(1956, 1, 1)
                      select p).ToList();

            Assert.AreEqual("Bill", people[0].Name);
        }

        [TestMethod]
        public void ShouldSelectEmployees()
        {
            var q = from Employee e in oc
                    select e;

            Assert.AreEqual(6, q.Count());


            var q2 = from Person p in oc
                     select p.Employees;

            Assert.AreEqual(6, q2.Count());
        }

        [TestMethod]
        public void ShouldTakeOnePerson()
        {
            var q = from Person p in oc
                    select p;

            Assert.AreEqual(1, q.Take(1).ToList().Count);
        }

        [TestMethod]
        public void ShouldConvertValuesExplicitly()
        {
            var q = from Person p in oc
                    where p.Salary == 10.3 // double should be casted to float
                    select p.Name;

            Assert.AreEqual("Bill", q.First());
        }

        [TestMethod]
        public void ShouldEscapeSpecialCharacters()
        {
            var q = from Person p in oc
                    where p.Name.Contains("O'Shea")
                    select p.Name;

            Assert.AreEqual("John O'Shea", q.First());
        }

        [TestMethod]
        public void ShouldLoadUsingEnums()
        {
            var q = from Person p in oc
                    where p.Title == TitleEnum.Sir
                    select p.Name;

            Assert.AreEqual("Bill", q.First());
        }

        [TestMethod]
        public void ShouldHandleToOneRelationship()
        {
            Person steve = (from Person p in oc where p.Name == "Steve" select p).First();

            Assert.AreEqual("Bill", (from Person p in oc
                                     where p.BestFriend.Name == "Steve"
                                     select p.Name).First());

            Assert.AreEqual("Bill", (from Person p in oc
                                     where p.BestFriend.PersonId == steve.PersonId
                                     select p.Name).FirstOrDefault());

            Assert.AreEqual("Bill", (from Person p in oc
                                     where p.BestFriend.PersonId == steve.PersonId
                                     where p.IsRetired == true
                                     select p.Name).FirstOrDefault());

            Address a = (from Person p in oc
                         where p.Name == "Steve"
                         select p.Address).FirstOrDefault();

            Assert.IsNotNull(a);
            Assert.AreEqual("Chicago", a.City);

            string name = (from Person p in oc
                           from e in p.Employees
                           where e.Address.City == "New York"
                           select p).First().Name;

            Assert.IsTrue(name == "Steve" || name == "Employee 6", "name should be Steve or Employee 6 but was {0}", name);

            // Currently this test is not handled but could be by splitting the && into two Where() clauses

            //Assert.AreEqual("Bill", (from Person p in oc
            //                         where p.IsRetired == true && p.BestFriend.PersonId == steve.PersonId
            //                         select p.Name).First()); 
        }

        [TestMethod]
        public void ShouldHandleAllBooleanConstructions()
        {

            Assert.AreEqual("Employee 4", (from Person p in oc
                                           where p.IsRetired
                                           orderby p.Name descending
                                           select p.Name).First());

            List<Person> ps = (from Person p in oc
                               where p.IsRetired
                               orderby p.Name descending
                               select p).ToList();

            Assert.AreEqual("Employee 4", ps[0].Name);

            Assert.AreEqual("Bill", ps[1].Name);

            Assert.AreEqual("Steve", (from Person p in oc
                                      where p.IsRetired == false
                                      orderby p.Name descending
                                      select p.Name).First());

            Assert.AreEqual("Employee 4", (from Person p in oc
                                           where p.IsRetired == true
                                           orderby p.Name descending
                                           select p.Name).First());

            Assert.AreEqual("Steve", (from Person p in oc
                                      where !p.IsRetired
                                      orderby p.Name descending
                                      select p.Name).First());

            Assert.AreEqual("Bill", (from Person p in oc
                                     where p.IsRetired && p.Name == "Bill"
                                     select p.Name).First());

            Assert.AreEqual("Bill", (from Person p in oc
                                     where p.IsRetired == true && p.Name == "Bill"
                                     select p.Name).First());

        }

        [TestMethod]
        public void ShouldTransformCascaginWhere()
        {
            Assert.AreEqual(2, (from Person p in oc
                                where p.IsRetired
                                select p.Name).Count());

            Assert.AreEqual(1, (from Person p in oc
                                where p.Name == "Bill"
                                select p.Name).Count());

            Assert.AreEqual(1, (from Person p in oc
                                where p.Name == "Bill"
                                where p.IsRetired
                                select p.Name).Count());

        }

        [TestMethod]
        public void ShouldReturnTheCountOfAllPerson()
        {
            var q = (from Person p in oc select p).Count();
            Assert.AreEqual(8, q);

            q = (from Person p in oc select p).Count(p => p.Name == "Bill");
            Assert.AreEqual(1, q);

        }

        [TestMethod]
        public void ShouldReturnTheMaximumSalary()
        {
            Assert.AreEqual(10.3f, (from Person p in oc
                                   select p).Max(p => p.Salary));
        }

        [TestMethod]
        public void ShouldReturnTheMinimumSalary()
        {
            Assert.AreEqual(0.1f, (from Person p in oc
                                  select p).Min(p => p.Salary));
        }

        [TestMethod]
        public void ShouldReturnTheSumSalary()
        {
            Assert.AreEqual(22f, (from Person p in oc
                                 select p).Sum(p => p.Salary));
        }

        [TestMethod]
        public void ShouldReturnTheAverageSalary()
        {
            Assert.AreEqual(2.75f, (from Person p in oc
                                    select p).Average(p => p.Salary));
        }

        [TestMethod]
        public void ShouldImplementFirst()
        {
            var bill = (from Person p in oc
                        orderby p.Name
                        select p).First();

            Assert.IsNotNull(bill);

            try
            {
                bill = (from Person p in oc
                        where p.Name == "Billy"
                        orderby p.Name
                        select p).First();
            }
            catch (InvalidOperationException e)
            {
                return;
            }

            Assert.Fail();
        }

        [TestMethod]
        public void ShouldImplementFirstOrDefault()
        {
            var bill = (from Person p in oc
                        orderby p.Name
                        select p).FirstOrDefault();

            Assert.IsNotNull(bill);

            bill = (from Person p in oc
                    where p.Name == "Billy"
                    orderby p.Name
                    select p).FirstOrDefault();

            Assert.IsNull(bill);
        }

        [TestMethod]
        public void ShouldLoadFromId()
        {
            var bill = (from Person p in oc
                        where p.Name == "Bill"
                        select p).First();

            var id = oc.GetId(bill);

            Assert.AreEqual("Bill", (from Person p in oc where p.PersonId == id select p.Name).First());

            Assert.AreEqual(0, (from name in (from Person p in oc where p.PersonId != id select p.Name).ToArray() where name == "Bill" select name).Count());
        }

        [TestMethod]
        public void ShouldPageResults()
        {
            var employees = (from Employee e in oc
                             orderby e.Name
                             select e).Skip(2).Take(3).ToList();

            Assert.AreEqual(3, employees.Count());
            Assert.AreEqual("Employee 4", employees[0].Name);
            Assert.AreEqual("Employee 5", employees[1].Name);
        }

        [TestMethod]
        public void ShouldSortResults()
        {
            var employees = (from Employee e in oc
                             orderby e.Name descending
                             select e).Skip(2).Take(3).ToList();

            Assert.AreEqual(3, employees.Count());
            Assert.AreEqual("Employee 5", employees[0].Name);
            Assert.AreEqual("Employee 4", employees[1].Name);
        }

        [TestMethod]
        public void ShouldReturnAllEmployeesOfPersons()
        {
            var employees = (from Person p in oc
                             from Employee e in p.Employees
                             select e).ToList();

            Assert.AreEqual(6, employees.Count());

            Assert.AreEqual(6, (from Person p in oc
                                from Employee e in p.Employees
                                select e).Count());

        }

        [TestMethod]
        public void ShouldReturnAllEmployeesForBill()
        {
            var employees = (from Person p in oc
                             where p.Name == "Bill"
                             from Employee e in p.Employees
                             select e).OrderBy(e => e.Name);

            Assert.AreEqual(4, employees.Count());

            var employees2 = from Person p in oc
                             where p.Name == "Bill"
                             from Employee e in p.Employees
                             orderby e.Name
                             select e;

            Assert.AreEqual(4, employees2.Count());

            IEnumerable<Employee> es = employees2.ToList();

            Assert.AreEqual(4, es.Count());


        }

        [TestMethod]
        public void ShouldReturnAllFilteredSubItems()
        {
            var employees = (from Person p in oc
                             where p.Name == "Bill"
                             from Employee e in p.Employees
                             select e).Where(e => e.Name.StartsWith("Employee")).ToList();

            Assert.AreEqual(3, employees.Count());

            employees =
                        (from Person p in oc
                         where p.Name == "Bill"
                         from Employee e in p.Employees
                         where e.Name.StartsWith("Employee")
                         select e).ToList();

            // Transformed to 
            // oc.Where(p => p.Name == "Bill")
            // .SelectMany(p => p.Employees, , (p, e) => new { p, e })
            // .Where(x => x.e.Name.StartsWith("Employee"))
            // .Select(x => x.e);

            Assert.AreEqual(3, employees.Count());
        }

        [TestMethod]
        public void ShouldReturnEmployeesFilteredByName()
        {
            var employees = (from Employee e in oc
                             where e.Name == ("Employee 2")
                             select e).ToList();

            Assert.AreEqual(1, employees.Count());
        }

        [TestMethod]
        public void ShouldReturnHandleContains()
        {
            var employees = (from Employee e in oc
                             where e.Name.Contains("Employee")
                             select e).ToList();

            Assert.AreEqual(5, employees.Count());
        }

        [TestMethod]
        public void ShouldReturnHandleStartsWith()
        {
            var employees = (from Employee e in oc
                             where e.Name.StartsWith("Employee")
                             select e).ToList();

            Assert.AreEqual(5, employees.Count());
        }

        [TestMethod]
        public void ShouldReturnHandleEndsWith()
        {
            var employees = (from Employee e in oc
                             where e.Name.EndsWith("2")
                             select e).ToList();

            Assert.AreEqual(1, employees.Count());
        }

        [TestMethod]
        public void ShouldReturnAnonymousTypes()
        {
            var employees1 = (from Employee e in oc select e.Name).ToList();
            var employees2 = (from Employee e in oc select new { e.Name, e.Kind }).ToList();
        }

        [TestMethod]
        public void ShouldBeAbleToReturnDetachedObjects()
        {
            oc = os.CreateObjectContext();

            QueryableEussData<Employee> queryable = new QueryableEussData<Employee>(oc, true);

            var employees = (from e in queryable
                             select e).ToList();

            Assert.AreEqual(6, employees.Count());
            Assert.AreEqual(typeof(Employee), employees[0].GetType());

            oc = os.CreateObjectContext();

            employees = (from e in new QueryableEussData<Employee>(oc, false)
                         select e).ToList();

            Assert.AreEqual(6, employees.Count());
            Assert.IsTrue(typeof(Employee) != employees[0].GetType());

            employees = (from e in new QueryableEussData<Employee>(oc)
                         select e).ToList();

            Assert.AreEqual(6, employees.Count());
            Assert.IsTrue(typeof(Employee) != employees[0].GetType());

        }

        public IQueryable<T> Find<T>(Expression<Func<T, bool>> expression)
        {
            return new QueryableEussData<T>(oc).Where(expression);
        }

        [TestMethod]
        public void ShouldHandleInheritance()
        {
            var query = Find<Person>(p => p.IsRetired).OrderBy(p => p.Name);

            Assert.AreEqual(2, query.Count());

            var people = query.ToList();

            Assert.IsTrue(typeof(Person).IsAssignableFrom(people[0].GetType()));
            Assert.IsTrue(typeof(Employee).IsAssignableFrom(people[1].GetType()));

            Assert.AreEqual("Bill", people[0].Name);
            Assert.AreEqual("Employee 4", people[1].Name);

            var employee = Find<Person>(p => p.IsRetired).OrderByDescending(p => p.Name).FirstOrDefault();

            Assert.IsNotNull(employee);
            Assert.AreEqual("Employee 4", employee.Name);
        }
    }
}
