using System;
using System.Collections;

using NUnit.Framework;
using UniversalStorageServices.PersistenceManager;
using UniversalStorageServices;


namespace NUnitWeaver.Weaving
{
	/// <summary>
	/// Description rsume de Proxy.
	/// </summary>
	[TestFixture]
	public class Weaver
	{

		public UniversalStorageServices.PersistenceManager.PersistenceManager pm;

		public Weaver()
		{
			PersistenceManagerFactory factory = new PersistenceManagerFactory("engines.config");
			this.pm = factory.CreatePersistenceManager();
			factory.AddAssembly("NUnitWeaver");
		}

		[Test]
		public void Overral()
		{

			pm.InitializeRepository();

            Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Person partner = new Person();
			partner.Name = "Joey Triviani";
			partner.Age = 32;

			p.Partners.Add(partner);
	
			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

            Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);

            p = pm.Load<Person>(typeof(Person), "[_Name = 'John Smith']")[0] as Person;
			Assert.IsNotNull(p);

			Assert.AreEqual("John Smith", p.Name);
			Assert.AreEqual(50, p.Age);

			Assert.AreEqual(1, p.Partners.Count);
			Assert.AreEqual("Joey Triviani", ((Person)p.Partners[0]).Name);

			p.Age = 49;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[_Name = 'John Smith']").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[m_Age = 49]").Count);
		}

		[Test]
		public void OneToOneSerialize()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Address a = new Address("New York");
			p.Address = a;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

            Assert.AreEqual(1, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Address)).Count);
            Assert.AreEqual("New York", ((Address)pm.Load<Address>(typeof(Address))[0]).City);
		}

		[Test]
		public void OneToOneLazyLoad()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Address a = new Address("New York");
			p.Address = a;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

            p = pm.LoadSingle<Person>(typeof(Person));
			Assert.IsFalse (p.BaseAddressGetterCalled);
			Assert.IsNotNull(p.Address);
			Assert.IsTrue(p.BaseAddressGetterCalled);
			Assert.AreEqual("New York", p.Address.City);
		}

		[Test]
		public void Update()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;
			p.Birth = new DateTime(2005, 12, 28);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

            p = pm.LoadSingle<Person>(typeof(Person));

			Assert.AreEqual("John Smith", p.Name);
			Assert.AreEqual(50, p.Age);
			Assert.AreEqual(new DateTime(2005, 12, 28), p.Birth);

			p.Name = "Doug Ross";
			p.Age = 40;
			p.Birth = new DateTime(2005, 12, 29);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

            Assert.AreEqual(1, pm.Load<Person>(typeof(Person)).Count);
            p = pm.LoadSingle<Person>(typeof(Person));
			Assert.AreEqual("Doug Ross", p.Name);
			Assert.AreEqual(40, p.Age);
			Assert.AreEqual(new DateTime(2005, 12, 29), p.Birth);
		}

		[Test]
		public void IdentityMap()
		{			
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

            Person p1 = pm.LoadSingle<Person>(typeof(Person));
            Person p2 = pm.LoadSingle<Person>(typeof(Person));
			
			Assert.IsNotNull(p1);
			Assert.IsNotNull(p2);
			Assert.IsTrue(p1 == p2);
		}

		[Test]
		public void CascadeUpdate()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Address a1 = new Address();
			a1.City = "a1";

			Person p2 = new Person();
			p2.Name = "p2";

			p1.Partners.Add(p2);
			p1.Address = a1;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

            p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");
			p2 = ((Person)p1.Partners[0]);
			a1 = p1.Address;
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual("p2", p2.Name);
			Assert.AreEqual("a1", a1.City);

			p2.Name = "p2 changed";
			a1.City = "a1 changed";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

            p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");
			p2 = ((Person)p1.Partners[0]);
			a1 = p1.Address;
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual("p2 changed", p2.Name);
			Assert.AreEqual("a1 changed", a1.City);

		}

		[Test]
		public void Inheritance()
		{
			pm.InitializeRepository();

			Employee bill = new Employee();
			bill.Name = "Bill Gates";
			bill.Salary = float.MaxValue;

			Address redmond = new Address();
			redmond.City = "Redmond";
			bill.Address = redmond;

			Address paloalto = new Address();
			paloalto.City = "Palo Alto";

			Person steve = new Person();
			steve.Name = "Steve Jobs";
			steve.Address = paloalto;

			bill.Partners.Add(steve);
            
			pm.BeginTransaction();
			pm.Serialize(bill);
			pm.CommitTransaction();

			pm.Clear();

            Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(2, pm.Load<Person>(typeof(Address)).Count);

            bill = pm.LoadSingle<Employee>(typeof(Employee), "[Name = 'Bill Gates']");
			Assert.IsNotNull(bill);
			Assert.AreEqual(float.MaxValue, bill.Salary);
			Assert.AreEqual("Steve Jobs", ((Person)bill.Partners[0]).Name);
			Assert.AreEqual("Redmond", bill.Address.City);

		}
	}
}
