using System;
using System.Reflection;
using System.Text;
using System.Xml;
using System.Diagnostics;
using System.Collections;

using System.Runtime.InteropServices;
using System.Diagnostics.SymbolStore;

using Mono.Cecil;

using Pdb.Model;

namespace Pdb.Tools
{
	/// <summary>
	/// Description rsume de PdbReader.
	/// </summary>
	/// 
	[Serializable]
	public class PdbReader
	{
		public PdbReader()
		{
		
		}

		public PdbReader(string assemblyName)
		{
			_assemblyName = assemblyName;
		}

		private string _assemblyName;
		public string AssemblyName
		{
			get
			{
				return _assemblyName;
			}
			set
			{
				_assemblyName = value;
			}
		}

		private string CreatePdbFromAssembly(string assemblyName)
		{
			string noExt = assemblyName.Substring(0, assemblyName.LastIndexOf('.'));
			
			return String.Concat(noExt, ".pdb");
		}

		public PdbInfos GetPdbModel()
		{
			string pdbFile = CreatePdbFromAssembly(_assemblyName);
			if(!System.IO.File.Exists(pdbFile))
				throw new System.IO.FileNotFoundException("Cannot find the debug information file", pdbFile);

			//Assembly assembly = Assembly.LoadFile(_assemblyName);
			AssemblyDefinition assembly = AssemblyFactory.GetAssembly(_assemblyName);

			PdbInfos pdb = new PdbInfos(_assemblyName);
			
			ISymbolReader reader = SymUtil.GetSymbolReaderForFile(_assemblyName, null);

			pdb.Files = this.GetFiles(reader);

			foreach (TypeDefinition currentType in assembly.MainModule.Types)
			{
				Hashtable methods = SymUtil.GetMethodsSymbolFromType(currentType, reader);

				ArrayList meths = new ArrayList ();

				foreach (MethodDefinition methodReflection in currentType.Methods)
				{
					meths.Add (methodReflection);
				}

				foreach (MethodDefinition ctorReflection in currentType.Constructors)
				{
					meths.Add (ctorReflection);
				}

				foreach (MethodDefinition methodReflection in meths)
				{
					ISymbolMethod methodSymbol = (ISymbolMethod)methods[methodReflection.ToString()];

                    if (methodSymbol != null)
                    {
                        Method methodModel = new Method();
                        methodModel.Namespace = currentType.Namespace;
                        methodModel.Name = methodReflection.Name;
                        methodModel.FullName = methodReflection.ToString();
                        methodModel.TypeName = methodReflection.DeclaringType.FullName;

                        methodModel.Token = methodSymbol.Token.GetToken();
                        methodModel.StartScope = methodSymbol.RootScope.StartOffset;
                        methodModel.EndScope = methodSymbol.RootScope.EndOffset;

                        methodModel.SequencePoints = this.GetSequencePoints(methodSymbol, pdb.Files);
                        LocalCollection locals = this.GetLocals(methodSymbol.RootScope);

                        if (locals != null)
                        {
                            foreach (Local local in locals)
                            {
                                methodModel.Locals.Add(local);
                            }
                        }

                        pdb.Methods.Add(methodModel);
                    }
				}
			}

			pdb.EntryPoint = this.GetEntryPoint(reader);

			return pdb;
		}

		private EntryPoint GetEntryPoint(ISymbolReader reader)
		{
			try
			{
				SymbolToken token = reader.UserEntryPoint;
				ISymbolMethod m = reader.GetMethod(token);

				EntryPoint entryPoint = new EntryPoint();
				Pdb.Model.MethodReference mRef = new Pdb.Model.MethodReference();
				mRef.Value = String.Format(System.Globalization.CultureInfo.InvariantCulture, "0x{0:x}", token.GetToken());

				entryPoint.MethodRef = mRef;

				return entryPoint;
			}
			catch (System.Runtime.InteropServices.COMException)
			{
				return null;
			}
		}

		private FileCollection GetFiles(ISymbolReader reader)
		{
			FileCollection files = new FileCollection();

			int id = 0;
            
			ISymbolDocument[] docs = reader.GetDocuments();

			foreach (ISymbolDocument doc in docs)
			{
				if (!files.ContainsFile(doc.URL))
				{
					File file = new File();
					file.Name = doc.URL;
					file.Id = id;

					file.DocumentType = doc.DocumentType;
					file.Language = doc.Language;
					file.LanguageVendor = doc.LanguageVendor;

					files.Add(file);

					id++;
				}
			}

			return files;
		}

		private LocalCollection GetLocals(ISymbolScope scope)
		{
			LocalCollection locals = new LocalCollection();

			foreach (SymVariable l in scope.GetLocals())
			{
				Local local = new Local();

				local.Name = l.Name;

				local.Attributes = (System.Reflection.FieldAttributes)l.Attributes;

				local.AddressKind = l.AddressKind;

				local.Index = l.AddressField1;
				local.Start = scope.StartOffset;
				local.End = scope.EndOffset;

				locals.Add(local);
			}

			foreach (ISymbolScope childScope in scope.GetChildren())
			{
				LocalCollection locals2 = GetLocals(childScope);

				if (locals2 != null)
				{
					foreach (Local lmodel in locals2)
						locals.Add(lmodel);
				}
			}
			return locals;
		}

		private EntryCollection GetSequencePoints(ISymbolMethod method, FileCollection files)
		{
			EntryCollection entries = new EntryCollection();

			int[] offsets = new int[method.SequencePointCount];
			int[] startColumn = new int[method.SequencePointCount];
			int[] endColumn = new int[method.SequencePointCount];
			int[] startRow = new int[method.SequencePointCount];
			int[] endRow = new int[method.SequencePointCount];

			ISymbolDocument[] docs = new ISymbolDocument[method.SequencePointCount];
            
			method.GetSequencePoints(offsets, docs, startRow, startColumn, endRow, endColumn);

			for (int i = 0; i < method.SequencePointCount; i++)
			{
				Entry entry = new Entry();

				entry.FileId = files.GetFileByName(docs[0].URL).Id;

				entry.Offset = offsets[i];
                
				if (startRow[i] == 0xFeeFee)
				{
					entry.Hidden = true;
				}
				else
				{
					entry.StartRow = startRow[i];
					entry.EndRow = endRow[i];
					entry.StartColumn = startColumn[i];
					entry.EndColumn = endColumn[i];
				}

				entries.Add(entry);
			}
			return entries;
		}
	}
}
