using System;
using System.Collections;

namespace Evaluant.OPath.Expressions.Collections
{
	/// <summary>
	///     <para>
	///       A collection that stores <see cref="Identifier"/> objects.
	///    </para>
	/// </summary>
	/// <seealso cref="IdentifierCollection"/>
	[Serializable()]
    public class IdentifierCollection : CollectionBase 
	{
		
		/// <summary>
		///     <para>
		///       Initializes a new instance of <see cref="IdentifierCollection"/>.
		///    </para>
		/// </summary>
		public IdentifierCollection()
		{
		}
		
		/// <summary>
		///     <para>
		///       Initializes a new instance of <see cref="IdentifierCollection"/> based on another <see cref="IdentifierCollection"/>.
		///    </para>
		/// </summary>
		/// <param name="value">
		///       A <see cref="IdentifierCollection"/> from which the contents are copied
		/// </param>
		public IdentifierCollection(IdentifierCollection value)
		{
			this.AddRange(value);
		}
		
		/// <summary>
		///     <para>
		///       Initializes a new instance of <see cref="IdentifierCollection"/> containing any array of <see cref="Identifier"/> objects.
		///    </para>
		/// </summary>
		/// <param name="value">
		///       A array of <see cref="Identifier"/> objects with which to intialize the collection
		/// </param>
		public IdentifierCollection(Identifier[] value)
		{
			this.AddRange(value);
		}
		
		/// <summary>
		/// <para>Represents the entry at the specified index of the <see cref="Identifier"/>.</para>
		/// </summary>
		/// <param name="index"><para>The zero-based index of the entry to locate in the collection.</para></param>
		/// <value>
		///    <para> The entry at the specified index of the collection.</para>
		/// </value>
		/// <exception cref="System.ArgumentOutOfRangeException"><paramref name="index"/> is outside the valid range of indexes for the collection.</exception>
		public Identifier this[int index] 
		{
			get 
			{
				return ((Identifier)(List[index]));
			}
			set 
			{
				List[index] = value;
			}
		}
		
		/// <summary>
		///    <para>Adds a <see cref="Identifier"/> with the specified value to the 
		///    <see cref="IdentifierCollection"/> .</para>
		/// </summary>
		/// <param name="value">The <see cref="Identifier"/> to add.</param>
		/// <returns>
		///    <para>The index at which the new element was inserted.</para>
		/// </returns>
		/// <seealso cref="IdentifierCollection.AddRange"/>
		public int Add(Identifier value)
		{
			return List.Add(value);
		}
		
		/// <summary>
		/// <para>Copies the elements of an array to the end of the <see cref="IdentifierCollection"/>.</para>
		/// </summary>
		/// <param name="value">
		///    An array of type <see cref="Identifier"/> containing the objects to add to the collection.
		/// </param>
		/// <returns>
		///   <para>None.</para>
		/// </returns>
		/// <seealso cref="IdentifierCollection.Add"/>
		public void AddRange(Identifier[] value)
		{
			for (int i = 0; i < value.Length; i++) 
			{
				this.Add(value[i]);
			}
		}
		
		/// <summary>
		///     <para>
		///       Adds the contents of another <see cref="IdentifierCollection"/> to the end of the collection.
		///    </para>
		/// </summary>
		/// <param name="value">
		///    A <see cref="IdentifierCollection"/> containing the objects to add to the collection.
		/// </param>
		/// <returns>
		///   <para>None.</para>
		/// </returns>
		/// <seealso cref="IdentifierCollection.Add"/>
		public void AddRange(IdentifierCollection value)
		{
			for (int i = 0; i < value.Count; i++)
			{
				this.Add(value[i]);
			}
		}
		
		/// <summary>
		/// <para>Gets a value indicating whether the 
		///    <see cref="IdentifierCollection"/> contains the specified <see cref="Identifier"/>.</para>
		/// </summary>
		/// <param name="value">The <see cref="Identifier"/> to locate.</param>
		/// <returns>
		/// <para><see langword='true"/> if the <see cref="Identifier"/> is contained in the collection; 
		///   otherwise, <see langword='false"/>.</para>
		/// </returns>
		/// <seealso cref="IdentifierCollection.IndexOf"/>
		public bool Contains(Identifier value)
		{
			return List.Contains(value);
		}
		
		/// <summary>
		/// <para>Copies the <see cref="IdentifierCollection"/> values to a one-dimensional <see cref="System.Array"/> instance at the 
		///    specified index.</para>
		/// </summary>
		/// <param name="array"><para>The one-dimensional <see cref="System.Array"/> that is the destination of the values copied from <see cref="IdentifierCollection"/> .</para></param>
		/// <param name="index">The index in <paramref name="array"/> where copying begins.</param>
		/// <returns>
		///   <para>None.</para>
		/// </returns>
		/// <exception cref="System.ArgumentException"><para><paramref name="array"/> is multidimensional.</para> <para>-or-</para> <para>The number of elements in the <see cref="IdentifierCollection"/> is greater than the available space between <paramref name="arrayIndex"/> and the end of <paramref name="array"/>.</para></exception>
		/// <exception cref="System.ArgumentNullException"><paramref name="array"/> is <see langword='null"/>. </exception>
		/// <exception cref="System.ArgumentOutOfRangeException"><paramref name="index"/> is less than <paramref name="array"/>'s lowbound. </exception>
		/// <seealso cref="System.Array"/>
		public void CopyTo(Identifier[] array, int index)
		{
			List.CopyTo(array, index);
		}
		
		/// <summary>
		///    <para>Returns the index of a <see cref="Identifier"/> in 
		///       the <see cref="IdentifierCollection"/> .</para>
		/// </summary>
		/// <param name="value">The <see cref="Identifier"/> to locate.</param>
		/// <returns>
		/// <para>The index of the <see cref="Identifier"/> of <paramref name="value"/> in the 
		/// <see cref="IdentifierCollection"/>, if found; otherwise, -1.</para>
		/// </returns>
		/// <seealso cref="IdentifierCollection.Contains"/>
		public int IndexOf(Identifier value)
		{
			return List.IndexOf(value);
		}
		
		/// <summary>
		/// <para>Inserts a <see cref="Identifier"/> into the <see cref="IdentifierCollection"/> at the specified index.</para>
		/// </summary>
		/// <param name="index">The zero-based index where <paramref name="value"/> should be inserted.</param>
		/// <param name=" value">The <see cref="Identifier"/> to insert.</param>
		/// <returns><para>None.</para></returns>
		/// <seealso cref="IdentifierCollection.Add"/>
		public void Insert(int index, Identifier value)
		{
			List.Insert(index, value);
		}
		
		/// <summary>
		///    <para>Returns an enumerator that can iterate through 
		///       the <see cref="IdentifierCollection"/> .</para>
		/// </summary>
		/// <returns><para>None.</para></returns>
		/// <seealso cref="System.Collections.IEnumerator"/>
		internal new IdentifierEnumerator GetEnumerator()
		{
			return new IdentifierEnumerator(this);
		}
		
		/// <summary>
		///    <para> Removes a specific <see cref="Identifier"/> from the 
		///    <see cref="IdentifierCollection"/> .</para>
		/// </summary>
		/// <param name="value">The <see cref="Identifier"/> to remove from the <see cref="IdentifierCollection"/> .</param>
		/// <returns><para>None.</para></returns>
		/// <exception cref="System.ArgumentException"><paramref name="value"/> is not found in the Collection. </exception>
		public void Remove(Identifier value)
		{
			List.Remove(value);
		}
		
		internal class IdentifierEnumerator : IEnumerator
		{
			IEnumerator baseEnumerator;
			IEnumerable temp;
			
			public IdentifierEnumerator(IdentifierCollection mappings)
			{
				this.temp = ((IEnumerable)(mappings));
				this.baseEnumerator = temp.GetEnumerator();
			}
			
			public Identifier Current 
			{
				get 
				{
					return ((Identifier)(baseEnumerator.Current));
				}
			}
			
			object IEnumerator.Current 
			{
				get 
				{
					return baseEnumerator.Current;
				}
			}
			
			public bool MoveNext()
			{
				return baseEnumerator.MoveNext();
			}
			
			bool IEnumerator.MoveNext()
			{
				return baseEnumerator.MoveNext();
			}
			
			public void Reset()
			{
				baseEnumerator.Reset();
			}
			
			void IEnumerator.Reset()
			{
				baseEnumerator.Reset();
			}
		}
	}
}