grammar OPath;

options
{
	output=AST;
	ASTLabelType=CommonTree;
	language=CSharp;
}

@header {
using System.Text;
using Evaluant.OPath.Expressions;
using Evaluant.OPath.Expressions.Collections;
}

@members {
const char BS = '\\';

private string extractString(string text) {
    
    StringBuilder sb = new StringBuilder(text);
    int startIndex = 1; // Skip initial quote
    int slashIndex = -1;

    while ((slashIndex = sb.ToString().IndexOf(BS, startIndex)) != -1)
    {
        char escapeType = sb[slashIndex + 1];
        switch (escapeType)
        {
            case 'u':
              string hcode = String.Concat(sb[slashIndex+4], sb[slashIndex+5]);
              string lcode = String.Concat(sb[slashIndex+2], sb[slashIndex+3]);
              char unicodeChar = Encoding.Unicode.GetChars(new byte[] { System.Convert.ToByte(hcode, 16), System.Convert.ToByte(lcode, 16)} )[0];
              sb.Remove(slashIndex, 6).Insert(slashIndex, unicodeChar); 
              break;
            case 'n': sb.Remove(slashIndex, 2).Insert(slashIndex, '\n'); break;
            case 'r': sb.Remove(slashIndex, 2).Insert(slashIndex, '\r'); break;
            case 't': sb.Remove(slashIndex, 2).Insert(slashIndex, '\t'); break;
            case '\'': sb.Remove(slashIndex, 2).Insert(slashIndex, '\''); break;
            case '\\': sb.Remove(slashIndex, 2).Insert(slashIndex, '\\'); break;
            default: throw new RecognitionException("Unvalid escape sequence: \\" + escapeType);
        }

        startIndex = slashIndex + 1;

    }

    sb.Remove(0, 1);
    sb.Remove(sb.Length - 1, 1);

    return sb.ToString();
}
}

@init {
    numberFormatInfo.NumberDecimalSeparator = ".";
}

pathExpression returns [Path value]
	: path EOF { $value = $path.value; }
	;

path returns [Path value]
@init {
IdentifierCollection identifiers = new IdentifierCollection();
}
	: first=identifier { identifiers.Add($first.value); } ( '.' next=identifier { identifiers.Add($next.value); })*  { $value = new Path(identifiers); }
	;

identifier returns [Identifier value]
@init {
Constraint constraint = null;
}
	: ID ('[' conditionalExpression ']' { constraint = $conditionalExpression.value; })? { $value = new Identifier($ID.text, constraint); }
	;
		
conditionalExpression returns [Constraint value]
@init {
BinaryOperatorEnum type = BinaryOperatorEnum.Unknown;
}
	:	left=booleanAndExpression { $value = $left.value; } (
			'or' { type = BinaryOperatorEnum.Or; } 
			right=booleanAndExpression { $value = new BinaryOperator(type, $value, $right.value); } 
			)* 
	;
		
booleanAndExpression returns [Constraint value]
@init {
BinaryOperatorEnum type = BinaryOperatorEnum.Unknown;
}
	:	left=equalityExpression { $value = $left.value; } (
			'and' { type = BinaryOperatorEnum.And; } 
			right=equalityExpression { $value = new BinaryOperator(type, $value, $right.value); } 
			)* 
	;

equalityExpression returns [Constraint value]
@init {
BinaryOperatorEnum type = BinaryOperatorEnum.Unknown;
}
	:	left=relationalExpression { $value = $left.value; } (
			( '=' { type = BinaryOperatorEnum.Equal; } 
			| '!=' { type = BinaryOperatorEnum.NotEqual; } 
			| '~' { type = BinaryOperatorEnum.Contains; } 
			| '^' { type = BinaryOperatorEnum.BeginsWith; } 
			| '$' { type = BinaryOperatorEnum.EndsWith; } ) 
			right=relationalExpression { $value = new BinaryOperator(type, $value, $right.value); } 
			)* 
	;
	
relationalExpression returns [Constraint value]
@init {
BinaryOperatorEnum type = BinaryOperatorEnum.Unknown;
}
	:	left=additiveExpression { $value = $left.value; } (
			( '<' { type = BinaryOperatorEnum.Lesser; } 
			| '<=' { type = BinaryOperatorEnum.LesserOrEqual; }  
			| '>' { type = BinaryOperatorEnum.Greater; } 
			| '>=' { type = BinaryOperatorEnum.GreaterOrEqual; } ) 
			right=additiveExpression { $value = new BinaryOperator(type, $value, $right.value); } 
			)* 
	;

additiveExpression returns [Constraint value]
@init {
BinaryOperatorEnum type = BinaryOperatorEnum.Unknown;
}
	:	left=multiplicativeExpression { $value = $left.value; } (
			( '+' { type = BinaryOperatorEnum.Plus; } 
			| '-' { type = BinaryOperatorEnum.Minus; } ) 
			right=multiplicativeExpression { $value = new BinaryOperator(type, $value, $right.value); } 
			)* 
	;

multiplicativeExpression returns [Constraint value]
@init {
BinaryOperatorEnum type = BinaryOperatorEnum.Unknown;
}
	:	left=unaryExpression { $value = $left.value; } (
			( '*' { type = BinaryOperatorEnum.Times; } 
			| '/' { type = BinaryOperatorEnum.Div; } 
			| '%' { type = BinaryOperatorEnum.Modulo; } ) 
			right=unaryExpression { $value = new BinaryOperator(type, $left.value, $right.value); } 
			)* 
	;

unaryExpression returns [Constraint value]
	:	primaryExpression { $value = $primaryExpression.value; }
    	|	'not' primaryExpression { $value = new UnaryOperator(UnaryOperatorEnum.Not, $primaryExpression.value); }
    	|	'-' primaryExpression { $value = new UnaryOperator(UnaryOperatorEnum.Minus, $primaryExpression.value); }
   	;
		
primaryExpression returns [Constraint value]
	:	
	|	'(' conditionalExpression ')' 	{ $value = $conditionalExpression.value; }
	|	expr=value			{ $value = $expr.value; }
	|	function			{ $value = $function.value; }
	|	path				{ $value = $path.value; }
	;

function returns [Expression value]
@init{
FunctionEnum type = FunctionEnum.Unknown;
ConstraintCollection constraints = new ConstraintCollection();
}
	:	
	( 'max' { type = FunctionEnum.Max; }
	| 'min' { type = FunctionEnum.Min; }
	| 'count' { type = FunctionEnum.Count; }
	| 'avg' { type = FunctionEnum.Average; }
	| 'sum' { type = FunctionEnum.Sum; }
	| 'exists' { type = FunctionEnum.Exists; }
	| 'isNull' { type = FunctionEnum.IsNull; } 
	) '(' path ')' { $value = new Function($path.value, type); }
	| ID '(' first=conditionalExpression { constraints.Add($first.value); } (',' next=conditionalExpression { constraints.Add($next.value); } )* ')'	
		{ $value = new Call($ID.text, constraints);  }
	;

value returns [Value value]
	: 	INTEGER		{ $value = new Value($INTEGER.text, ValueEnum.Integer); }
	|	FLOAT		{ $value = new Value($FLOAT.text, ValueEnum.Float); }
	| 	DATE		{ $value = new Value($DATE.text, ValueEnum.Date); }
	|	BOOLEAN		{ $value = new Value($BOOLEAN.text, ValueEnum.Boolean); }
	|	STRING		{ $value = new Value(extractString($STRING.text), ValueEnum.String); }
	;

STRING
    	:  	'\'' ( EscapeSequence | (options {greedy=false;} : ~('\u0000'..'\u001f' | '\\' | '\'' ) ) )* '\''
    	;

BOOLEAN 
	:	'true'	
	|	'false'	
	;

ID 
	: 	LETTER ( LETTER | DIGIT | '{' | '}' | ':' )*
	;

INTEGER
	:	DIGIT+
	;

FLOAT 
	:	DIGIT* '.' DIGIT+
	;

DATE
	:	'#' (~('\r' | '\n' | '#'))* '#'
	;
	
fragment LETTER
	:	'a'..'z'
	|	'A'..'Z'
	|	'_'
	;

fragment DIGIT
	:	'0'..'9'
	;

fragment EscapeSequence 
	:	'\\'
  	(	
  		'n' 
	|	'r' 
	|	't'
	|	'\'' 
	|	'\\'
	|	UnicodeEscape
	)
  ;

fragment HexDigit 
	: 	('0'..'9'|'a'..'f'|'A'..'F') ;


fragment UnicodeEscape
    	:    	'u' HexDigit HexDigit HexDigit HexDigit 
    	;

/* Ignore white spaces */	
WS	:  (' '|'\r'|'\t'|'\u000C'|'\n') {$channel=HIDDEN;}
	;
