using System.Collections;
using Evaluant.Uss.Commands;
using Evaluant.Uss.Common;
using System.Threading;

namespace Evaluant.Uss.Memory
{
    /// <summary>
    /// Visitor pattern to execute concrete commands
    /// </summary>
    public class MemoryCommandProcessor : ICommandProcessor
    {
        private Hashtable _Entities;
        private ReaderWriterLock _RWL;

        private string _NewId;
        public string NewId
        {
            get { return _NewId; }
            set { _NewId = value; }
        }

        public MemoryCommandProcessor(Hashtable entities, ReaderWriterLock rwl)
        {
            _Entities = entities;
            _RWL = rwl;
        }

        public void Process(Command c)
        {
        }

        public void Process(CreateEntityCommand c)
        {
            Entity entity = new Entity(c.Type);
            entity.Id = c.ParentId;
            entity.State = State.UpToDate;
            entity.MetaData = Entity.LoadMetaData.AttributesLoaded | Entity.LoadMetaData.AttributesLoaded;

            _RWL.AcquireWriterLock(Timeout.Infinite);

            try
            {
                _Entities.Add(Utils.GetCacheKey(c.Type, c.ParentId), entity);
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }

        }

        public void Process(CompoundCreateCommand c)
        {
            Process((CreateEntityCommand)c);
            foreach (CreateAttributeCommand cc in c.InnerCommands)
                cc.Accept(this);
        }

        public void Process(CompoundUpdateCommand c)
        {
            foreach (Command cc in c.InnerCommands)
                cc.Accept(this);
        }

        public void Process(DeleteEntityCommand c)
        {
            _RWL.AcquireWriterLock(Timeout.Infinite);

            try
            {
                Entity processedEntity = GetEntity(c.Type, c.ParentId);
                if (processedEntity == null)
                    return;

                _Entities.Remove(Utils.GetCacheKey(c.Type, c.ParentId));

                foreach (Entity entity in _Entities.Values)
                    entity.RemoveReference(processedEntity);
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }

        }

        public void Process(CreateAttributeCommand c)
        {
            _RWL.AcquireWriterLock(Timeout.Infinite);

            try
            {
                Entity entity = GetEntity(c.ParentType, c.ParentId);
                State tmpState = entity.State;
                entity.AddValue(c.Name, c.Value, c.Type, State.UpToDate);
                entity.State = tmpState;
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }

        }

        public void Process(UpdateAttributeCommand c)
        {
            _RWL.AcquireWriterLock(Timeout.Infinite);

            try
            {
                Entity entity = GetEntity(c.ParentType, c.ParentId);
                State tmpState = entity.State;
                entity[c.Name] = c.Value;
                entity.State = tmpState;
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }
        }

        public void Process(DeleteAttributeCommand c)
        {
            _RWL.AcquireWriterLock(Timeout.Infinite);

            try
            {
                Entity entity = GetEntity(c.ParentType, c.ParentId);
                State tmpState = entity.State;
                entity.RemoveAttribute(c.Name);
                entity.State = tmpState;
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }
        }

        public void Process(CreateReferenceCommand c)
        {
            _RWL.AcquireWriterLock(Timeout.Infinite);

            try
            {
                Entity childEntity = GetEntity(c.ChildType, c.ChildId);
                Entity parentEntity = GetEntity(c.ParentType, c.ParentId);

                State tmpState = parentEntity.State;
                parentEntity.AddValue(c.Role, childEntity);
                parentEntity.State = tmpState;
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }
        }

        public void Process(DeleteReferenceCommand c)
        {
            _RWL.AcquireWriterLock(Timeout.Infinite);

            try
            {
                Entity entity = GetEntity(c.ParentType, c.ParentId);

                for (int i = entity.EntityEntries.Count - 1; i >= 0; i--)
                {
                    EntityEntry ee = (EntityEntry)entity.EntityEntries[i];
                    if (ee.IsEntity && ((Entity)ee.Value).Id == c.ChildId)
                        entity.EntityEntries.Remove(ee);
                }
            }
            finally
            {
                _RWL.ReleaseWriterLock();
            }
        }

        protected Entity GetEntity(string type, string id)
        {
            return (Entity)_Entities[Utils.GetCacheKey(type, id)];
        }
    }
}
