using System;
using System.Collections;
using System.Xml.Serialization;

namespace Evaluant.Uss.SqlMapper
{
    /// <summary>
    /// Description rsume de Mapping.
    /// </summary>
    /// 
    [XmlRoot]
    public class Mapping
    {
        IDriver _Driver;
        EntityMappingCollection _Entities;

        /// <summary>
        /// Initializes a new instance of the <see cref="Mapping"/> class.
        /// </summary>
        public Mapping()
        {
            _Entities = new EntityMappingCollection();
        }

        private string Prefix(string tableName, string tablePrefix)
        {
            if (tableName != null && tableName.Trim() != String.Empty)
            {
                return string.Concat(tablePrefix, tableName);
            }
            else
            {
                return tableName;
            }
        }

        /// <summary>
        /// Initializes the entities of this instance after deserializing
        /// </summary>
        public void Initialize(string tablePrefix)
        {
            foreach (EntityMapping e in _Entities)
                e.ParentMapping = this;

            foreach (EntityMapping e in _Entities)
                CopyNodeTo(e);

            //	Add prefix to each table if defined
            if (tablePrefix != null && tablePrefix != string.Empty)
            {
                foreach (EntityMapping em in _Entities)
                {
                    em.Table = Prefix(em.Table, tablePrefix);

                    foreach (AttributeMapping am in em.Attributes)
                    {
                        if (am.ParentEntity == null)
                            am.ParentEntity = em;

                        if (am.Table != null && am.Table != am.ParentEntity.Table)
                            am.Table = Prefix(am.Table, tablePrefix);
                    }

                    foreach (ReferenceMapping rm in em.References)
                    {
                        if (rm.EntityParent == null)
                            rm.EntityParent = em;

                        else if (rm.EntityParent.Table != null && rm.EntityParent.Table != string.Empty)
                            rm.EntityParent.Table = Prefix(rm.EntityParent.Table, tablePrefix);

                        foreach (RuleMapping rule in rm.Rules)
                        {
                            rule.ParentTable = Prefix(rule.ParentTable, tablePrefix);
                            rule.ChildTable = Prefix(rule.ChildTable, tablePrefix);
                        }
                    }
                }
            }

            foreach (EntityMapping e in _Entities)
                e.Initialize();


            ProcessRules();
        }

        /// <summary>
        /// Gets or sets the factory.
        /// </summary>
        /// <value></value>
        [XmlIgnore()]
        public IDriver Driver
        {
            get { return _Driver; }
            set { _Driver = value; }
        }

        /// <summary>
        /// Gets or sets the entities.
        /// </summary>
        /// <value></value>
        [XmlElementAttribute("Entity")]
        public EntityMappingCollection Entities
        {
            get { return _Entities; }
            set { _Entities = value; }
        }

        private void ProcessRules()
        {
            foreach (EntityMapping entity in _Entities)
            {
                foreach (ReferenceMapping map in entity.References)
                {
                    if (!string.IsNullOrEmpty(entity.DiscriminatorField) && entity.Ids[0].Generator.Name == GeneratorMapping.GeneratorType.business)
                        map.Rules[0].IsParentId = map.Rules[0].ParentField + SqlMapperProvider.IDSEP + entity.DiscriminatorField == map.EntityParent.IdFields;
                    else
                        map.Rules[0].IsParentId = map.Rules[0].ParentField == map.EntityParent.IdFields;

                    EntityMapping childEntity = _Entities[map.EntityChild];
                    //if (map.Rules[0].IsParentId)
                    //{
                    //    map.Rules[0].ParentField += SqlMapperProvider.IDSEP + entity.DiscriminatorField;
                    //}

                    for (int index = 1; index < map.Rules.Count; index++)
                    {
                        //if (!string.IsNullOrEmpty(entity.DiscriminatorField))
                        //    map.Rules[index].IsParentId = map.Rules[index].ParentField + SqlMapperProvider.IDSEP + entity.DiscriminatorField == map.Rules[index - 1].ChildField;
                        //else
                        map.Rules[index].IsParentId = map.Rules[index].ParentField == map.Rules[index - 1].ChildField;

                        //if (map.Rules[0].IsParentId)
                        //{
                        //    map.Rules[0].ParentField += SqlMapperProvider.IDSEP + entity.DiscriminatorField;
                        //}
                    }
                }
            }
        }

        public void CopyNodeTo(EntityMapping map)
        {
            if (map.CopyNode != String.Empty)
            {
                EntityMapping copyEntity = _Entities[map.CopyNode];
                CopyNodeTo(copyEntity);

                foreach (AttributeMapping atrb in copyEntity.Attributes)
                {
                    if (map.Attributes[atrb.Name] == null)
                    {
                        AttributeMapping copyAtrb = new AttributeMapping(atrb.Name, atrb.Type, map);

                        copyAtrb.Table = atrb.ParentField != null ? atrb.Table : map.Table; // if ParentField != null : attribute table is not a same table of the entity table
                        copyAtrb.Field = atrb.Field;
                        copyAtrb.ParentField = atrb.ParentField;
                        copyAtrb.DbType = atrb.DbType;
                        copyAtrb.DefaultValue = atrb.DefaultValue;
                        copyAtrb.Discriminator = atrb.Discriminator;
                        copyAtrb.DiscriminatorValue = atrb.DiscriminatorValue;
                        copyAtrb.IsNotNull = atrb.IsNotNull;
                        copyAtrb.Precision = atrb.Precision;
                        copyAtrb.Scale = atrb.Scale;
                        copyAtrb.Size = atrb.Size;
                        copyAtrb.XmlType = atrb.XmlType;

                        map.Attributes.Add(copyAtrb);
                    }
                }

                foreach (ReferenceMapping refm in copyEntity.References)
                {
                    if (map.References[refm.Name, refm.EntityChild] == null)
                    {
                        ReferenceMapping copyRefm = new ReferenceMapping(refm.Name, refm.EntityChild, refm.EntityParent);
                        copyRefm.DiscriminatorField = refm.DiscriminatorField;
                        copyRefm.DiscriminatorValue = refm.DiscriminatorValue;
                        foreach (RuleMapping rulem in refm.Rules)
                        {
                            RuleMapping copyRule = new RuleMapping(rulem.ParentField, rulem.ChildField, copyRefm);
                            copyRule.ChildTable = rulem.ChildTable;
                            copyRule.ParentTable = rulem.ParentTable;
                            copyRule.ParentDefaultValue = rulem.ParentDefaultValue;
                            copyRule.ChildDefaultValue = rulem.ChildDefaultValue;
                            copyRefm.Rules.Add(copyRule);
                        }

                        map.References.Add(copyRefm);
                    }
                }
            }
        }

        /// <summary>
        /// Truncates any information to a specified number of chars
        /// </summary>
        /// <param name="size">The maximum length of an identifier for the dialect</param>
        public void Normalize(int size)
        {
            int length = size - 2;

            Hashtable truncated = new Hashtable();
            ArrayList names = new ArrayList();

            foreach (EntityMapping em in this.Entities)
            {
                em.Table = GetTruncatedTerm(truncated, names, "table", em.Table, length);

                foreach (PrimaryKeyMapping pkm in em.Ids)
                {
                    pkm.Field = GetTruncatedTerm(truncated, names, "field", pkm.Field, length);
                }

                foreach (AttributeMapping am in em.Attributes)
                {
                    am.Field = GetTruncatedTerm(truncated, names, "field", am.Field, length);
                }

                foreach (ReferenceMapping rem in em.References)
                {
                    foreach (RuleMapping rm in rem.Rules)
                    {
                        for (int i = 0; i < rm.ChildFields.Length; i++)
                            rm.ChildFields[i] = GetTruncatedTerm(truncated, names, "field", rm.ChildFields[i], length);
                        rm.ChildTable = rm.ChildSchema + GetTruncatedTerm(truncated, names, "table", rm.ChildTable, length);
                        for (int i = 0; i < rm.ParentFields.Length; i++)
                            rm.ParentFields[i] = GetTruncatedTerm(truncated, names, "field", rm.ParentFields[i], length);
                        rm.ParentTable = rm.ParentSchema + GetTruncatedTerm(truncated, names, "table", rm.ParentTable, length);
                        rm.Constraint = GetTruncatedTerm(truncated, names, "cs", rm.Constraint, length);
                    }
                }
            }
        }

        public void Normalize(DBDialect dialect)
        {
            Normalize(dialect.GetMaxIdentifierLength());
        }

        private string GetTruncatedTerm(Hashtable truncated, ArrayList names, string type, string term, int length)
        {
            if (term == null || term == String.Empty || term.Length <= length)
                return term;

            if (!truncated.Contains(String.Concat(type, ":", term)))
            {
                string newterm = term.Substring(0, length);
                int i = 1;
                while (names.Contains(String.Concat(newterm, i.ToString()))) i++;
                truncated.Add(String.Concat(type, ":", term), String.Concat(newterm, i.ToString()));
                names.Add(String.Concat(newterm, i.ToString()));
            }

            return (string)truncated[String.Concat(type, ":", term)];
        }
    }
}
