using System;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Xsl;

namespace Evaluant.Uss.Xmi
{
	/// <summary>
	/// Utility class that manipulates
	/// xmi files
	/// </summary>
	public class Converter
	{
		static Stream GetResource (string res)
		{
            //string fullName = string.Concat ("Evaluant.Uss.", typeof (Converter).Namespace, ".Xsl.", res);
            //return Assembly.GetExecutingAssembly().GetManifestResourceStream(fullName);

            string fullName = string.Concat(typeof(Converter).Namespace, ".Xsl.", res);
			return Assembly.GetAssembly(typeof(Converter)).GetManifestResourceStream (fullName);
		}

		static string GetTempFile (string file)
		{
			return Path.Combine (Path.GetDirectoryName (
				Assembly.GetAssembly (typeof (Converter)).Location), file);
		}

		static void SaveTempResource (string res)
		{
			using (StreamWriter sw = new StreamWriter (GetTempFile (res))) {
				using (StreamReader sr = new StreamReader (GetResource (res))) {
					sw.Write (sr.ReadToEnd ());
				}
			}
		}

		static string CleanXml (string xml)
		{
			return DeleteSchema (DeleteDTD (xml));
		}

		static void LoadDependantXsl (string spec, XmlDocument xsl)
		{
			SaveTempResource (spec);
			SaveTempResource ("toXMI.xsl");
			SaveTempResource ("toXMIFunction.xsl");

			xsl.Load (GetTempFile (spec));
		}

		/// <summary>
		/// Extracts the exporter from an xml string
		/// </summary>
		/// <param name="fileContent"></param>
		/// <returns>the exporter</returns>
		public static Exporter GetExporter(string fileContent)
		{
			XmlDocument xmlDoc = new XmlDocument();
			xmlDoc.LoadXml (CleanXml (fileContent));
			string exporter = string.Empty;
			try {
				exporter = xmlDoc.SelectSingleNode("XMI/XMI.header/XMI.documentation/XMI.exporter").InnerXml.Trim();
				switch(exporter) {
                    case "StarUML.XMI-Addin":
                        return Exporter.StarUML;

                    case "PowerAMC":
                        return Exporter.PowerAMC;

                    case "PowerDesigner":
						return Exporter.PowerDesigner;
			
					case "Objecteering/UML XMI Module":
						return Exporter.ObjectTeering;

					case "TogetherSoft":
						return Exporter.Together;

					case "Visual Paradigm for UML":
						return Exporter.VisualParadigm;

					case "Netbeans XMI Writer":
						return Exporter.Poseidon;

					case "Enterprise Architect":
						return Exporter.EA;

					default:
						return Exporter.Rose;
				}
			} catch {
				return Exporter.Rose;
			}
		}

		/// <summary>
		/// Extracts the schema type from an xml string
		/// </summary>
		/// <param name="fileContent">a whole xml file</param>
		/// <returns>the schema type</returns>
		public static SchemaType GetSchemaType(string fileContent)
		{
			XmlDocument xmlDoc = new XmlDocument();
			xmlDoc.LoadXml(CleanXml (fileContent));

			if (GetExporter (fileContent) == Exporter.Other) {
				//create a collection of XML Schema
				XmlSchemaCollection xsdColl = new XmlSchemaCollection();

				//the validating reader
				XmlValidatingReader vr = new XmlValidatingReader(new XmlNodeReader(xmlDoc.SelectSingleNode("//")));

				//we'll use XSD
				vr.ValidationType = ValidationType.Schema;

				switch(xmlDoc.FirstChild.Name)
				{
					case "Package" :
						//add the XSD file to the collection
						xsdColl.Add("", new XmlTextReader("SimpleXML.xsd"));

						//the XSD files are here
						vr.Schemas.Add(xsdColl);

						//read the file through
						while(vr.Read()) { }
						return SchemaType.Model;

			
					case "ORMapping" :
						//add the XSD file to the collection
						xsdColl.Add("", new XmlTextReader("MappingXML.xsd"));

						//the XSD files are here
						vr.Schemas.Add(xsdColl);

						//read the file through
						while(vr.Read()) { }
						return SchemaType.Mapping;
					default:
						break;
				}
			}
			
			return SchemaType.Other;
		}

		/// <summary>
		/// Extracts the version from a xml file
		/// </summary>
		/// <param name="fileContent">the whole xml file</param>
		/// <returns>the version</returns>
		public static XmiVersion GetVersion (string fileContent)
		{
			XmlDocument xmlDoc = new XmlDocument();
			xmlDoc.LoadXml(CleanXml (fileContent));

			string version = string.Empty;

			try {
				if (GetExporter(fileContent) == Exporter.ObjectTeering) {
					Regex regex = new Regex(@"<!DOCTYPE XMI SYSTEM '(?<version>[^']*)'",RegexOptions.Multiline);
					Match m = regex.Match(fileContent);
					if (m.Success)
						version = m.Groups["version"].Value;
				}
				else
					version = xmlDoc.SelectSingleNode("XMI/XMI.header/XMI.metamodel").Attributes["xmi.version"].Value;

				switch(version) {
					case "1.1":
						return XmiVersion.V11;

					case "1.3":
						return XmiVersion.V13;
					
					case "uml13.dtd":
						return XmiVersion.V13;

					case "1.4":
						return XmiVersion.V14;

					case "uml14.dtd":
						return XmiVersion.V14;

					default : 
						return XmiVersion.Other;
				}
			} catch {
				// not xmi
				try {
					XmlElement rootElement = xmlDoc.DocumentElement;
					switch(rootElement.LocalName) {
						case "UMLEditor":
							return XmiVersion.XmlMeta;

						case "Model":
							return XmiVersion.SimpleXmi;

						default :
							return XmiVersion.Other;

					}
				} catch {
					return XmiVersion.Other;
				}
			}
		}

		public static string ConvertToMeta(string xml, XmiVersion ver, Exporter exp)
		{
			XmlDocument xsl = new XmlDocument();
			
			switch(exp) {
                case Exporter.StarUML:
                    xsl.Load(GetResource("XMI2MetaStarUML.xsl"));
                    break;

                case Exporter.PowerAMC:
                    xsl.Load(GetResource("XMI2MetaPowerAMC.xsl"));
                    break;

                case Exporter.PowerDesigner:
					xsl.Load (GetResource ("XMI2MetaPowerAMC.xsl"));
					break;

				case Exporter.Poseidon:
					LoadDependantXsl ("XMI2MetaPoseidon.xsl", xsl);
					break;

				case Exporter.Together:
					LoadDependantXsl ("XMI2MetaTogether.xsl", xsl);
					break;

				case Exporter.VisualParadigm:
					LoadDependantXsl ("XMI2MetaVParadigm.xsl", xsl);
					break;			

				case Exporter.ObjectTeering:
					switch(ver) {
						case XmiVersion.V13:
							xsl.Load (GetResource ("XMI2MetaObjecteering13.xsl"));
							break;

						case XmiVersion.V14:
							xsl.Load (GetResource ("XMI2MetaObjecteering14.xsl"));
							break;

						default : 
							return CleanXml (xml);
					}
					break;
					
				case Exporter.Rose:
					switch(ver) {
						case XmiVersion.V11 :				
							xsl.Load (GetResource ("XMI2MetaConvertor11.xsl"));
							break;

						case XmiVersion.V13:
							xsl.Load (GetResource ("XMI2MetaConvertor13.xsl"));
							break;

						case XmiVersion.V14:
							xsl.Load (GetResource ("XMI2MetaConvertor14.xsl"));
							break;

						case XmiVersion.SimpleXmi:
							xsl.Load (GetResource ("SimpleToMeta.xsl"));
							break;

						default :
							return CleanXml (xml);
					}
					break;

				case Exporter.EA:
					LoadDependantXsl ("XMI2MetaEArchitect.xsl", xsl);
					break;

				case Exporter.Visio:
					break;

				case Exporter.Other:
					break;

			}
			
			XmlDocument xmlDoc = new XmlDocument();

			xml = CleanXml (xml);

			if (exp == Exporter.PowerDesigner || exp == Exporter.Together || 
				exp == Exporter.VisualParadigm || exp == Exporter.Poseidon || exp == Exporter.StarUML) {

				xml = DeleteNamespacePowerDesigner (xml);
			}
		
			xmlDoc.LoadXml (xml);

			using (StringWriter sw = new StringWriter ()) {
				XmlTextWriter sb = new XmlTextWriter (sw);
				sb.Formatting = Formatting.Indented;

				XslTransform xslt = new XslTransform();
				xslt.Load (xsl);
				xslt.Transform (xmlDoc, null, sb);
		
				Dispose();

				return sw.ToString ();
			}
		}

		public static string ConvertToSimple(string xml, XmiVersion ver, Exporter exp)
		{
			if (ver != XmiVersion.SimpleXmi)
			{

				string xmlMeta = ConvertToMeta(xml,ver,exp);
                File.WriteAllText("meta.xml", xmlMeta);
				XmlDocument xmlDoc = new XmlDocument();
				xmlDoc.LoadXml(xmlMeta);
			
				XmlDocument xslDoc = new XmlDocument();

				xslDoc.Load(GetResource ("MetaToSimple.xsl"));

				StringWriter sw = new StringWriter ();

				XslTransform myXsl = new XslTransform();
				myXsl.Load(xslDoc);
				myXsl.Transform(xmlDoc, null,  sw);
				
				Dispose();

				return sw.ToString();
			}
			else
				return xml; 

		}

		public static string DeleteNamespacePowerDesigner(string fileStr)
		{
			fileStr = Regex.Replace(fileStr, @"UML:", String.Empty, RegexOptions.Singleline);
			fileStr = Regex.Replace(fileStr, @"UML:", String.Empty, RegexOptions.Multiline);

			fileStr = Regex.Replace(fileStr, @"UML2:", String.Empty, RegexOptions.Singleline);
			fileStr = Regex.Replace(fileStr, @"UML2:", String.Empty, RegexOptions.Multiline);

			fileStr = Regex.Replace(fileStr, @"Foundation.Core.", String.Empty, RegexOptions.Singleline);
			fileStr = Regex.Replace(fileStr, @"Foundation.Core.", String.Empty, RegexOptions.Multiline);

			fileStr = Regex.Replace(fileStr, @"Model_Management.", String.Empty, RegexOptions.Singleline);
			fileStr = Regex.Replace(fileStr, @"Model_Management.", String.Empty, RegexOptions.Multiline);

			return fileStr;
		}

		public static string DeleteDTD(string fileStr)
		{
			fileStr = Regex.Replace (fileStr, @"<!DOCTYPE.*]>", String.Empty, RegexOptions.Singleline);
			fileStr = Regex.Replace (fileStr, @"<!DOCTYPE.*>", String.Empty, RegexOptions.Multiline);

			return fileStr;
		}

		public static string DeleteSchema(string fileStr)
		{
			//supprime le schema
			//supprime tous les xmlns (\s => n'importe quel caractre de type espace)
			fileStr = Regex.Replace(fileStr,@"\sxmlns[^""]*\""[^""]*\""", String.Empty, RegexOptions.IgnoreCase | RegexOptions.IgnorePatternWhitespace);
			//supprime tous les xsi
			fileStr = Regex.Replace(fileStr,@"\sxsi[^""]*\""[^""]*\""", String.Empty, RegexOptions.IgnoreCase | RegexOptions.IgnorePatternWhitespace);
			fileStr = Regex.Replace(fileStr,@"UML:", String.Empty, RegexOptions.IgnoreCase | RegexOptions.IgnorePatternWhitespace);

			// supprime le schema
			//fileStrResult = Regex.Replace(fileStrResult,@" xmlns=\"".*\""", String.Empty, RegexOptions.None);

			return fileStr;
		}

		static void Dispose()
		{
			File.Delete (GetTempFile ("toXMI.xsl"));
			File.Delete (GetTempFile ("toXMIFunction.xsl"));
			File.Delete (GetTempFile ("XMI2MetaTogether.xsl"));
			File.Delete (GetTempFile ("XMI2MetaVParadigm.xsl"));
			File.Delete (GetTempFile ("XMI2MetaVParadigm.xsl"));
			File.Delete (GetTempFile ("XMI2MetaPoseidon.xsl"));
			File.Delete (GetTempFile ("XMI2MetaEArchitect.xsl"));
		}

	}
}
