using System;
using System.IO;
using System.Reflection;
using System.Runtime.Serialization.Formatters.Binary;
using System.Xml.Serialization;
using System.Text.RegularExpressions;
using System.Collections;

namespace Evaluant.Uss.Common
{
	/// <summary>
	/// Description rsume de Utils.
	/// </summary>
	public class Utils
	{
        static System.Globalization.NumberFormatInfo sNumberFormat;
        
		static Utils()
		{
            sNumberFormat = new System.Globalization.NumberFormatInfo();
            sNumberFormat.NumberDecimalSeparator = ".";
		}

		/// <summary>
		/// Gets a type.
		/// </summary>
		/// <param name="typeName">Name of the type.</param>
		/// <returns></returns>
		public static Type GetType(string typeName)
		{
            // If it's a nullable type, gets the underlying type
            if (typeName[typeName.Length - 1] == Constants.QUESTION_MARK)
                typeName = typeName.Substring(0, typeName.Length - 1);

			Type type = Type.GetType(typeName);

            if (type != null)
            {
                return type;
            }

			switch(typeName.ToLower())
			{
				case "string" : return typeof(string);
				case "datetime" : return typeof(DateTime);
				case "int" : return typeof(int);
				case "sbyte" : return typeof(sbyte);
				case "byte" : return typeof(byte);
				case "char" : return typeof(char);
				case "short" : return typeof(short);
				case "ushort" : return typeof(ushort);
				case "uint" : return typeof(uint);
				case "long" : return typeof(long);
				case "ulong" : return typeof(ulong);
				case "bool" : return typeof(bool);
				case "double" : return typeof(double);
				case "float" : return typeof(float);
                case "guid": return typeof(Guid);
                case "byte[]": return typeof(byte[]);
			}

			string[] values = typeName.Split(Constants.COLON);

            if (values.Length >= 2)
            {
                Assembly assembly = Assembly.LoadWithPartialName(values[1].Trim());
                if (assembly != null)
                    return assembly.GetType(values[0].Trim());

                throw new NullReferenceException("The type " + typeName + " was not found");
            }

			return typeof(object);
		}

		public static string GetCacheKey(Entity entity)
		{
			return GetCacheKey(entity.Type, entity.Id);
		}

		public static string GetCacheKey(string type, string id)
		{
            return String.Concat(type, Constants.COMMA, id); 
		}

        /// <summary>
        /// Determines whether a type is standard .
        /// </summary>
        /// <param name="type">The type.</param>
        /// <returns>
        /// 	<c>true</c> if the specified type is standard ; otherwise, <c>false</c>.
        /// </returns>
        public static bool IsStandardType(Type type)
        {
#if !EUSS11
            return type.IsValueType || type.Equals(typeof(string)) || (type.IsGenericType && type.GetGenericTypeDefinition().Equals(typeof(Nullable<>)));
#else
            return type.IsValueType || type.Equals(typeof(string));
#endif
        }

		/// <summary>
		/// Converts any object to string representation
		/// </summary>
		/// <param name="o">The object to convert</param>
		/// <param name="type">The target type</param>
		/// <returns>The string representation of an object</returns>
        public static string ConvertToString(object o, Type type)
        {
#if !EUSS11
            if (type.IsGenericType && type.GetGenericTypeDefinition().Equals(typeof(Nullable<>)))
            {
                return ConvertToString(o, type.GetGenericArguments()[0]);
            }
#endif
            if (type.IsArray)
            {
                Array array = (Array)o;

                String[] sa = new String[array.Length];
                for (int i = 0; i < array.Length; i++)
                    sa[i] = array.GetValue(i).ToString();

                return String.Join(",", sa);
            }
            else
            {
                if (type == typeof(double))
                    return ((double)o).ToString(sNumberFormat);

                else if (type == typeof(float))
                    return ((float)o).ToString(sNumberFormat);

                else if (type == typeof(DateTime))
                    return ((DateTime)o).Ticks.ToString();

                else if (type == typeof(decimal))
                    return ((decimal)o).ToString(sNumberFormat);

                else if (type == typeof(TimeSpan))
                    return ((TimeSpan)o).Ticks.ToString();

                else
                    return Convert.ToString(o);
            }

        }

		/// <summary>
		/// Converts an string to an object with a specific type
		/// </summary>
		/// <param name="s">The string representation if the object to create</param>
		/// <param name="type">The target type</param>
		/// <returns>A new object with the specified type</returns>
		public static object StringToObject(string s, Type type)
		{

            if (s == null || s == String.Empty)
                return null;

			if(type.IsArray)
			{
                string[] values = s.Split(Constants.COLON);
				Type utype = type.GetElementType();
				Array result = Array.CreateInstance(utype, values.Length);
				for(int i=0; i< values.Length; i++)
					result.SetValue(Convert.ChangeType(values[i], utype), i);

				return result;
			}
			else if(type.IsEnum)
			{
				return s;
			}
			else if (type == typeof(DateTime))
			{
				return new DateTime(long.Parse(s));
			}
            else if (type == typeof(Guid))
            {
                return new Guid(s);
            }
            else if (type == typeof(TimeSpan))
            {
                return TimeSpan.FromTicks(long.Parse(s));
            }
            else
            {
#if !EUSS11
                if (type.IsGenericType && type.GetGenericTypeDefinition().Equals(typeof(Nullable<>)))
                {
					if(type==typeof(DateTime?))
					{
						if (!string.IsNullOrEmpty(s))
							return new DateTime(long.Parse(s));
						return null;
					}
                    System.ComponentModel.NullableConverter nc = new System.ComponentModel.NullableConverter(type);
                    return nc.ConvertFrom(s);
                }
                else
                {
                    return Convert.ChangeType(s, type, sNumberFormat);
                }
#else
                return Convert.ChangeType(s, type, sNumberFormat);
#endif
            }
		}

		/// <summary>
		/// Changes all the entities of a graph to a specific state
		/// </summary>
		/// <param name="entity">The entity graph to change the state</param>
		/// <param name="newState">The new state to apply to any entity</param>
		public static void ChangeGraphStates(Entity entity, State newState)
		{
			entity.State = State.Updating;

			foreach(EntityEntry ee in entity)
			{
				if(ee.Type == typeof(Entity) && ((Entity)ee.Value).State != State.Updating) // Check cyclic references
					ChangeGraphStates((Entity)ee.Value, newState);
			
				// Change the state of the entry, as it represents a Reference if the value is an entity.
				ee.State = newState;
			}

			entity.State = newState;
		}

		/// <summary>
		/// Transforms any Serializable object into a string
		/// </summary>
		/// <param name="o">The object to serialize</param>
		/// <returns>The string representation of the object</returns>
		public static string SerializeToString(object o)
		{
            if (o == null)
                return null;

			BinaryFormatter bf = new BinaryFormatter();
            using (MemoryStream ms = new MemoryStream())
            {
                bf.Serialize(ms, o);
                return Convert.ToBase64String(ms.ToArray());
            }
		}

		/// <summary>
		/// Transforms any Serializable object into a byte array
		/// </summary>
		/// <param name="o">The object to serialize</param>
		/// <returns>The byte array representation of the object</returns>
		public static byte[] SerializeToArray(object o)
		{
            if (o == null)
                return null;
            
            BinaryFormatter bf = new BinaryFormatter();
			MemoryStream ms = new MemoryStream();
			
			bf.Serialize(ms, o);
			byte[] result = ms.ToArray();
			ms.Close();
			
			return result;
		}

		/// <summary>
		/// Unserializes an object from its string representation
		/// </summary>
		/// <param name="s">The string representing the object to unserialize</param>
		/// <returns>The object represented by the string</returns>
		public static object UnSerialize(string s)
		{
            if (s == null || s == String.Empty)
                return null;

            BinaryFormatter bf = new BinaryFormatter();
			MemoryStream ms = new MemoryStream();

			byte[] bytes = Convert.FromBase64String(s);
			ms.Write(bytes, 0, bytes.Length);
			ms.Position = 0;
			object obj = bf.Deserialize(ms);
			ms.Close();

			return obj;
		}

		/// <summary>
		/// Unserializes an object from its byte array representation
		/// </summary>
		/// <param name="bytes">The string representing the object to unserialize</param>
		/// <returns>The object represented by the string</returns>
		public static object UnSerialize(byte[] bytes)
		{
            if (bytes == null)
                return null;

            BinaryFormatter bf = new BinaryFormatter();
			MemoryStream ms = new MemoryStream();

			ms.Write(bytes, 0, bytes.Length);
			ms.Position = 0;
			object obj = bf.Deserialize(ms);
			ms.Close();

			return obj;
		}

		/// <summary>
		/// Clones an object using BinarySerialization
		/// </summary>
		/// <param name="o">The object to clone</param>
		public static object Clone(object o)
		{
			BinaryFormatter bf = new BinaryFormatter();
            using (MemoryStream ms = new MemoryStream())
            {
                bf.Serialize(ms, o);
                ms.Position = 0;
                return bf.Deserialize(ms);
            }
		}

		/// <summary>
		/// Get the string representation of a given Type instance
		/// </summary>
		/// <param name="type">The type to get the string from</param>
		/// <returns></returns>
		/// <remarks>
		/// To get the Type back you must use Type.GetType(). The returned value doesn't contains the Assembly version and token
		/// to prevent re-compilations to make type unloadable due to version number's change.
		/// The Assembly must be already loaded. If it is not loaded it will return null.
		/// </remarks>
		public static string GetFullName(Type type)
		{
			// For primitive types the name is enough as they are defined in mscorlib which will be used by GetType() by default
			if(	type.IsPrimitive 
				|| type == typeof(string) 
				|| type == typeof(DateTime) 
				|| (type.IsArray && type.GetElementType().IsPrimitive))
				return type.FullName;

			return String.Concat(type.FullName, Constants.COLON, Constants.SPACE, type.Assembly.FullName.Split(Constants.COLON)[0]);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="filename"></param>
		/// <param name="type"></param>
		/// <returns></returns>
		public static object DeserializeXmlFile(string filename, Type type)
		{
			TextReader tr = new StreamReader(filename);
			string fileStrResult = Regex.Replace(tr.ReadToEnd(),@"\sxmlns[^""]*\""[^""]*\""", String.Empty, RegexOptions.IgnoreCase | RegexOptions.IgnorePatternWhitespace);
			
			StringReader reader = new StringReader(fileStrResult);

			XmlSerializer deserialize = new XmlSerializer(type);
			object result = deserialize.Deserialize(reader);
			reader.Close();

			return result;
		}

        /// <summary>
        /// Truncate a list of elements specifing the bounds
        /// </summary>
        /// <param name="listIn">the original list</param>
        /// <param name="listOut">the truncated list</param>
        /// <param name="first">the first element to extract</param>
        /// <param name="max">the number of element to extract (if 0, no limit)</param>
        public static void TruncateList(IList listIn, IList listOut, int first, int max)
        {
            if (listIn.Count >= first)
            {
                first = (first < 1) ? 1 : first;
                int end = (max == 0) ? listIn.Count : Math.Min(listIn.Count, first +  max - 1);
                for (int i = first; i <= end; i++)
                {
                    listOut.Add(listIn[i - 1]);
                }
            }
        }
	}
}