using System;
using System.Collections;
using System.Collections.Specialized;
using System.Runtime.Serialization;
using System.ComponentModel;

using Evaluant.Uss.Commands;
using Evaluant.Uss.Collections;
using System.Globalization;

namespace Evaluant.Uss
{
	/// <summary>
	/// Description rsume de Entity.
	/// </summary>
	[Serializable]
	public class Entity : IEditableObject	
	{
		#region Members

		protected ArrayList _Entries;
		protected string _Type;
		protected string _Id; 
		protected State _State;
		protected StringCollection _InferredReferences;
		
		[Flags]
		public enum LoadMetaData { AttributesLoaded = 1, ReferenceLoaded = 2 };

		[NonSerialized]
		protected LoadMetaData _MetaData;
		
		#endregion

		#region Properties

		public LoadMetaData MetaData
		{
			get { return _MetaData; }
			set { _MetaData = value; }
		}

		/// <summary>
		/// Gets the entity entries.
		/// </summary>
		/// <value>The entity entries.</value>
		public ArrayList EntityEntries
		{
			get { return _Entries;  }
			set { _Entries = value; }		// used in RemotingPersistenceEngine
		}

		/// <summary>
		/// Get a collection listing inferred references
		/// </summary>
		public StringCollection InferredReferences
		{
			get 
            {
                // Lazy initialization for memory optimization
                if (_InferredReferences == null)
                    _InferredReferences = new StringCollection();

                return _InferredReferences;  
            }
			set { _InferredReferences = value; }
		}

		#endregion

		#region Ctor

		/// <summary>
		/// Creates a new <see cref="Entity"/> instance.
		/// </summary>
		/// <param name="type">Type.</param>
		public Entity(string type)
		{
			if (type == null)
			{
				throw new ArgumentNullException("type");
			}
			
			this._Entries = new ArrayList(4);
			this._Type = type;
			this._State = State.New;

			GenerateId();

		}

		#endregion

		private void GenerateId()
		{
			// Force the result to begin with a letter for XML compatibility
			_Id = Guid.NewGuid().ToString();
			while( _Id[0] <= '9') 
				_Id = Guid.NewGuid().ToString();
		}

		#region AddValue Methods

		public void AddValue(string name, bool value)
		{
			this.AddValue(name, value, typeof(bool));
		}
 
		public void AddValue(string name, byte value)
		{
			this.AddValue(name, value, typeof(byte));
		}
 
		public void AddValue(string name, char value)
		{
			this.AddValue(name, value, typeof(char));
		}
 
		public void AddValue(string name, DateTime value)
		{
			this.AddValue(name, value, typeof(DateTime));
		}
 
		public void AddValue(string name, decimal value)
		{
			this.AddValue(name, value, typeof(decimal));
		}
 
		public void AddValue(string name, double value)
		{
			this.AddValue(name, value, typeof(double));
		}
 
		public void AddValue(string name, short value)
		{
			this.AddValue(name, value, typeof(short));
		}
 
		public void AddValue(string name, int value)
		{
			this.AddValue(name, value, typeof(int));
		}
 
		public void AddValue(string name, long value)
		{
			this.AddValue(name, value, typeof(long));
		}
 
		public void AddValue(string name, object value)
		{
			if (value == null)
			{
				this.AddValue(name, value, typeof(object));
				return;
			}

			this.AddValue(name, value, value.GetType());
		}
 
		public void AddValue(string name, sbyte value)
		{
			this.AddValue(name, value, typeof(sbyte));
		}
 
		public void AddValue(string name, float value)
		{
			this.AddValue(name, value, typeof(float));
		}
 
		public void AddValue(string name, ushort value)
		{
			this.AddValue(name, value, typeof(ushort));
		}
 
		public void AddValue(string name, uint value)
		{
			this.AddValue(name, value, typeof(uint));
		}
 
		public void AddValue(string name, ulong value)
		{
			this.AddValue(name, value, typeof(ulong));
		}
 
		public void AddValue(string name, object value, Type type)
		{
			AddValue(name, value, type, State.New);
		}

        public void AddValue(string name, Entity value)
        {
            AddValue(name, value, State.New);
        }

		public void AddValue(string name, object value, Type type, State refState)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			if (type == null)
			{
				throw new ArgumentNullException("type");
			}
			
			this._Entries.Add(new EntityEntry(name, value, type, refState, this));

			if(_State == State.UpToDate)
			{
				_State = State.Modified;
			}

		}

		public void AddValue(string name, Entity entity, State refState)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			if (entity == null)
			{
				throw new ArgumentNullException("entity");
			}

            // Ignore relationships to already referenced entities
            foreach (EntityEntry ee in _Entries)
            {
                if (ee.IsEntity && ee.Name == name && ((Entity)ee.Value).Type == entity.Type && ((Entity)ee.Value).Id == entity.Id)
                {
                    // If this entity has been previously deleted, go back to the original state
                    if (ee.State == State.Deleted)
                    {
                        ee.State = State.Modified;
                    }

                    return;
                }
            }

			this._Entries.Add(new EntityEntry(name, entity, typeof(Entity), refState, this));

			if(_State == State.UpToDate)
			{
				_State = State.Modified;
			}
		}
 
		#endregion

		#region Getters

		public bool GetBoolean(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(bool))
			{
				return ((bool) obj1);
			}

            return Convert.ToBoolean(obj1, CultureInfo.InvariantCulture);
		}
 
		public byte GetByte(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(byte))
			{
				return ((byte) obj1);
			}
            
            return Convert.ToByte(obj1, CultureInfo.InvariantCulture);
        }
 
		public char GetChar(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(char))
			{
				return ((char) obj1);
			}

            return Convert.ToChar(obj1, CultureInfo.InvariantCulture);
		}
 
		public DateTime GetDateTime(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(DateTime))
			{
				return ((DateTime) obj1);
			}

            return Convert.ToDateTime(obj1, CultureInfo.InvariantCulture);
		}
 
		public decimal GetDecimal(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(decimal))
			{
				return ((decimal) obj1);
			}

            return Convert.ToDecimal(obj1, CultureInfo.InvariantCulture);
		}
 
		public double GetDouble(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(double))
			{
				return ((double) obj1);
			}

            return Convert.ToDouble(obj1, CultureInfo.InvariantCulture);
		}
 
		public short GetInt16(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(short))
			{
				return ((short) obj1);
			}

            return Convert.ToInt16(obj1, CultureInfo.InvariantCulture);
		}
 
		public int GetInt32(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(int))
			{
				return ((int) obj1);
			}

            return Convert.ToInt32(obj1, CultureInfo.InvariantCulture);
		}
 
		public long GetInt64(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(long))
			{
				return ((long) obj1);
			}

            return Convert.ToInt64(obj1, CultureInfo.InvariantCulture);
		}
 
		public sbyte GetSByte(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(sbyte))
			{
				return ((sbyte) obj1);
			}

            return Convert.ToSByte(obj1, CultureInfo.InvariantCulture);
		}
 
		public float GetSingle(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(float))
			{
				return ((float) obj1);
			}

            return Convert.ToSingle(obj1, CultureInfo.InvariantCulture);
		}
 
		public string GetString(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if ((type1 == typeof(string)) || (obj1 == null))
			{
				return ((string) obj1);
			}

            return Convert.ToString(obj1, CultureInfo.InvariantCulture);
		}
 
		public ushort GetUInt16(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(ushort))
			{
				return ((ushort) obj1);
			}

            return Convert.ToUInt16(obj1, CultureInfo.InvariantCulture);
		}
 
		public uint GetUInt32(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(uint))
			{
				return ((uint) obj1);
			}

            return Convert.ToUInt32(obj1, CultureInfo.InvariantCulture);
		}
 
		public ulong GetUInt64(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			if (type1 == typeof(ulong))
			{
				return ((ulong) obj1);
			}

            return Convert.ToUInt64(obj1, CultureInfo.InvariantCulture);
		}
 
		public Entity GetEntity(string name)
		{
			Type type1;
			object obj1 = this.GetElement(name, out type1);
			return obj1 as Entity;
		}
 
		public EntitySet GetEntitySet(string name)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			EntitySet entities = new EntitySet();

			foreach(EntityEntry ee in this)
			{
				if(ee.IsEntity && ee.Name == name && ee.State != State.Deleted && !entities.Contains(ee.Value)) 
					entities.Add(ee.Value);
			}

			return entities;
		}
 
		#endregion

        /// <summary>
        /// Finds the element among the old entries.
        /// </summary>
        /// <param name="name">Name.</param>
        /// <returns></returns>
        public EntityEntry FindOldEntry(string name)
        {
            if (name == null)
            {
                throw new ArgumentNullException("name");
            }

            for (int i = 0; i < this._OldEntries.Count; ++i)
            {
                if (((EntityEntry)this._OldEntries[i]).Name.Equals(name))
                {
                    return (EntityEntry)this._OldEntries[i];
                }
            }

            return null;
        }

        /// <summary>
        /// Finds a collection of elements among the old entries.
        /// </summary>
        /// <param name="name">Name.</param>
        /// <returns></returns>
        public EntityEntry[] FindOldEntries(string name)
        {
            if (name == null)
            {
                throw new ArgumentNullException("name");
            }

            ArrayList entries = new ArrayList();

            foreach (EntityEntry ee in _OldEntries)
                if (ee.Name == name)
                    entries.Add(ee);

            return (EntityEntry[])entries.ToArray(typeof(EntityEntry));
        }

		/// <summary>
		/// Finds the element among the entries.
		/// </summary>
		/// <param name="name">Name.</param>
		/// <returns></returns>
		public EntityEntry FindEntry(string name)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			for (int i = 0; i < this._Entries.Count; ++i)
			{
				if( ((EntityEntry)this._Entries[i]).Name.Equals(name) )
				{
					return (EntityEntry)this._Entries[i];
				}
			}

			return null;
		}
		/// <summary>
		/// Finds a collection of elements among the entries.
		/// </summary>
		/// <param name="name">Name.</param>
		/// <returns></returns>
		public EntityEntry[] FindEntries(string name)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			ArrayList entries = new ArrayList();

			foreach(EntityEntry ee in this)
				if(ee.Name == name)
					entries.Add(ee);

			return (EntityEntry[])entries.ToArray(typeof(EntityEntry));
		}
		
		/// <summary>
		/// Gets the value of the specified element.
		/// </summary>
		/// <param name="name">Name.</param>
		/// <param name="type">Type.</param>
		/// <returns></returns>
		public object GetValue(string name, Type type)
		{
			Type type1;
			
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			if (type == null)
			{
				throw new ArgumentNullException("type");
			}
			
			object obj1 = this.GetElement(name, out type1);
			
			if (((type1 == type) || type.IsAssignableFrom(type1)) || (obj1 == null))
			{
				return obj1;
			}

            return Convert.ChangeType(obj1, type, CultureInfo.InvariantCulture);
		}
 
		/// <summary>
		/// Gets the value of the specified element.
		/// </summary>
		/// <param name="name">Name.</param>
		/// <returns></returns>
		public object GetValue(string name)
		{
			EntityEntry entry = this.FindEntry(name);
			if (entry == null)
			{
				return null;
			}

			return entry.Value;
		}

        /// <summary>
        /// Determines whether the specified attribute is null.
        /// </summary>
        /// <param name="attribute">The attribute.</param>
        /// <returns>
        /// 	<c>true</c> if the specified name is null; otherwise, <c>false</c>.
        /// </returns>
        public bool IsNull(string attribute)
        {
            EntityEntry ee = this.FindEntry(attribute) ;
            return ee == null ? true : ee.Value == null;
        }
 
		/// <summary>
		/// Mark the value as Deleted.
		/// </summary>
		/// <remarks>
		/// If Detele is called on an entity, the Reference is Deleted, not the Entity
		/// </remarks>
		/// <param name="name">Name.</param>
		public void DeleteElement(string name)
		{
			EntityEntry[] entries = this.FindEntries(name);
			
			if (entries.Length == 0)
		                return;

			// If Detele is called on an entity, the Reference is Deleted, not the Entity
			for(int i= entries.Length-1; i>=0; i--)
			{
				if(entries[i].State != State.New)
					entries[i].State = State.Deleted;
				else
					_Entries.Remove(entries[i]);
			}

			if(_State == State.UpToDate)
			{
				_State = State.Modified;
			}

		}
 
		/// <summary>
		/// Mark the value as Deleted.
		/// </summary>
		/// <remarks>
		/// If Detele is called on an entity, the Reference is Deleted, not the Entity
		/// </remarks>
		/// <param name="name">Name.</param>
		public void DeleteElement(string name, Entity e)
		{
			EntityEntry[] entries = this.FindEntries(name);
			
			if (entries.Length == 0)
                return;

			// If Detele is called on an entity, the Reference is Deleted, not the Entity
			for(int i= entries.Length-1; i>=0; i--)
				if(entries[i].Value == e)
				{
					if(entries[i].State != State.New)
						entries[i].State = State.Deleted;
					else
						_Entries.Remove(entries[i]);
				}

			if(_State == State.UpToDate)
			{
				_State = State.Modified;
			}

		}
 
		/// <summary>
		/// Removes an element from the entries.
		/// </summary>
		/// <param name="name">Name.</param>
		public void RemoveAny(string name)
		{
			EntityEntry[] entries = this.FindEntries(name);
			
			if (entries.Length == 0)
				return;
			
			foreach(EntityEntry entry in entries)
			{
				_Entries.Remove(entry);
			}
		}
 
		/// <summary>
		/// Removes all entries, attributes and references.
		/// </summary>
		public void RemoveAny()
		{
			_Entries.Clear();
		}
 
		/// <summary>
		/// Removes all attributes.
		/// </summary>
		public void RemoveAttribute()
		{
			int i = 0;
			while(i < _Entries.Count)
			{
				if( !(((EntityEntry)_Entries[i]).IsEntity))
					_Entries.RemoveAt(i);
				else
					i++;
			}
		}
 
		/// <summary>
		/// Removes all attributes.
		/// </summary>
		public void RemoveAttribute(string name)
		{
			EntityEntry[] entries = this.FindEntries(name);
			
			if (entries.Length == 0)
			{
				return;
			}
			
			foreach(EntityEntry entry in entries)
			{
				if( !(entry.IsEntity))
					_Entries.Remove(entry);
			}
		}
 
		/// <summary>
		/// Removes all references without marking them as deleted
		/// </summary>
		public void RemoveReference()
		{
			int i = 0;
			while(i < _Entries.Count)
			{
				if( ((EntityEntry)_Entries[i]).IsEntity)
					_Entries.RemoveAt(i);
				else
					i++;
			}
		}
 
        /// <summary>
		/// Removes all references without marking them as deleted
		/// </summary>
		/// <param name="names">The names of the references to remove</param>
        public void RemoveReference(string[] names)
        {
            // Removes all references that will be updated
            if (names == null || names.Length == 0)
                RemoveReference();
            else
                foreach (string name in names)
                    RemoveReference(name);
        }

		/// <summary>
		/// Removes all references without marking them as deleted
		/// </summary>
		/// <param name="name">The name of the references to remove</param>
		public void RemoveReference(string name)
		{
			EntityEntry[] entries = this.FindEntries(name);
			
			if (entries.Length == 0)
			{
				return;
			}
			
			foreach(EntityEntry entry in entries)
			{
				if(entry.IsEntity)
					_Entries.Remove(entry);
			}
		}
 
		/// <summary>
		/// Removes a reference without marking it as deleted
		/// </summary>
		/// <param name="name">The name of the references to remove</param>
		/// <param name="id">The entity's id to remove in the named references</param>
		public void RemoveReference(string name, string id)
		{
			EntityEntry[] entries = this.FindEntries(name);
			
			if (entries.Length == 0)
			{
				return;
			}
			
			foreach(EntityEntry entry in entries)
			{
				if(entry.IsEntity && (entry.Value as Entity).Id == id)
					_Entries.Remove(entry);
			}
		}
 
		/// <summary>
		/// Removes all references without marking them as deleted
		/// </summary>
		/// <param name="entity">The entity to remove from the references</param>
		public void RemoveReference(Entity entity)
		{
			int i = 0;
			while(i < _Entries.Count)
			{
				EntityEntry ee = (EntityEntry)_Entries[i];

				if( ee.IsEntity && ee.Value == entity)
					_Entries.RemoveAt(i);
				else
					i++;
			}
		}
 
		/// <summary>
		/// Gets the type.
		/// </summary>
		/// <value></value>
		public string Type
		{
			get
			{
				return this._Type;
			}
		}
 
		/// <summary>
		/// Gets the member count.
		/// </summary>
		/// <value></value>
		public int MemberCount
		{
			get
			{
				return this._Entries.Count;
			}
		}
 
		/// <summary>
		/// Gets or sets the related value with the specified name.
		/// </summary>
		/// <value></value>
		public object this [string name]
		{
			get
			{
				return GetValue(name);
			}
			
			set
			{
                SetValue(name, value);
			}
		}

		/// <summary>
		/// Sets the value of an attribute if the values differ.
		/// </summary>
		/// <param name="name">The name.</param>
		/// <param name="value">The value.</param>
		public void SetValue(string name, object value)
		{
			EntityEntry entry = this.FindEntry(name);

			if (entry != null)
			{
				if(entry.Value == null || !entry.Value.Equals(value))
				{
					if(value == null)
						entry.Type = typeof(object);
					else	
						entry.Type = value.GetType();
					
					entry.Value = value;

					if(entry.State == State.UpToDate)
						entry.State = State.Modified;
				}
			}
			else
				AddValue(name, value);
		}

		/// <summary>
		/// Gets or sets the <see cref="Evaluant.Uss.EntityEntry"/> with the specified name.
		/// </summary>
		/// <value></value>
		public object this [string name, int index]
		{
			get
			{
				EntityEntry[] entries = this.FindEntries(name);
				if (entries.Length > index )
				{
					throw new SerializationException(string.Format("{0} not found", name));
				}

				return entries[index].Value;
			}
		}

		/// <summary>
		/// Gets a value with its underlying type.
		/// </summary>
		/// <param name="name">Name.</param>
		/// <param name="type">Type.</param>
		/// <returns></returns>
		private object GetElement(string name, out Type type)
		{
			EntityEntry entry = this.FindEntry(name);
			if (entry == null)
			{
				type = null;
				return null;
			}
			
			type = entry.Type;
			return entry.Value;
		}
 
		public EntityEnumerator GetEnumerator()
		{
			return new EntityEnumerator(this._Entries);
		}

		/// <summary>
		/// Gets or sets the id. Every objects have an Id, even if they are not yet synchronized
		/// </summary>
		/// <value></value>
		public string Id
		{
			get { return _Id; }
			set 
			{ 
				if(_State != State.New)
					throw new PersistenceEngineException("Cannot change Entity'Id if it's state is not New");

				_Id = value; 
			}
		}

		public State State
		{
			get { return _State; }
			set { _State = value; }
		}

		public static CommandCollection ComputeChanges(Entity entity)
		{
			CommandCollection commands = new CommandCollection();
			HashedList processing = new HashedList();

			ComputeChanges(entity, processing, commands);

			return commands;
		}

        internal static CommandCollection ComputeChanges(Entity entity, bool ignoreReferences)
        {
            CommandCollection commands = new CommandCollection();
            HashedList processing = new HashedList();
            if (ignoreReferences)
            {
                foreach (EntityEntry entry in entity)
                {
                    if (!(!entry.IsEntity || processing.Contains(entry.Value)))
                    {
                        processing.Add(entry.Value);
                    }
                }
            }
            ComputeChanges(entity, processing, commands);
            return commands;
        }

		/// <summary>
		/// Generated a set of commands to be processed to take into account any modification applied to an entity graph
		/// </summary>
		/// <param name="entity">The Entity to compute</param>
		/// <param name="processing">A collection of all currently processed entities</param>
		/// <param name="commands">A collection of all currently created commands</param>
		internal static void ComputeChanges(Entity entity, HashedList processing, CommandCollection commands)
		{
			// Nothing to do if the entity is currently processed
			if(processing.Contains(entity))
				return ;

			processing.Add(entity);
			
			// To know if we can optimize the command with a compound
			bool canCreateCompound = true;

			// Will contain the commands created for the attributes of the current entity
			ArrayList attributeCommands = new ArrayList(entity.MemberCount);

			// First process all attributes not to increase the stack size because of deep hierarchies in the object graph and the recursive algorithm
			if(entity.State == State.Modified || entity.State == State.New)
				foreach(EntityEntry e in entity)
				{
					if(!e.IsEntity)
					{
						// The current entry is an Attribute
						switch(e.State)
						{
							case State.Deleted:
                                attributeCommands.Add(new UpdateAttributeCommand(entity.Id, entity.Type, e.Name, e.Type, null));
								break;

							case State.Modified:
								attributeCommands.Add(new UpdateAttributeCommand(entity.Id, entity.Type, e.Name, e.Type, e.Value));
								break;

							case State.New:
								attributeCommands.Add(new CreateAttributeCommand(entity.Id, entity.Type, e.Name, e.Type, e.Value));
								break;
							default:
								break;
						}
					}
				}

			// Selects the optimized command
			/// TODO: Add a property to PE so that the user could specify whether he wants all attributes to be sent with a compound command
			switch(entity.State)
			{
				case State.New :
					if(canCreateCompound)
					{
						// Loops over all inner attributes to add their commands
						attributeCommands.Clear();
						foreach(EntityEntry e in entity)
							if(!e.IsEntity)
								attributeCommands.Add(new CreateAttributeCommand(entity.Id, entity.Type, e.Name, e.Type, e.Value));
					
						commands.Add(new CompoundCreateCommand(entity.Id, entity.Type, attributeCommands));
					}
					else
					{
						commands.Add(new CreateEntityCommand(entity.Id, entity.Type));
						commands.AddRange(attributeCommands);
					}
						
					break;
				
				case State.Deleted :
					commands.Add(new DeleteEntityCommand(entity.Id, entity.Type));
					break;

				case State.Modified :
				
					if(attributeCommands.Count > 1)
					{
						CompoundUpdateCommand cuc = new CompoundUpdateCommand(entity.Id, entity.Type, attributeCommands);
						commands.Add(cuc);
					}
					else
						commands.AddRange(attributeCommands);
					break;
				
				default:
					commands.AddRange(attributeCommands);
					break;
			}

			foreach(EntityEntry e in entity)
			{
				if(e.IsEntity)
				{
					// The current entry is an Entity
					Entity et = (Entity)e.Value;
					ComputeChanges(et, processing, commands);
					
					switch(e.State)
					{
						case State.New :
							CreateReferenceCommand crc1 = new CreateReferenceCommand(e.Name, entity.Id, entity.Type, et.Id, et.Type);
							commands.Add(crc1);
							break;
											
						case State.Deleted :
							DeleteReferenceCommand drc = new DeleteReferenceCommand(e.Name, entity.Id, entity.Type, et.Id, et.Type);
							commands.Add(drc);
							break;
					}
				}
			}
		}

		public void AcceptChanges()
		{
			if(_State == State.Updating)
				return;

			_State = State.Updating;

			// Removes all entries which are marked as deleted
			for(int i = _Entries.Count - 1; i >= 0 ; i--)
				if(((EntityEntry)_Entries[i]).State == State.Deleted)
					_Entries.RemoveAt(i);

			// Change State to UpToDate to all entries
			foreach(EntityEntry entry in _Entries)
			{
				if(entry.IsEntity)
					((Entity)entry.Value).AcceptChanges();

				entry.State = State.UpToDate;
			}

			_State = State.UpToDate;
		}

		/// <summary>
		/// Gets the changes.
		/// </summary>
		/// <returns></returns>
		public CommandCollection GetChanges()
		{
			CommandCollection commands = new CommandCollection();
			HashedList processed = new HashedList();
			
			ComputeChanges(this, processed, commands);

			return commands;
		}

        
		/// <summary>
		/// Clones this instance.
		/// </summary>
		/// <returns></returns>
        public Entity Clone()
        {
            return Clone(null);
        }

        /// <summary>
        /// Clones this instance.
        /// </summary>
        /// <param name="attributes">
        /// null	        -> do not load attribute
        /// new string[0]	-> load all attributes
        /// other values    -> load the specified attributes</param>
        /// <returns></returns>
		public Entity Clone(string[] attributes)
		{

            Entity entity = new Entity(this.Type);
            entity.Id = this.Id;
            
            if (attributes != null)
            {
                foreach (EntityEntry ee in this)
                {
                    if (!ee.IsEntity && (attributes.Length == 0 || Array.IndexOf(attributes, ee.Name) != -1))
                        entity.AddValue(ee.Name, ee.Value, ee.Type, ee.State);
                }
            }

            entity.State = this.State;

			return entity;
		}

		public override string ToString()
		{
			return String.Concat(Type, " (", Id, ")");
		}

		#region IEditableObject members

		protected ArrayList _OldEntries;
		protected bool _Editing;

		// Explicit implementation to hide those methods to the standard user

		void IEditableObject.BeginEdit()
		{
			if(!_Editing)
			{
				// If entries point to an entity the reference is kept
				ArrayList cloned = new ArrayList(_Entries.Count);
				foreach(EntityEntry ee in _Entries)
					cloned.Add(ee.Clone());

				_OldEntries = _Entries;
				_Entries = cloned;

				_Editing = true;
			}
		}

		void IEditableObject.EndEdit()
		{
			_Editing = false;
		}

		void IEditableObject.CancelEdit()
		{
			if(_Editing)
			{
				_Entries = _OldEntries;
				_Editing = false;
			}
		}

		#endregion
	}
}
