using System;
using System.Collections;
using System.Runtime.Serialization.Formatters.Soap;
using System.IO;

using NUnit.Framework;
using Evaluant.Uss.ObjectContext;
using Evaluant.Uss;
using Evaluant.Uss.ObjectContext.Descriptors;

namespace NUnit.DynamicProxy
{
	/// <summary>
	/// Description rsume de Proxy.
	/// </summary>
	[TestFixture]
	public class Proxy
	{

		public Evaluant.Uss.ObjectContext.ObjectContext pm;

		public Proxy()
		{
			Configure();
		}

		public virtual void Configure()
		{
            IPersistenceProvider provider = XmlConfigLoader.LoadXmlConfig("engines.config", "Prevalence");
            provider.RegisterMetaData(MetaDataFactory.FromAssembly(typeof(Person).Assembly));
            provider.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\PersistenceEngine\Models\Person.xml"));
            provider.InitializeConfiguration();

            this.pm = new ObjectService(provider).CreateObjectContext();
		}

		[Test]
		public virtual void PropertyTypeIsNotFieldType()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";
			p1.Marks = new int[] { 1, 10, 19 };

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = (Person)pm.Load(typeof(Person))[0];

			Assert.IsNotNull(p1);
			Assert.AreEqual(3, p1.Marks.Length);
			Assert.AreEqual(10, p1.Marks[1]);
		}

		[Test]
		public virtual void TestEnumeration()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";
			p1.Kind = KindEnum.Nice;
			p1.Title = TitleEnum.Sir;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Person p = (Person) pm.Load(typeof(Person), "[Name = 'p1']")[0];

			Assert.IsNotNull(p);
			Assert.AreEqual(TitleEnum.Sir, p.Title);
			Assert.AreEqual(KindEnum.Nice, p.Kind);

			p.Title = TitleEnum.Madam;
			p.Kind = KindEnum.Bad;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = (Person) pm.Load(typeof(Person), "[Name = 'p1']")[0];

			Assert.IsNotNull(p);
			Assert.AreEqual(TitleEnum.Madam, p.Title);
			Assert.AreEqual(KindEnum.Bad, p.Kind);

			p = (Person) pm.Load(typeof(Person), string.Format("[Title = '{0}']", Enum.GetName(typeof(TitleEnum), TitleEnum.Madam)))[0];

			Assert.IsNotNull(p);
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual(KindEnum.Bad, p.Kind);

			p = (Person) pm.Load(typeof(Person), string.Format("[Kind = {0}]", (int)p.Kind))[0];

			Assert.IsNotNull(p);
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual(TitleEnum.Madam, p.Title);
		}

		[Test]
		public void Overral()
		{

			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Person partner = new Person();
			partner.Name = "Joey Triviani";
			partner.Age = 32;

			p.Partners.Add(partner);
	
			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(2, pm.Load(typeof(Person)).Count);
			
			p = pm.Load(typeof(Person), "[Name = 'John Smith']")[0] as Person;
			Assert.IsNotNull(p);

			Assert.AreEqual("John Smith", p.Name);
			Assert.AreEqual(50, p.Age);

			Assert.AreEqual(1, p.Partners.Count);
			Assert.AreEqual("Joey Triviani", ((Person)p.Partners[0]).Name);

			p.Age = 49;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load(typeof(Person), "[Name = 'John Smith']").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[Age = 49]").Count);
		}

		[Test]
		public virtual void OPathException()
		{
			try
			{
				pm.Load(typeof(Person), "Person[ Test and or Toto ]");
			}
			catch(Exception e)
			{
				Assert.AreEqual(e.GetType(), typeof(OPath.OPathException));
				return;
			}

			Assert.Fail();
		}

		[Test]
		public void ManyToOne()
		{

			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			Address a = new Address("Paris");

			p1.Address = a;
			p2.Address = a;
			p3.Address = a;
            
			pm.BeginTransaction();
			pm.Serialize(new object[] { p1, p2, p3 });
			pm.CommitTransaction();
            
			Assert.AreEqual(1, pm.Load(typeof(Address)).Count);
		}

		[Test]
		public void RemoveRelationship()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			p1.Partners.Add(p2);
			p1.Partners.Add(p3);
            
			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
            
			pm.Clear();

			Assert.AreEqual(3, pm.Load(typeof(Person)).Count);

			p1 = pm.LoadSingle(typeof(Person), "[Name = 'p1']") as Person;

			Assert.AreEqual(2, p1.Partners.Count);

			p1.Partners.Clear();

			Assert.AreEqual(0, p1.Partners.Count);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(3, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(0, pm.Load(typeof(Person), "[Name = 'p1'].Partners").Count);

			p1 = pm.LoadSingle(typeof(Person), "[Name = 'p1']") as Person;
			
			Assert.AreEqual(0, p1.Partners.Count);
		}

		[Test]
		public void Id()
		{
			pm.InitializeRepository();

			Address a = new Address();
			a.City = "New York";
			a.Id = "Address01";

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;
			p.Address = a;
			p.PersonId = "Person01";

			Person partner = new Person();
			partner.Name = "Joey Triviani";
			partner.Age = 32;
			partner.PersonId = "Person02";

			p.Partners.Add(partner);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = (Person)pm.LoadWithId(typeof(Person), "Person01");

			Assert.IsNotNull(p);
			Assert.AreEqual("Person01", p.PersonId);
			Assert.AreEqual("Address01", p.Address.Id);
			Assert.AreEqual("Person02", ((Person)p.Partners[0]).PersonId);

			pm.Clear();

			p = pm.Load(typeof(Person), "[Name = 'John Smith']")[0] as Person;

			Assert.IsNotNull(p);
			Assert.AreEqual("Person01", p.PersonId);
			Assert.AreEqual("Address01", p.Address.Id);
			Assert.AreEqual("Person02", ((Person)p.Partners[0]).PersonId);
		}

		[Test]
		public void OneToOneSerialize()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Address a = new Address("New York");
			p.Address = a;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Address)).Count);
			Assert.AreEqual("New York", ((Address)pm.Load(typeof(Address))[0]).City);
		}

		[Test]
		public void OneToOneLazyLoad()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Address a = new Address("New York");
			p.Address = a;

			Country c = new Country("United States");
			a.Country = c;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = (Person)pm.LoadSingle(typeof(Person));
			
			Assert.IsFalse (p.BaseAddressGetterCalled);
			Assert.IsNotNull(p.Address);
			Assert.IsTrue(p.BaseAddressGetterCalled);
			Assert.AreEqual("New York", p.Address.City);

			Assert.IsFalse (p.Address.BaseCountryGetterCalled);
			Assert.IsNotNull(p.Address.Country);
			Assert.IsTrue(p.Address.BaseCountryGetterCalled);
			Assert.AreEqual("United States", p.Address.Country.Name);

		}

		[Test]
		public void Update()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;
			p.Birth = new DateTime(2005, 12, 28);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = (Person)pm.LoadSingle(typeof(Person));

			Assert.AreEqual("John Smith", p.Name);
			Assert.AreEqual(50, p.Age);
			Assert.AreEqual(new DateTime(2005, 12, 28), p.Birth);

			p.Name = "Doug Ross";
			p.Age = 40;
			p.Birth = new DateTime(2005, 12, 29);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load(typeof(Person)).Count);
			p = (Person)pm.LoadSingle(typeof(Person));
			Assert.AreEqual("Doug Ross", p.Name);
			Assert.AreEqual(40, p.Age);
			Assert.AreEqual(new DateTime(2005, 12, 29), p.Birth);
		}

		[Test]
		public void IdentityMap()
		{			
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Person p1 = (Person)pm.LoadSingle(typeof(Person));
			Person p2 = (Person)pm.LoadSingle(typeof(Person));
			
			Assert.IsNotNull(p1);
			Assert.IsNotNull(p2);
			Assert.IsTrue(p1 == p2);
		}

		[Test]
		public void CascadeUpdate()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Address a1 = new Address();
			a1.City = "a1";

			Person p2 = new Person();
			p2.Name = "p2";

			p1.Partners.Add(p2);
			p1.Address = a1;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle(typeof(Person), "[Name = 'p1']") as Person;
			p2 = ((Person)p1.Partners[0]);
			a1 = p1.Address;
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual("p2", p2.Name);
			Assert.AreEqual("a1", a1.City);

			p2.Name = "p2 changed";
			a1.City = "a1 changed";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle(typeof(Person), "[Name = 'p1']") as Person;
			p2 = ((Person)p1.Partners[0]);
			a1 = p1.Address;
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual("p2 changed", p2.Name);
			Assert.AreEqual("a1 changed", a1.City);

		}

		[Test]
		public void Inheritance()
		{
			pm.InitializeRepository();

			Employee bill = new Employee();
			bill.Name = "Bill Gates";
			bill.Salary = 999999f;

			Address redmond = new Address();
			redmond.City = "Redmond";
			bill.Address = redmond;

			Address paloalto = new Address();
			paloalto.City = "Palo Alto";

			Person steve = new Person();
			steve.Name = "Steve Jobs";
			steve.Address = paloalto;

			bill.Partners.Add(steve);
            
			pm.BeginTransaction();
			pm.Serialize(bill);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(2, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(2, pm.Load(typeof(Address)).Count);

			bill = (Employee)pm.LoadSingle(typeof(Employee), "[Name = 'Bill Gates']");
			Assert.IsNotNull(bill);
			Assert.AreEqual(999999f, bill.Salary);
			Assert.AreEqual("Steve Jobs", ((Person)bill.Partners[0]).Name);
			Assert.AreEqual("Redmond", bill.Address.City);

		}

		[Test]
		public void Composition()
		{
			pm.InitializeRepository();

			Address paloalto = new Address("Palo Alto");

			Person steve = new Person();
			steve.Name = "Steve Jobs";
			steve.Address = paloalto;

			pm.BeginTransaction();
			pm.Serialize(steve);
			pm.CommitTransaction();

			Assert.AreEqual(1, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Address)).Count);

			steve = pm.LoadSingle(typeof(Person)) as Person;

			pm.BeginTransaction();
			pm.Delete(steve);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(0, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(0, pm.Load(typeof(Address)).Count);
		}

		[Test]
		public void Id2()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.PersonId = "p1";
			p1.Name = "Mike Tyson";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.IsNotNull(pm.LoadWithId(typeof(Person), "p1"));

			pm.InitializeRepository();

			Employee e1 = new Employee();
			e1.PersonId = "e1";
			e1.Name = "Peter Pan";

			pm.BeginTransaction();
			pm.Serialize(e1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.IsNotNull(pm.LoadWithId(typeof(Person), "e1"));
			Assert.IsNotNull(pm.LoadWithId(typeof(Employee), "e1"));
		}


		public void ExistingObjects()
		{
			//	Person and Country exists in database
			//	Creates a new Address and reserialize the Person
			//	Add the Country to the new Address
			//	Add the Address to the new Person

			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "Steve Jobs";
			p.Age = 49;

			Country c = new Country();
			c.Name = "United States";

			pm.BeginTransaction();
			pm.Serialize(new object[] { p, c });
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Country)).Count);

			p = (Person) pm.LoadSingle(typeof(Person));
			c = (Country) pm.LoadSingle(typeof(Country));

			Address a = new Address();
			a.City = "New York";

			a.Country = c;

			p.Address = a;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = (Person) pm.LoadSingle(typeof(Person));

			Assert.IsNotNull(p);
			Assert.IsNotNull(p.Address);
			Assert.IsNotNull(p.Address.Country);
		}

		[Test]
		public void ExistingObject2()
		{
			//	Address exists in database
			//	Creates a Person and a Country
			//	Add the Address to the new Person
			//	Add the Country to the Address
			//	Serialize the Person

			pm.InitializeRepository();

			Address a = new Address();
			a.City = "New York";

			pm.BeginTransaction();
			pm.Serialize(a);
			pm.CommitTransaction();

			pm.Clear();

			a = (Address) pm.LoadSingle(typeof(Address));

			Person p = new Person();
			p.Name = "Steve Jobs";

			Country c = new Country();
			c.Name = "United States";

			a.Country = c;

			p.Address = a;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.IsNotNull(p);
			Assert.IsNotNull(p.Address);
			Assert.IsNotNull(p.Address.Country);
		}

		[Test]
		public void ExistingObjects3()
		{ 
			//	Person p1 p2 and p4 exist in database
			//	p2 is a partner of p1
			//	Load p1 and creates a new Person p3 and add it to p1 as partner
			//	Load p1 and add p4 as partners of p1 at least

			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "Bill Gates";

			Person p2 = new Person();
			p2.Name = "Steve Jobs";

			Person p4 = new Person();
			p4.Name = "John Smith";

			p1.Partners.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p4);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(3, pm.Load(typeof(Person)).Count);

			pm.Clear();

			p1 = (Person) pm.Load(typeof(Person), "[Name = 'Bill Gates']")[0];

			Assert.IsNotNull(p1);
			Assert.AreEqual(1, p1.Partners.Count);

			Person p3 = new Person();
			p3.Name = "Joey Triviani";

			p1.Partners.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(4, pm.Load(typeof(Person)).Count);

			p1 = (Person) pm.Load(typeof(Person), "[Name = 'Bill Gates']")[0];

			Assert.IsNotNull(p1);
			Assert.AreEqual(2, p1.Partners.Count);

			pm.Clear();

			p1 = (Person) pm.Load(typeof(Person), "[Name = 'Bill Gates']")[0];
			p4 = (Person) pm.Load(typeof(Person), "[Name = 'John Smith']")[0];

			Assert.IsNotNull(p1);
			Assert.IsNotNull(p4);

			p1.Partners.Add(p4);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(4, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(3, pm.Load(typeof(Person), "[Name = 'Bill Gates'].Partners").Count);
		}

		[Test]
		public virtual void Enumeration()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "Elton";
			p1.Title = TitleEnum.Sir;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();
			
			IList coll = pm.Load(typeof(Person), "[Title = 'Sir']");
			Assert.AreEqual(1, coll.Count);

			coll = pm.Load(typeof(Person), "[Name = 'Elton']");
			Assert.AreEqual("Elton", ((Person)coll[0]).Name);
		}

		[Test]
		public void Detach()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			p1.Partners.Add(p2);
			p1.Partners.Add(p3);

			p3.Partners.Add(p1);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
	
			pm.Clear();

			p1 = pm.LoadSingle(typeof(Person), "[Name = 'p1']") as Person;
			Assert.AreEqual(2, p1.Partners.Count); // Lazyload partners for p1

			p1 = pm.Detach(p1) as Person;

			Assert.AreEqual(typeof(Person), p1.GetType());
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual(2, p1.Partners.Count); // Check lazyloaded items are there
			Assert.AreEqual(0, ((Person)p1.Partners[1]).Partners.Count); // Check none lazyloaded items are not there
		}

		[Test]
		public virtual void LoadWithQuerySort()
		{
			// one field
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.PersonId = "1";
			p1.Name = "p01";
			p1.Birth = DateTime.Today;

			Person p2 = new Person();
			p2.PersonId = "3";
			p2.Name = "p02";
			p2.Birth = DateTime.Today.AddDays(1);

			Person p3 = new Person();
			p3.PersonId = "2";
			p3.Name = "p02";
			p3.Birth = DateTime.Today;

			Person p4 = new Person();
			p4.PersonId = "4";
			p4.Name = "p04";
			p4.Birth = DateTime.Today.AddDays(1);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.Serialize(p3);
			pm.Serialize(p4);
			pm.CommitTransaction();

			Query query = new Query(typeof(Person));
			query.OrderBy.Add("Name");
			IList results =	pm.Load(query);
			
			Assert.AreEqual("p01", ((Person)results[0]).Name);
			Assert.AreEqual("p02", ((Person)results[1]).Name);
			Assert.AreEqual("p02", ((Person)results[2]).Name);
			Assert.AreEqual("p04", ((Person)results[3]).Name);

			query = new Query(typeof(Person));
			query.OrderBy.Add("Name desc");
			results = pm.Load(query);

			Assert.AreEqual("p04", ((Person)results[0]).Name);
			Assert.AreEqual("p02", ((Person)results[1]).Name);
			Assert.AreEqual("p02", ((Person)results[2]).Name);
			Assert.AreEqual("p01", ((Person)results[3]).Name);

			// two fields
			query = new Query(typeof(Person));
			query.OrderBy.Add("Name");
			query.OrderBy.Add("Birth");
			results =	pm.Load(query);
			
			Assert.AreEqual("1", ((Person)results[0]).PersonId);
			Assert.AreEqual("2", ((Person)results[1]).PersonId);
			Assert.AreEqual("3", ((Person)results[2]).PersonId);
			Assert.AreEqual("4", ((Person)results[3]).PersonId);

			query = new Query(typeof(Person));
			query.OrderBy.Add("Name desc");
			query.OrderBy.Add("Birth desc");
			results = pm.Load(query);

			Assert.AreEqual("4", ((Person)results[0]).PersonId);
			Assert.AreEqual("3", ((Person)results[1]).PersonId);
			Assert.AreEqual("2", ((Person)results[2]).PersonId);
			Assert.AreEqual("1", ((Person)results[3]).PersonId);
		}

		[Test]
		public virtual void LoadWithQueryMinMax()
		{
			pm.InitializeRepository();

			ArrayList personColl = new ArrayList();

			for(int i=1; i<=50; i++)
			{
				Person p = new Person();
				p.Name = "p" + i.ToString("00");
				personColl.Add(p);
			}

			pm.BeginTransaction();
			pm.Serialize(personColl);
			pm.CommitTransaction();

			Query query = new Query(typeof(Person));
			query.OrderBy.Add("Name");
			query.FirstResult = 1;
			query.MaxResults = 100;
			IList results =	pm.Load(query);

			Assert.AreEqual(50, results.Count);

			query.FirstResult = 10;
			query.MaxResults = 21;
			results =	pm.Load(query);

			Assert.AreEqual(21, results.Count);
			Assert.AreEqual("p10", ((Person)results[0]).Name);
			Assert.AreEqual("p20", ((Person)results[10]).Name);
			Assert.AreEqual("p30", ((Person)results[20]).Name);

			query.FirstResult = 10;
			query.MaxResults = 0;
			results = pm.Load(query);

			Assert.AreEqual(41, results.Count);
			Assert.AreEqual("p10", ((Person)results[0]).Name);
			Assert.AreEqual("p20", ((Person)results[10]).Name);
			Assert.AreEqual("p50", ((Person)results[40]).Name);

			query = new Query(typeof(Person), "[Name ^ 'p']");
			query.OrderBy.Add("Name");
			query.MaxResults = 10;
			results = pm.Load(query);

			Assert.AreEqual(10, results.Count);
		}

		[Test]
		public virtual void NullValueAttribute()
		{
			pm.InitializeRepository();

			// serialize an object Person with an attribute which has a value null
			Person p1 = new Person();
			p1.PersonId = "p1";
			p1.Name = null;
			p1.Birth = DateTime.Now;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			// load the object PersonProxy
			pm.Clear();
			Person pp1 = (Person)pm.LoadSingle(typeof(Person), "[id('p1')]");

			Assert.IsNull(pp1.Name);

			pp1.Age = 30;

			// serialize the object PersonProxy   (CreateAttributCommand with null value)
			pm.BeginTransaction();
			pm.Serialize(pp1);
			pm.CommitTransaction();

			Assert.IsNull(pp1.Name);
			pp1.Name = "Joe";

			// serialize the object PersonProxy again (UpdateAttributeCommand for an non existing attribute in repository)
			pm.BeginTransaction();
			pm.Serialize(pp1);
			pm.CommitTransaction();

			pm.Clear();
			Person ppp1 = (Person)pm.LoadSingle(typeof(Person), "[id('p1')]");

			Assert.AreEqual("Joe", ppp1.Name);

			ppp1.Name = null;

			// serialize the object PersonProxy again
			pm.BeginTransaction();
			pm.Serialize(ppp1);
			pm.CommitTransaction();

			pm.Clear();
			Person pppp1 = (Person)pm.LoadSingle(typeof(Person), "[id('p1')]");
			Assert.IsNull(pppp1.Name);
		}


		[Test]
		public void EagerLoading()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "John";

			Person p2 = new Person();
			p2.Name = "Eric";
			p2.Age = 24;

			Person p3 = new Person();
			p3.Name = "Mark";
			p3.Age = 24;

			Address adr1 = new Address();
			adr1.City = "Paris";

			p1.Partners.Add(p2);
			p1.Address = adr1;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p3);
			pm.CommitTransaction();

			// The code above should not execute a query

			int c1 = p1.Partners.Count;
			int c2 = p2.Partners.Count;
			object obj1 = p1.Address;
			object obj2 = p2.Address;

			//pm.Clear();

			Query q = new Query(typeof(Person), "[Name = 'John']");
			q.InferredRelationships.Add("Partners");
			q.InferredRelationships.Add("Address");

			p1 = (Person)pm.Load(q)[0];

			// The code above should not execute a query

			Assert.IsNotNull(p1.Partners);
			Assert.AreEqual(1, p1.Partners.Count);
			Assert.AreEqual("Eric", ((Person)p1.Partners[0]).Name);
			Assert.AreEqual(24, ((Person)p1.Partners[0]).Age);
			Assert.AreEqual("Paris", p1.Address.City);

			//------------------------------------------
			// test with empty references without clearing the cache

			q = new Query(typeof(Person), "[Name = 'Eric']");
			q.InferredRelationships.Add("Partners");
			q.InferredRelationships.Add("Address");

			p2 = (Person)pm.Load(q)[0];
			
			// The code above should not execute a query

			Assert.IsNotNull(p2.Partners);
			Assert.AreEqual(0, p2.Partners.Count);
			Assert.IsNull(p2.Address);

			//------------------------------------------
			// test with empty references after clearing the cache
			pm.Clear();

			q = new Query(typeof(Person), "[Name = 'Mark']");
			q.InferredRelationships.Add("Partners");
			q.InferredRelationships.Add("Address");

			p2 = (Person)pm.Load(q)[0];
			
			// The code above should not execute a query

			Assert.IsNotNull(p2.Partners);
			Assert.AreEqual(0, p2.Partners.Count);
			Assert.IsNull(p2.Address);

			//------------------------------------------
			// test with none empty references after clearing the cache
			pm.Clear();

			q = new Query(typeof(Person), "[Name = 'John']");
			q.InferredRelationships.Add("Partners");
			q.InferredRelationships.Add("Address");

			p1 = (Person)pm.Load(q)[0];

			// The code above should not execute a query

			Assert.IsNotNull(p1.Partners);
			Assert.AreEqual(1, p1.Partners.Count);
			Assert.AreEqual("Eric", ((Person)p1.Partners[0]).Name);
			Assert.AreEqual(24, ((Person)p1.Partners[0]).Age);
			Assert.AreEqual("Paris", p1.Address.City);
		}

        [Test]
        public void Import()
        {
            ObjectService ss = new ObjectService("engines.config", "Xml");
            ss.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\PersistenceEngine\Models\Person.xml"));
            ObjectContext sc = ss.CreateObjectContext();

            ObjectService ts = new ObjectService("engines.config", "Sql");
            ts.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\PersistenceEngine\Models\Person.xml"));
            ObjectContext tc = ts.CreateObjectContext();

            sc.InitializeRepository();
            tc.InitializeRepository();

            Person p1 = new Person();
            p1.Name = "John";

            Person p2 = new Person();
            p2.Name = "Eric";
            p2.Age = 24;

            Person p3 = new Person();
            p3.Name = "Mark";
            p3.Age = 24;

            p1.Partners.Add(p2);

            sc.BeginTransaction();
            sc.Serialize(p1);
            sc.Serialize(p3);
            sc.CommitTransaction();

            sc.Clear();
            tc.Clear();

            tc.BeginTransaction();
            foreach (Person p in sc.Load(typeof(Person)))
            {
                tc.Import(p as IPersistable);
                tc.Serialize(p);
            }
            tc.CommitTransaction();
            tc.Clear();

            Assert.AreEqual(3, tc.Load(typeof(Person)).Count);
        }

        [Test]
        public void ImportRepository()
        {
            ObjectService ss = new ObjectService("engines.config", "Xml");
            ss.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\PersistenceEngine\Models\Person.xml"));
            ObjectContext sc = ss.CreateObjectContext();

            ObjectService ts = new ObjectService("engines.config", "Sql");
            ts.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\PersistenceEngine\Models\Person.xml"));
            ObjectContext tc = ts.CreateObjectContext();

            sc.InitializeRepository();
            tc.InitializeRepository();

            Person p1 = new Person();
            p1.Name = "John";

            Person p2 = new Person();
            p2.Name = "Eric";
            p2.Age = 24;

            Person p3 = new Person();
            p3.Name = "Mark";
            p3.Age = 24;

            p1.Partners.Add(p2);

            sc.BeginTransaction();
            sc.Serialize(p1);
            sc.Serialize(p3);
            sc.CommitTransaction();

            sc.Clear();
            tc.Clear();

            ObjectService.ImportRepository(sc, tc, false);

            Assert.AreEqual(3, tc.Load(typeof(Person)).Count);
            Assert.AreEqual(1, tc.Load(typeof(Person), ".Partners").Count);
        }

		[Test]
		public virtual void OPath()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";
			
			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			Person p = new Person();
			p.Name = "p";

			p.Partners.Add(p1);
			p.Partners.Add(p2);
			p.Partners.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			IList pColl = pm.Load(typeof(Person), "Person[Name='p'].Partners[Name ^ 'p']");
			Assert.AreEqual(3, pColl.Count);

			pm.Clear();
			Query q = new Query(typeof(Person), "Person[Name='p'].Partners[Name ^ 'p']");
			pColl = pm.Load(q);
			Assert.AreEqual(3, pColl.Count);
		}

		[Test]
		public virtual void CreateRelationship()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.CreateRelationship(p1, p2, "Partners");
			pm.CommitTransaction();

			pm.Clear();

			p1 = (Person)pm.Load(typeof(Person), "[Name = 'p1']")[0];

			Assert.IsNotNull(p1);
			Assert.AreEqual(1, p1.Partners.Count);
			Assert.AreEqual("p2", ((Person)p1.Partners[0]).Name);
		}

		[Test]
		public virtual void CreateRelationShipAndReload()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			p1 = (Person)pm.LoadSingle(typeof(Person), "[Name = 'p1']");
			Assert.IsNotNull(p1);
			Assert.AreEqual(0, p1.Partners.Count);

			Person p2 = new Person();
			p2.Name = "p2";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.CreateRelationship(p1, p2, "Partners");
			pm.CommitTransaction();

			// Mandatory as unless the original entity would be returned and it doesn't have a reference to the other (just in database)
			pm.Clear();

			p1 = (Person)pm.LoadSingle(typeof(Person), "[Name = 'p1']");

			Assert.IsNotNull(p1);
			Assert.AreEqual(1, p1.Partners.Count);
			Assert.AreEqual("p2", ((Person)p1.Partners[0]).Name);
		}

		[Test]
		public virtual void CreateAndReloadWithId()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.PersonId = "p1";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			// Load without any clear
			p1 = (Person)pm.LoadWithId(typeof(Person), "p1");

			Assert.IsNotNull(p1);
		}

	}
}
