using System;
using System.Collections;

using Microsoft.VisualStudio.TestTools.UnitTesting;
using Evaluant.Uss.ObjectContext;
using Evaluant.Uss;
using Evaluant.Uss.ObjectContext.Descriptors;
using System.IO;
using System.Collections.Generic;
using NUnit.PersistenceManager.DynamicProxy;


namespace NUnit.DynamicProxy
{
	/// <summary>
	/// Description rsume de Proxy.
	/// </summary>
	[TestClass]
	public class Proxy
	{
		public static ObjectService os;

		public Evaluant.Uss.ObjectContext.ObjectContext pm;

		public Proxy()
		{
			Console.WriteLine(Environment.CurrentDirectory);
            provider = XmlConfigLoader.LoadXmlConfig("engines.config", "Prevalence");
			Configure();
		}

		IPersistenceProvider provider;

		public virtual void Configure()
		{
			provider.RegisterMetaData(MetaDataFactory.FromAssembly(typeof(Person).Assembly,new ReflectionDescriptor()));
			provider.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\..\NUnit\PersistenceEngine\Models\Proxy.xml"));

			provider.InitializeConfiguration();

			os = new ObjectService(provider);
			os.PersistentDescriptor.SetIdDescriptor(typeof(Custom1), "MyId", "myId1", typeof(string), false);
			os.PersistentDescriptor.SetIdDescriptor(typeof(Custom2), "MyId", "myId2", typeof(string), false);
			os.PersistentDescriptor.SetIdDescriptor(typeof(Custom3), "MyId", "myId3", typeof(string), false);

			pm = os.CreateObjectContext();

			// Uncomment to test with events. Warning: HugeQuantity is very very long with it
			// pm.EnableEvents = true;
		}

		[TestMethod]
		public void ShouldAllowToSpecifyTheId()
		{
			pm.InitializeRepository();

			Custom1 c1 = new Custom1();

			pm.BeginTransaction();
			pm.Serialize(c1);
			pm.CommitTransaction();

			Custom2 c2 = new Custom2();
			Custom3 c3 = new Custom3();
			Custom3 c4 = new Custom3();

			c1.Customs.Add(c2);
			c2.Customs.Add(c3);

			pm.BeginTransaction();
			pm.Serialize(c1);
			pm.CommitTransaction();

			c1.Custom = c4;

			pm.BeginTransaction();
			pm.Serialize(c1);
			pm.CommitTransaction();

			Assert.AreEqual(1, pm.Load<Custom1>().Count);
			Assert.AreEqual(1, pm.Load<Custom2>().Count);
			Assert.AreEqual(2, pm.Load<Custom3>().Count);
		}

		[TestMethod]
		public virtual void NonStandardValues()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "John";
			p1.Picture = new System.Drawing.Bitmap(200, 100);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Person p = pm.Load<Person>(typeof(Person))[0];

			Assert.IsNotNull(p);
			Assert.IsNotNull(p.Picture);
			Assert.IsTrue(p.Picture.GetType().IsSubclassOf(typeof(System.Drawing.Image)));
			Assert.AreEqual(200, p.Picture.Width);
			Assert.AreEqual(100, p.Picture.Height);
		}

		[TestMethod]
		public void ShouldCreateRestoredRelationships()
		{
			pm.InitializeRepository();

			Company c = new Company();

			Employee p1 = new Employee();
			p1.Name = "p1";

			Employee p2 = new Employee();
			p2.Name = "p2";

			c.Employees.Add(p1);
			c.Employees.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(c);
			pm.CommitTransaction();

			pm.Clear();

			c = pm.LoadSingle<Company>();

			Assert.AreEqual(2, c.Employees.Count);

			Employee e = c.Employees[0];

			c.Employees.Remove(e);

			c.Employees.Add(e);

			pm.BeginTransaction();
			pm.Serialize(c);
			pm.CommitTransaction();

			pm.Clear();

			c = pm.LoadSingle<Company>();

			Assert.AreEqual(2, c.Employees.Count);
		}

		[TestMethod]
		public void NullableTypes()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Age = 12;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			p1 = pm.Load<Person>(typeof(Person))[0];
			Assert.AreEqual(12, p1.Age);

			p1.Age = null;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			p1 = pm.Load<Person>(typeof(Person))[0];
			Assert.AreEqual(null, p1.Age);

		}

		[TestMethod]
		public void ShouldNotLoadUnitOfWork()
		{
			pm.InitializeRepository();

			Employee e1 = new Employee();
			e1.Name = "e1";

			Employee e2 = new Employee();
			e2.Name = "e2";

			pm.BeginTransaction();
			pm.Serialize(e1);
			pm.Serialize(e2);
			pm.CommitTransaction();

			Assert.AreEqual(e1, pm.LoadWithId<Employee>(e1.PersonId));
			Assert.AreEqual(e2, pm.LoadWithId<Employee>(e2.PersonId));

			Person p1 = new Person();

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CreateRelationship(p1, e1, "Employees");
			pm.CreateRelationship(p1, e2, "Employees");
			pm.CommitTransaction();

			Person p2 = pm.LoadWithId<Person>(p1.PersonId);

			Assert.AreEqual(p1, p2);
			Assert.AreEqual(0, p2.Employees.Count);

			pm.Clear();
			p2 = pm.LoadWithId<Person>(p1.PersonId);

			Assert.IsTrue(p1 != p2);
			Assert.AreEqual(2, p2.Employees.Count);

		}

		[TestMethod]
		public virtual void PropertyTypeIsNotFieldType()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";
			p1.Marks = new int[] { 1, 10, 19 };

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.Load<Person>(typeof(Person))[0];

			Assert.IsNotNull(p1);
			Assert.AreEqual(3, p1.Marks.Length);
			Assert.AreEqual(10, p1.Marks[1]);
		}

		[TestMethod]
		public virtual void TestEnumeration()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";
			p1.Kind = KindEnum.Nice;
			p1.Title = TitleEnum.Sir;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Person p = (Person)pm.Load<Person>(typeof(Person), "[Name = 'p1']")[0];

			Assert.IsNotNull(p);
			Assert.AreEqual(TitleEnum.Sir, p.Title);
			Assert.AreEqual(KindEnum.Nice, p.Kind);

			p.Title = TitleEnum.Madam;
			p.Kind = KindEnum.Bad;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = (Person)pm.Load<Person>(typeof(Person), "[Name = 'p1']")[0];

			Assert.IsNotNull(p);
			Assert.AreEqual(TitleEnum.Madam, p.Title);
			Assert.AreEqual(KindEnum.Bad, p.Kind);

			p = (Person)pm.Load<Person>(typeof(Person), string.Format("[Title = '{0}']", Enum.GetName(typeof(TitleEnum), TitleEnum.Madam)))[0];

			Assert.IsNotNull(p);
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual(KindEnum.Bad, p.Kind);

			p = pm.Load<Person>(typeof(Person), string.Format("[Kind = {0}]", (int)p.Kind))[0];

			Assert.IsNotNull(p);
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual(TitleEnum.Madam, p.Title);
		}

		[TestMethod]
		public void Id()
		{
			pm.InitializeRepository();

			Address a = new Address();
			a.City = "New York";
			a.Id = "Address01";

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;
			p.Address = a;
			p.PersonId = "Person01";

			Person partner = new Person();
			partner.Name = "Joey Triviani";
			partner.Age = 32;
			partner.PersonId = "Person02";

			p.Partners.Add(partner);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = pm.Load<Person>(typeof(Person), "[Name = 'John Smith']")[0] as Person;

			Assert.IsNotNull(p);
			Assert.AreEqual("Person01", p.PersonId);
			Assert.AreEqual("Address01", p.Address.Id);
			Assert.AreEqual("Person02", ((Person)p.Partners[0]).PersonId);
		}

		[TestMethod]
		public void LoadParents()
		{
			pm.InitializeRepository();

			Address a = new Address();
			a.City = "New York";
			a.Id = "Address01";

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;
			p.Address = a;
			p.PersonId = "Person01";

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			IList<Person> ps = pm.LoadParents<Person>(a, "Address");

			Assert.AreEqual(1, ps.Count);
			Assert.AreEqual(ps[0], p);
		}

		[TestMethod]
		public void ListGeneric()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Employee emp = new Employee();
			emp.Name = "Joey Triviani";
			emp.Age = 32;

			p.Employees.Add(emp);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);

			p = pm.Load<Person>(typeof(Person), "[Name = 'John Smith']")[0];
			Assert.IsNotNull(p);

			Assert.AreEqual("John Smith", p.Name);
			Assert.AreEqual(50, p.Age);

			Assert.AreEqual(1, p.Employees.Count);
			Assert.AreEqual("Joey Triviani", (p.Employees[0]).Name);
		}

		[TestMethod]
		public void Overral()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Person partner = new Person();
			partner.Name = "Joey Triviani";
			partner.Age = 32;

			p.Partners.Add(partner);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);

			p = pm.Load<Person>(typeof(Person), "[Name = 'John Smith']")[0];
			Assert.IsNotNull(p);

			Assert.AreEqual("John Smith", p.Name);
			Assert.AreEqual(50, p.Age);

			Assert.AreEqual(1, p.Partners.Count);
			Assert.AreEqual("Joey Triviani", ((Person)p.Partners[0]).Name);

			p.Age = 49;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[Name = 'John Smith']").Count);
			Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[Age = 49]").Count);
		}

		[TestMethod]
		public void ShouldHandleAbsractBaseClassWithNoProperty()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load<HumanBeing>(typeof(HumanBeing)).Count);
		}

		[TestMethod]
		public void OneToOneSerialize()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Address a = new Address("New York");
			p.Address = a;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load<Person>(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load<Address>(typeof(Address)).Count);
			Assert.AreEqual("New York", (pm.Load<Address>(typeof(Address))[0]).City);
		}

		[TestMethod]
		public void OneToOneLazyLoad()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;

			Address a = new Address("New York");
			p.Address = a;

			Country c = new Country("United States");
			a.Country = c;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = pm.LoadSingle<Person>(typeof(Person));

			Assert.IsFalse(p.BaseAddressGetterCalled);
			Assert.IsNotNull(p.Address);
			Assert.IsTrue(p.BaseAddressGetterCalled);
			Assert.AreEqual("New York", p.Address.City);

			Assert.IsFalse(p.Address.BaseCountryGetterCalled);
			Assert.IsNotNull(p.Address.Country);
			Assert.IsTrue(p.Address.BaseCountryGetterCalled);
			Assert.AreEqual("United States", p.Address.Country.Name);
		}

		[TestMethod]
		public void Update()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";
			p.Age = 50;
			p.Birth = new DateTime(2005, 12, 28);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = pm.LoadSingle<Person>(typeof(Person));

			Assert.AreEqual("John Smith", p.Name);
			Assert.AreEqual(50, p.Age);
			Assert.AreEqual(new DateTime(2005, 12, 28), p.Birth);

			p.Name = "Doug Ross";
			p.Age = 40;
			p.Birth = new DateTime(2005, 12, 29);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load<Person>(typeof(Person)).Count);
			p = pm.LoadSingle<Person>(typeof(Person));
			Assert.AreEqual("Doug Ross", p.Name);
			Assert.AreEqual(40, p.Age);
			Assert.AreEqual(new DateTime(2005, 12, 29), p.Birth);
		}

		[TestMethod]
		public void ShouldSerializeAddedReferences()
		{
			pm.InitializeRepository();

			Person t = new Person();
			t.Age = 10;

			pm.BeginTransaction();
			pm.Serialize(t);
			pm.CommitTransaction();

			pm.Clear();

			t = pm.LoadSingle<Person>(typeof(Person));
			Person c = new Person();
			c.Age = 11;

			t.Partners.Add(c);

			pm.BeginTransaction();
			pm.Serialize(t);
			pm.CommitTransaction();
		}

		[TestMethod]
		public void IdentityMap()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "John Smith";

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Person p1 = pm.LoadSingle<Person>(typeof(Person));
			Person p2 = pm.LoadSingle<Person>(typeof(Person));

			Assert.IsNotNull(p1);
			Assert.IsNotNull(p2);
			Assert.IsTrue(p1 == p2);
		}

		[TestMethod]
		public void RemoveRelationship()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			p1.Partners.Add(p2);
			p1.Partners.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(3, pm.Load<Person>(typeof(Person)).Count);

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");

			Assert.AreEqual(2, p1.Partners.Count);

			p1.Partners.Clear();

			Assert.AreEqual(0, p1.Partners.Count);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(3, pm.Load<Person>(typeof(Person)).Count);
			Assert.AreEqual(0, pm.Load<Person>(typeof(Person), "[Name = 'p1'].Partners").Count);

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");

			Assert.AreEqual(0, p1.Partners.Count);
		}

		[TestMethod]
		public void CascadeUpdate()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Address a1 = new Address();
			a1.City = "a1";

			Person p2 = new Person();
			p2.Name = "p2";

			p1.Partners.Add(p2);
			p1.Address = a1;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");
			p2 = ((Person)p1.Partners[0]);
			a1 = p1.Address;
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual("p2", p2.Name);
			Assert.AreEqual("a1", a1.City);

			p2.Name = "p2 changed";
			a1.City = "a1 changed";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");
			p2 = ((Person)p1.Partners[0]);
			a1 = p1.Address;
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual("p2 changed", p2.Name);
			Assert.AreEqual("a1 changed", a1.City);

		}

		[TestMethod]
		public void ManyToOne()
		{

			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			Address a = new Address("Paris");

			p1.Address = a;
			p2.Address = a;
			p3.Address = a;

			pm.BeginTransaction();
			pm.Serialize(new object[] { p1, p2, p3 });
			pm.CommitTransaction();

			Assert.AreEqual(1, pm.Load<Address>(typeof(Address)).Count);
		}

		[TestMethod]
		public void Inheritance()
		{
			pm.InitializeRepository();

			Employee bill = new Employee();
			bill.Name = "Bill Gates";
			bill.Rank = 1;

			Address redmond = new Address();
			redmond.City = "Redmond";
			bill.Address = redmond;

			Address paloalto = new Address();
			paloalto.City = "Palo Alto";

			Person steve = new Person();
			steve.Name = "Steve Jobs";
			steve.Address = paloalto;

			bill.Partners.Add(steve);

			pm.BeginTransaction();
			pm.Serialize(bill);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);
			Assert.AreEqual(2, pm.Load<Address>(typeof(Address)).Count);

			bill = pm.LoadSingle<Employee>(typeof(Employee), "[Name = 'Bill Gates']");
			Assert.IsNotNull(bill);
			Assert.AreEqual(1, bill.Rank);
			Assert.AreEqual("Steve Jobs", ((Person)bill.Partners[0]).Name);
			Assert.AreEqual("Redmond", bill.Address.City);

		}

		[TestMethod]
		public void Id2()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.PersonId = "p1";
			p1.Name = "Mike Tyson";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.IsNotNull(pm.LoadWithId<Person>("p1"));

			pm.InitializeRepository();

			Employee e1 = new Employee();
			e1.PersonId = "e1";
			e1.Name = "Peter Pan";

			pm.BeginTransaction();
			pm.Serialize(e1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.IsNotNull(pm.LoadWithId<Person>("e1"));
			Assert.IsNotNull(pm.LoadWithId<Employee>("e1"));
		}


		[TestMethod]
		public void Id3()
		{
			// Id of type integer

			pm.InitializeRepository();

			Task t1 = new Task();
			t1.Id = 1;
			t1.Libelle = "Feed the dog";

			pm.BeginTransaction();
			pm.Serialize(t1);
			pm.CommitTransaction();

			pm.Clear();

			t1 = pm.LoadWithId<Task>(t1.Id.ToString());
			Assert.IsNotNull(t1);
			Assert.AreEqual(1, t1.Id);
		}

		public void ExistingObjects()
		{
			//	Person and Country exists in database
			//	Creates a new Address and reserialize the Person
			//	Add the Country to the new Address
			//	Add the Address to the new Person

			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "Steve Jobs";
			p.Age = 49;

			Country c = new Country();
			c.Name = "United States";

			pm.BeginTransaction();
			pm.Serialize(new object[] { p, c });
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load<Person>(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load<Country>(typeof(Country)).Count);

			p = pm.LoadSingle<Person>(typeof(Person));
			c = pm.LoadSingle<Country>(typeof(Country));

			Address a = new Address();
			a.City = "New York";

			a.Country = c;

			p.Address = a;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			p = pm.LoadSingle<Person>(typeof(Person));

			Assert.IsNotNull(p);
			Assert.IsNotNull(p.Address);
			Assert.IsNotNull(p.Address.Country);
		}

		[TestMethod]
		public void ExistingObject2()
		{
			//	Address exists in database
			//	Creates a Person and a Country
			//	Add the Address to the new Person
			//	Add the Country to the Address
			//	Serialize the Person

			pm.InitializeRepository();

			Address a = new Address();
			a.City = "New York";

			pm.BeginTransaction();
			pm.Serialize(a);
			pm.CommitTransaction();

			pm.Clear();

			a = pm.LoadSingle<Address>(typeof(Address));

			Person p = new Person();
			p.Name = "Steve Jobs";

			Country c = new Country();
			c.Name = "United States";

			a.Country = c;

			p.Address = a;

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.IsNotNull(p);
			Assert.IsNotNull(p.Address);
			Assert.IsNotNull(p.Address.Country);
		}

		[TestMethod]
		public void ExistingObjects3()
		{
			//	Person p1 p2 and p4 exist in database
			//	p2 is a partner of p1
			//	Load p1 and creates a new Person p3 and add it to p1 as partner
			//	Load p1 and add p4 as partners of p1 at least

			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "Bill Gates";
			p1.Title = TitleEnum.Sir;

			Person p2 = new Person();
			p2.Name = "Steve Jobs";
			p2.Title = TitleEnum.Madam;

			Person p4 = new Person();
			p4.Name = "John Smith";

			p1.Partners.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p4);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(3, pm.Load<Person>(typeof(Person)).Count);

			pm.Clear();

			File.Create("begin").Close();

			p1 = pm.Load<Person>(typeof(Person), "[Name = 'Bill Gates']")[0];

			Assert.IsNotNull(p1);
			Assert.AreEqual(1, p1.Partners.Count);

			Person p3 = new Person();
			p3.Name = "Joey Triviani";

			p1.Partners.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(4, pm.Load<Person>(typeof(Person)).Count);

			p1 = pm.Load<Person>(typeof(Person), "[Name = 'Bill Gates']")[0];

			Assert.IsNotNull(p1);
			Assert.AreEqual(2, p1.Partners.Count);

			pm.Clear();

			p1 = pm.Load<Person>(typeof(Person), "[Name = 'Bill Gates']")[0];
			p4 = pm.Load<Person>(typeof(Person), "[Name = 'John Smith']")[0];

			Assert.IsNotNull(p1);
			Assert.IsNotNull(p4);

			p1.Partners.Add(p4);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(4, pm.Load<Person>(typeof(Person)).Count);
			Assert.AreEqual(3, pm.Load<Person>(typeof(Person), "[Name = 'Bill Gates'].Partners").Count);
		}

		[TestMethod]
		public virtual void Enumeration()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "Elton";
			p1.Title = TitleEnum.Sir;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			System.Collections.Generic.List<Person> coll = pm.Load<Person>(typeof(Person), "[Title = 'Sir']");
			Assert.AreEqual(1, coll.Count);

			coll = pm.Load<Person>(typeof(Person), "[Name = 'Elton']");
			Assert.AreEqual("Elton", ((Person)coll[0]).Name);
		}

		[TestMethod]
		public void Detach()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			p1.Partners.Add(p2);
			p1.Partners.Add(p3);

			p3.Partners.Add(p1);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");
			Assert.AreEqual(2, p1.Partners.Count); // Lazyload partners for p1

			p1 = pm.Detach(p1) as Person;

			Assert.AreEqual(typeof(Person), p1.GetType());
			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual(2, p1.Partners.Count); // Check lazyloaded items are there
			Assert.AreEqual(0, ((Person)p1.Partners[0]).Partners.Count + ((Person)p1.Partners[1]).Partners.Count); // Check none lazyloaded items are not there
		}


		[TestMethod]
		public virtual void LoadWithQuerySort()
		{
			// one field
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.PersonId = "1";
			p1.Name = "p01";
			p1.Birth = DateTime.Today;

			Person p2 = new Person();
			p2.PersonId = "3";
			p2.Name = "p02";
			p2.Birth = DateTime.Today.AddDays(1);

			Person p3 = new Person();
			p3.PersonId = "2";
			p3.Name = "p02";
			p3.Birth = DateTime.Today;

			Person p4 = new Person();
			p4.PersonId = "4";
			p4.Name = "p04";
			p4.Birth = DateTime.Today.AddDays(1);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.Serialize(p3);
			pm.Serialize(p4);
			pm.CommitTransaction();

			Query query = new Query(typeof(Person));
			query.OrderBy.Add("Name");
			System.Collections.Generic.List<Person> results = pm.Load<Person>(query);

			Assert.AreEqual("p01", ((Person)results[0]).Name);
			Assert.AreEqual("p02", ((Person)results[1]).Name);
			Assert.AreEqual("p02", ((Person)results[2]).Name);
			Assert.AreEqual("p04", ((Person)results[3]).Name);

			query = new Query(typeof(Person));
			query.OrderBy.Add("Name desc");
			results = pm.Load<Person>(query);

			Assert.AreEqual("p04", ((Person)results[0]).Name);
			Assert.AreEqual("p02", ((Person)results[1]).Name);
			Assert.AreEqual("p02", ((Person)results[2]).Name);
			Assert.AreEqual("p01", ((Person)results[3]).Name);


			// two fields
			query = new Query(typeof(Person));
			query.OrderBy.Add("Name");
			query.OrderBy.Add("Birth");
			results = pm.Load<Person>(query);

			Assert.AreEqual("1", ((Person)results[0]).PersonId);
			Assert.AreEqual("2", ((Person)results[1]).PersonId);
			Assert.AreEqual("3", ((Person)results[2]).PersonId);
			Assert.AreEqual("4", ((Person)results[3]).PersonId);

			query = new Query(typeof(Person));
			query.OrderBy.Add("Name desc");
			query.OrderBy.Add("Birth desc");
			results = pm.Load<Person>(query);

			Assert.AreEqual("4", ((Person)results[0]).PersonId);
			Assert.AreEqual("3", ((Person)results[1]).PersonId);
			Assert.AreEqual("2", ((Person)results[2]).PersonId);
			Assert.AreEqual("1", ((Person)results[3]).PersonId);
		}

		[TestMethod]
		public virtual void LoadWithQueryMinMax()
		{
			pm.InitializeRepository();

			System.Collections.Generic.List<Person> personColl = new System.Collections.Generic.List<Person>();

			for (int i = 1; i <= 50; i++)
			{
				Person p = new Person();
				p.Name = "p" + i.ToString("00");
				personColl.Add(p);

				pm.BeginTransaction();
				pm.Serialize(p);
				pm.CommitTransaction();
			}

			//pm.BeginTransaction();
			//pm.Serialize(personColl);
			//pm.CommitTransaction();

			Query query = new Query(typeof(Person));
			query.OrderBy.Add("Name");
			query.FirstResult = 1;
			query.MaxResults = 100;
			System.Collections.Generic.List<Person> results = pm.Load<Person>(query);

			Assert.AreEqual(50, results.Count);

			query.FirstResult = 10;
			query.MaxResults = 21;
			results = pm.Load<Person>(query);

			Assert.AreEqual(21, results.Count);
			Assert.AreEqual("p10", ((Person)results[0]).Name);
			Assert.AreEqual("p20", ((Person)results[10]).Name);
			Assert.AreEqual("p30", ((Person)results[20]).Name);

			query.FirstResult = 10;
			query.MaxResults = 0;
			results = pm.Load<Person>(query);

			Assert.AreEqual(41, results.Count);
			Assert.AreEqual("p10", ((Person)results[0]).Name);
			Assert.AreEqual("p20", ((Person)results[10]).Name);
			Assert.AreEqual("p50", ((Person)results[40]).Name);

			query = new Query(typeof(Person), "[Name ^ 'p']");
			query.OrderBy.Add("Name");
			query.FirstResult = 1;
			query.MaxResults = 10;
			results = pm.Load<Person>(query);

			Assert.AreEqual(10, results.Count);
		}

		[TestMethod]
		public virtual void NullValueAttribute()
		{
			pm.InitializeRepository();

			// serialize an object Person with an attribute which has a value null
			Person p1 = new Person();
			p1.PersonId = "p1";
			p1.Name = null;
			p1.Birth = DateTime.Now;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			// load the object PersonProxy
			pm.Clear();
			Person pp1 = pm.LoadSingle<Person>(typeof(Person), "[id('p1')]");

			Assert.IsNull(pp1.Name);

			pp1.Age = 30;

			// serialize the object PersonProxy   (CreateAttributCommand with null value)
			pm.BeginTransaction();
			pm.Serialize(pp1);
			pm.CommitTransaction();

			Assert.IsNull(pp1.Name);
			pp1.Name = "Joe";

			// serialize the object PersonProxy again (UpdateAttributeCommand for an non existing attribute in repository)
			pm.BeginTransaction();
			pm.Serialize(pp1);
			pm.CommitTransaction();

			pm.Clear();
			Person ppp1 = pm.LoadSingle<Person>(typeof(Person), "[id('p1')]");

			Assert.AreEqual("Joe", ppp1.Name);

			ppp1.Name = null;

			// serialize the object PersonProxy again
			pm.BeginTransaction();
			pm.Serialize(ppp1);
			pm.CommitTransaction();

			pm.Clear();
			Person pppp1 = pm.LoadSingle<Person>(typeof(Person), "[id('p1')]");
			Assert.IsNull(pppp1.Name);
		}

		[TestMethod]
		public void ShouldEagerLoadWithInheritance()
		{
			pm.InitializeRepository();

			Person p1 = new Employee();
			p1.Name = "p1";

			Person p2 = new Consultant();
			p2.Name = "p2";

			Address a1 = new Address();
			a1.City = "Paris";

			Person p3 = new Person();
			p3.Name = "p3";

			p1.Partners.Add(p3);
			p2.Partners.Add(p3);

			p1.Address = a1;
			p2.Address = a1;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.CommitTransaction();

			pm.Clear();

			Query q = new Query(typeof(Person));
			q.InferredRelationships.Add("Partners");
			q.InferredRelationships.Add("Address");

			List<Person> ps = pm.Load<Person>(q);

			// To check the objects don't need the engine to get the relationships
			IPersistenceEngine engine = pm.PersistenceEngine;
			pm.PersistenceEngine = null;
			try
			{

				// The code above should not execute a query
				foreach (Person p in ps)
				{
					if (p.Partners != null && p.Address != null)
						;
				}
			}
			finally
			{
				pm.PersistenceEngine = engine;
			}
		}

		[TestMethod]
		public void EagerLoading()
		{
			pm.InitializeRepository();

			Person p1 = new Employee();
			p1.Name = "John";

			Person p2 = new Person();
			p2.Name = "Eric";
			p2.Age = 24;

			Person p3 = new Person();
			p3.Name = "Mark";
			p3.Age = 24;

			Address adr1 = new Address();
			adr1.City = "Paris";

			p1.Partners.Add(p2);
			p1.Address = adr1;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p3);
			pm.CommitTransaction();


			int c1 = p1.Partners.Count;
			int c2 = p2.Partners.Count;
			object obj1 = p1.Address;
			object obj2 = p2.Address;

			//pm.Clear();

			Query q = new Query(typeof(Person), "[Name = 'John']");
			q.InferredRelationships.Add("Partners");
			q.InferredRelationships.Add("Address");

			p1 = pm.Load<Person>(q)[0];

			System.Diagnostics.Trace.WriteLine("Start Eager ... no request trace should appear");
			// The code above should not execute a query

			Assert.IsNotNull(p1.Partners);
			Assert.AreEqual(1, p1.Partners.Count);
			Assert.AreEqual("Eric", ((Person)p1.Partners[0]).Name);
			Assert.AreEqual(24, ((Person)p1.Partners[0]).Age);
			Assert.AreEqual("Paris", p1.Address.City);

			System.Diagnostics.Trace.WriteLine("Stop Eager");

			//------------------------------------------
			// test with empty references without clearing the cache

			q = new Query(typeof(Person), "[Name = 'Eric']");
			q.InferredRelationships.Add("Partners");
			q.InferredRelationships.Add("Address");

			p2 = pm.Load<Person>(q)[0];

			// The code above should not execute a query

			Assert.IsNotNull(p2.Partners);
			Assert.AreEqual(0, p2.Partners.Count);
			Assert.IsNull(p2.Address);

			//------------------------------------------
			// test with empty references after clearing the cache
			pm.Clear();

			q = new Query(typeof(Person), "[Name = 'Mark']");
			q.InferredRelationships.Add("Partners");
			q.InferredRelationships.Add("Address");

			p2 = pm.Load<Person>(q)[0];

			// The code above should not execute a query

			Assert.IsNotNull(p2.Partners);
			Assert.AreEqual(0, p2.Partners.Count);
			Assert.IsNull(p2.Address);

			//------------------------------------------
			// test with none empty references after clearing the cache
			pm.Clear();

			q = new Query(typeof(Person), "[Name = 'John']");
			q.InferredRelationships.Add("Partners");
			q.InferredRelationships.Add("Address");

			p1 = pm.Load<Person>(q)[0];

			// The code above should not execute a query

			Assert.IsNotNull(p1.Partners);
			Assert.AreEqual(1, p1.Partners.Count);
			Assert.AreEqual("Eric", ((Person)p1.Partners[0]).Name);
			Assert.AreEqual(24, ((Person)p1.Partners[0]).Age);
			Assert.AreEqual("Paris", p1.Address.City);
		}

		[TestMethod]
		public void SerializeOrder()
		{
			pm.InitializeRepository();
			pm.BeginTransaction();

			Company c = new Company();
			c.Name = "Evaluant";

			pm.Serialize(c);

			for (int i = 0; i < 10; i++)
			{
				Employee e = new Employee();
				e.Name = "John";

				c.Employees.Add(e);
			}

			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual("Evaluant", pm.LoadSingle<Company>(typeof(Company)).Name);
			Assert.AreEqual(10, pm.Load<Person>(typeof(Company), ".Employees").Count);

		}

		[TestMethod]
		public void ReloadEntity()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "John";

			Person p2 = new Person();
			p2.Name = "Eric";
			p2.Age = 24;

			Person p3 = new Person();
			p3.Name = "Mark";
			p3.Age = 24;

			p1.Partners.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'John']");
			p2 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'Eric']");
			Assert.IsNotNull(p2);

			p2.Age = 10;

			// Loads p2 not directly
			Assert.AreEqual(1, p1.Partners.Count);

			// The lazy loading should not have refreshed the properties of p2
			Assert.AreEqual(10, p2.Age);
		}

		[TestMethod]
		public void Import()
		{
			ObjectService ss = new ObjectService("engines.config", "Xml");
			ss.RegisterMetaData(MetaDataFactory.FromAssembly(typeof(Person).Assembly));
			ss.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\..\NUnit\PersistenceEngine\Models\Proxy.xml"));
			ObjectContext sc = ss.CreateObjectContext();

			ObjectService ts = new ObjectService("engines.config", "Sql");
			ts.RegisterMetaData(MetaDataFactory.FromAssembly(typeof(Person).Assembly));
			ts.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\..\NUnit\PersistenceEngine\Models\Proxy.xml"));
			ObjectContext tc = ts.CreateObjectContext();

			sc.InitializeRepository();
			tc.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "John";

			Person p2 = new Person();
			p2.Name = "Eric";
			p2.Age = 24;

			Person p3 = new Person();
			p3.Name = "Mark";
			p3.Age = 24;

			p1.Partners.Add(p2);

			sc.BeginTransaction();
			sc.Serialize(p1);
			sc.Serialize(p3);
			sc.CommitTransaction();

			sc.Clear();
			tc.Clear();

			tc.BeginTransaction();
			foreach (Person p in sc.Load<Person>(typeof(Person)))
			{
				tc.Import(p);
				tc.Serialize(p);
			}
			tc.CommitTransaction();
			tc.Clear();

			Assert.AreEqual(3, tc.Load<Person>(typeof(Person)).Count);
		}

		[TestMethod]
		public void ImportRepository()
		{
			ObjectService ss = new ObjectService("engines.config", "Xml");
			ss.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\..\NUnit\PersistenceEngine\Models\Proxy.xml"));
			ObjectContext sc = ss.CreateObjectContext();

			ObjectService ts = new ObjectService("engines.config", "Sql");
			ts.RegisterMetaData(MetaDataFactory.FromMetaDataFile(@"..\..\..\NUnit\PersistenceEngine\Models\Proxy.xml"));
			ObjectContext tc = ts.CreateObjectContext();

			sc.InitializeRepository();
			tc.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "John";

			Person p2 = new Person();
			p2.Name = "Eric";
			p2.Age = 24;

			Employee p3 = new Employee();
			p3.Name = "Mark";
			p3.Age = 24;
			p3.Rank = 10;

			p1.Partners.Add(p2);

			sc.BeginTransaction();
			sc.Serialize(p1);
			sc.Serialize(p3);
			sc.CommitTransaction();

			sc.Clear();
			tc.Clear();

			ObjectService.ImportRepository(sc, tc, true);

			Assert.AreEqual(3, tc.Load<Person>(typeof(Person)).Count);
			Assert.AreEqual(1, tc.Load<Person>(typeof(Person), ".Partners").Count);
			Assert.AreEqual(10, tc.LoadSingle<Employee>(typeof(Employee)).Rank);
		}

		/// <summary>
		/// Using dynamic proxies, newly serialized entities should appear in the IdentityMap.
		/// </summary>
		[TestMethod]
		public void LoadAndSave()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.PersonId = "John";
			p1.Name = "John";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Person p2 = pm.LoadWithId<Person>("John");

			Assert.IsNotNull(p1);
			Assert.IsTrue(p1 == p2);

			p2.Name = "John2";

			p1 = pm.LoadWithId<Person>("John");
			Assert.IsTrue(p1 == p2);
		}

		[TestMethod]
		public void NewlyCreatedEntitiesShouldNotBeSerialzedTwoTimes()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.PersonId = "p1";
			p1.Name = "Titi";

			Person p2 = new Person();
			p2.PersonId = "p2";

			p1.Partners.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);

			Person p3 = new Person();
			p3.PersonId = "p3";

			p1.Partners.Add(p3);

			p1.Name = "Toto";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			// Assert.IsNotNull(pm.LoadWithId<Person>("p1"));
			Assert.AreEqual("Toto", pm.LoadWithId<Person>("p1").Name);
			Assert.IsNotNull(pm.LoadWithId<Person>("p2"));
			Assert.IsNotNull(pm.LoadWithId<Person>("p3"));

		}

		[TestMethod]
		public void HugeQuantity()
		{
			pm.InitializeRepository();

			DateTime dt = DateTime.Now;

			pm.BeginTransaction();

			int length = 100000;
			for (int i = 0; i < length; i++)
			{
				Person p = new Person();
				p.Name = "Name";
				p.Birth = dt;
				// p.Age = i; // Do not make instances different to check Hashcode

				pm.Serialize(p);
			}

			pm.CommitTransaction();
		}

		[TestMethod]
		public void AssemblyMetaData()
		{
			ObjectService factory = new ObjectService("engines.config", "Sql");
			factory.RegisterMetaData(MetaDataFactory.FromAssembly(typeof(Person).Assembly));
			ObjectContext oc = factory.CreateObjectContext();

			Evaluant.Uss.Models.Model model = oc.PersistenceEngine.Model;
			Evaluant.Uss.Models.Entity employee = model.GetEntity("NUnit:DynamicProxy:Employee");
			Assert.IsNotNull(employee);
			Assert.AreEqual("NUnit:DynamicProxy:Person", employee.Inherit);
			Assert.AreEqual(typeof(float), model.GetAttribute("NUnit:DynamicProxy:Employee", "Salary").Type);

			Assert.AreEqual("NUnit:DynamicProxy:Address", model.GetReference("NUnit:DynamicProxy:Person", "Address").ChildType);
			Assert.AreEqual("NUnit:DynamicProxy:Person", model.GetReference("NUnit:DynamicProxy:Person", "Address").ParentType);
			Assert.AreEqual(true, model.GetReference("NUnit:DynamicProxy:Person", "Address").FromMany);
			Assert.AreEqual(false, model.GetReference("NUnit:DynamicProxy:Person", "Address").ToMany);
			Assert.AreEqual(false, model.GetReference("NUnit:DynamicProxy:Person", "Address").IsComposition);

			Assert.AreEqual("NUnit:DynamicProxy:Employee", model.GetReference("NUnit:DynamicProxy:Person", "Employees").ChildType);
			Assert.AreEqual("NUnit:DynamicProxy:Person", model.GetReference("NUnit:DynamicProxy:Person", "Employees").ParentType);
			Assert.AreEqual(true, model.GetReference("NUnit:DynamicProxy:Person", "Employees").FromMany);
			Assert.AreEqual(true, model.GetReference("NUnit:DynamicProxy:Person", "Employees").ToMany);
			Assert.AreEqual(false, model.GetReference("NUnit:DynamicProxy:Person", "Employees").IsComposition);

			Assert.AreEqual("NUnit:DynamicProxy:Employee", model.GetReference("NUnit:DynamicProxy:Company", "Employees").ChildType);
			Assert.AreEqual("NUnit:DynamicProxy:Company", model.GetReference("NUnit:DynamicProxy:Company", "Employees").ParentType);
			Assert.AreEqual(true, model.GetReference("NUnit:DynamicProxy:Company", "Employees").FromMany);
			Assert.AreEqual(true, model.GetReference("NUnit:DynamicProxy:Company", "Employees").ToMany);
			Assert.AreEqual(false, model.GetReference("NUnit:DynamicProxy:Company", "Employees").IsComposition);

		}

		[TestMethod]
		public virtual void OPath()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			Person p = new Person();
			p.Name = "p";

			p.Partners.Add(p1);
			p.Partners.Add(p2);
			p.Partners.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			System.Collections.Generic.List<Person> pColl = pm.Load<Person>(typeof(Person), "Person[Name='p'].Partners[Name ^ 'p']");
			Assert.AreEqual(3, pColl.Count);

			pm.Clear();
			Query q = new Query(typeof(Person), "Person[Name='p'].Partners[Name ^ 'p']");
			pColl = pm.Load<Person>(q);
			Assert.AreEqual(3, pColl.Count);
		}

		[TestMethod]
		public void CachedQuery()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			p1.Partners.Add(p3);
			p2.Partners.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.CommitTransaction();

			string id1 = p1.PersonId;
			string id2 = p2.PersonId;
			string id3 = p3.PersonId;

			p1 = null;
			p2 = null;
			p3 = null;

			GC.Collect();

			for (int i = 0; i < 100; i++)
			{
				// Assert.AreEqual(1, pm.LoadScalar("count(NUnit:DynamicProxy:Person[Name = 'p1'].Partners)"));
			}

			DateTime begin = DateTime.Now;
			while (DateTime.Now - begin < new TimeSpan(0, 0, 5))
			{
				p1 = pm.LoadWithId<Person>(id1);
				p2 = pm.LoadWithId<Person>(id2);
				p3 = pm.LoadWithId<Person>(id3);

				Assert.AreEqual("p1", p1.Name);
				Assert.AreEqual("p2", p2.Name);
				Assert.AreEqual("p3", p3.Name);

				Assert.AreEqual(1, p1.Partners.Count);
				Assert.AreEqual(1, p2.Partners.Count);
				Assert.AreEqual(0, p3.Partners.Count);

				Assert.AreEqual("p3", ((Person)p1.Partners[0]).Name);
				Assert.AreEqual("p3", ((Person)p2.Partners[0]).Name);

				p1.Name = "xxx"; p1.Name = "p1";
				p2.Name = "xxx"; p2.Name = "p2";
				p3.Name = "xxx"; p3.Name = "p3";

				Assert.AreEqual("p1", p1.Name);
				Assert.AreEqual("p2", p2.Name);
				Assert.AreEqual("p3", p3.Name);

				pm.BeginTransaction();
				pm.Serialize(p1);
				pm.Serialize(p2);
				pm.Serialize(p3);
				pm.CommitTransaction();
			}
		}

		[TestMethod]
		public virtual void CreateRelationship()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.CreateRelationship(p1, p2, "Partners");
			pm.CommitTransaction();

			pm.Clear();

			p1 = (Person)pm.Load<Person>(typeof(Person), "[Name = 'p1']")[0];

			Assert.IsNotNull(p1);
			Assert.AreEqual(1, p1.Partners.Count);
			Assert.AreEqual("p2", ((Person)p1.Partners[0]).Name);

			pm.InitializeRepository();

			pm.Clear();

			p1 = new Person();
			p1.Name = "p1";

			p2 = new Person();
			p2.Name = "p2";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.CommitTransaction();

			pm.BeginTransaction();
			pm.CreateRelationship(p1, p2, "Partners");
			pm.CommitTransaction();

			Assert.AreEqual(p2, pm.Load<Person>(typeof(Person), "[Name = 'p1'].Partners")[0]);

			pm.InitializeRepository();

			pm.Clear();

			p2 = new Person();
			p2.Name = "p2";

			pm.BeginTransaction();
			pm.Serialize(p2);
			pm.CommitTransaction();

			pm.Clear();

			p1 = new Person();
			p1.Name = "p1";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			p2 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p2']");

			pm.BeginTransaction();
			pm.CreateRelationship(p1, p2, "Partners");
			pm.CommitTransaction();

			Assert.AreEqual(p2, pm.Load<Person>(typeof(Person), "[Name = 'p1'].Partners")[0]);

			p2.Name = "p3";

			pm.BeginTransaction();
			pm.Serialize(p2);
			pm.CommitTransaction();

			Assert.AreEqual(p2, pm.Load<Person>(typeof(Person), "[Name = 'p3']")[0]);

		}

		[TestMethod]
		public void ShouldSerializeNewToOne()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");

			Address a = new Address();
			a.City = "Paris";
			a.Street = "Champs Elyses";
			a.Country = new Country("France");

			p1.Address = a;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load<Address>(typeof(Address), "[City = 'Paris' and Street = 'Champs Elyses' and exists(Country[ Name = 'France']) ]").Count);

		}

		[TestMethod]
		public void ShouldSerializeReferenceToNewInstance()
		{
			ObjectService os = pm.Factory;
			pm = os.CreateObjectContext();
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "p1";
			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			p = pm.LoadSingle<Person>("[Name='p1']");

			p.Address = new Address();

			pm = os.CreateObjectContext();

			p = pm.LoadSingle<Person>("[Name='p1']");

			p.Address = new Address();
			Assert.IsNull(p.Address.City);

			p.Address.City = "Mulhouse";

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm = os.CreateObjectContext();

			p = pm.LoadSingle<Person>("[Name='p1']");

			Assert.IsNotNull(p.Address.City);
		}

		[TestMethod]
		public void ShouldDeleteNonProxyIfGloballyProcessed()
		{
			pm.InitializeRepository();

			Person p = new Person();

			p.Name = "p1";
			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			p = pm.LoadWithId<Person>(p.PersonId);

			try
			{
				pm.BeginTransaction();
				pm.Delete(p);
				pm.CommitTransaction();
			}
			catch (NullReferenceException ex)
			{
				Assert.Fail("Object is not globally processed");
			}
		}

		[TestMethod]
		public void ShouldDeleteNonProxy()
		{
			pm.InitializeRepository();

			Person p = new Person();

			p.Name = "p1";
			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.BeginTransaction();
			pm.Delete(p);
			pm.CommitTransaction();

			p = pm.LoadWithId<Person>(p.PersonId);

			Assert.IsNull(p);
		}

		[TestMethod]
		public void ShouldDeleteNonProxyChildren()
		{
			pm.InitializeRepository();

			Person p = new Person();
			Person q = new Person();
			Employee e = new Employee();

			p.BestFriend = q;
			p.Employees.Add(e);

			p.Name = "p1";
			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.BeginTransaction();
			pm.Delete(q);
			pm.Delete(e);
			pm.CommitTransaction();

			q = pm.LoadWithId<Person>(q.PersonId);
			e = pm.LoadWithId<Employee>(e.PersonId);

			Assert.IsNull(q);
			Assert.IsNull(e);
		}

		[TestMethod]
		public void ShouldLoadFromOtherObjectContext()
		{
			ObjectContext pm1 = new ObjectService(provider).CreateObjectContext();

			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "p1";

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			Employee e = new Employee();
			e.Name = "e1";
			p.Employees.Add(e);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(1, pm.Load<Employee>().Count);

			IList<Employee> employees = pm1.Load<Employee>();
			Assert.AreEqual(1, employees.Count);
		}

		[TestMethod]
		public void ShouldPersistDeepChains()
		{
			pm.InitializeRepository();

			Company c = new Company();
			c.Name = "Evaluant";

			pm.BeginTransaction();
			pm.Serialize(c);
			pm.CommitTransaction();

			pm.Clear();

			c = pm.LoadSingle<Company>();

			Employee e = new Employee();
			e.Name = "employee 1";

			Person p1 = new Person();
			p1.Name = "p1";

			e.Partners.Add(p1);
			c.Employees.Add(e);

			pm.BeginTransaction();
			pm.Serialize(c);
			pm.CommitTransaction();

			Assert.AreEqual(1, pm.Load<Company>().Count);
			Assert.AreEqual(1, pm.Load<Employee>().Count);
			Assert.AreEqual(2, pm.Load<Person>().Count);
		}

		[TestMethod]
		public void ShouldLoadReference()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.Name = "p1";

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm = pm.Factory.CreateObjectContext();
			p = pm.LoadSingle<Person>("[Name='p1']");

			Employee e = new Employee();
			e.Name = "e1";
			p.Employees.Add(e);

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.RemoveFromCache(p);
			p = pm.LoadSingle<Person>("[Name='p1']");

			Assert.AreEqual(1, p.Employees.Count);
		}

		[TestMethod]
		public void ShouldSerializeSubLevels()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			Person p4 = new Person();
			p4.Name = "p4";

			Person p5 = new Person();
			p5.Name = "p5";

			p1.Partners.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);

			p2.Partners.Add(p3);
			p3.Partners.Add(p4);

			pm.BeginTransaction();
			pm.Serialize(p2);
			pm.CommitTransaction();

			Assert.AreEqual(4, pm.Load<Person>(typeof(Person)).Count);
		}

		[TestMethod]
		public void ShouldSerializeSubLevels2()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			Person p4 = new Person();
			p4.Name = "p4";

			Person p5 = new Person();
			p5.Name = "p5";

			p1.Partners.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			p3.Partners.Add(p1);

			pm.BeginTransaction();
			pm.Serialize(p3);
			pm.CommitTransaction();

			p4.Partners.Add(p1);

			pm.BeginTransaction();
			pm.Serialize(p4);
			pm.CommitTransaction();

			p5.Partners.Add(p4);

			pm.BeginTransaction();
			pm.Serialize(p5);
			pm.CommitTransaction();

			Assert.AreEqual(5, pm.Load<Person>(typeof(Person)).Count);
		}

		[TestMethod]
		public void ShouldSerializeReferenceIfChildIsAlreadyPersisted()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			p1.Partners.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p2);
			pm.CommitTransaction();

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[exists(Partners)]").Count);

			pm.InitializeRepository();
			pm.Clear();

			pm.BeginTransaction();
			pm.Serialize(p2);
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[exists(Partners)]").Count);
		}

		[TestMethod]
		public void ShouldSerializeMultipleReferencesToTheSameObject()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			p1.Partners.Add(p2);
			p1.BestFriend = p2;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[exists(Partners)]").Count);
			Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[exists(BestFriend)]").Count);

			pm.InitializeRepository();
			pm.Clear();

			p1 = new Person();
			p1.Name = "p1";

			p2 = new Person();
			p2.Name = "p2";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");
			p2 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p2']");

			p1.Partners.Add(p2);
			p1.BestFriend = p2;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[exists(Partners)]").Count);
			Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[exists(BestFriend)]").Count);

            pm.InitializeRepository();
            pm.Clear();

            p1 = new Person();
            p1.Name = "p1";


            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            pm.Clear();

            p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");

            p2 = new Person();
            p2.Name = "p2";

            p1.Partners.Add(p2);
            p1.BestFriend = p2;

            pm.BeginTransaction();
            pm.Serialize(p1);
            pm.CommitTransaction();

            Assert.AreEqual(2, pm.Load<Person>(typeof(Person)).Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[exists(Partners)]").Count);
            Assert.AreEqual(1, pm.Load<Person>(typeof(Person), "[exists(BestFriend)]").Count);
		}

		[TestMethod]
		public void ShouldReplaceExistingReference()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			Person p3 = new Person();
			p3.Name = "p3";

			p1.BestFriend = p2;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p3);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");
			p3 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p3']");

			Assert.AreEqual("p1", p1.Name);
			Assert.AreEqual("p2", p1.BestFriend.Name);

			p1.BestFriend = p3;
			p1.Partners.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

		}

		[TestMethod]
		public void ShouldAllowClearingLists()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Person p2 = new Person();
			p2.Name = "p2";

			p1.Partners.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p1']");
			p2 = pm.LoadSingle<Person>(typeof(Person), "[Name = 'p2']");

			p1.Partners.Clear();
			p1.Partners.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
		}

		[TestMethod]
		public void ShouldAllowInheritanceWithLoadWithId()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "p1";

			Employee e1 = new Employee();
			e1.Name = "e1";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(e1);
			pm.CommitTransaction();

			pm = pm.Factory.CreateObjectContext();

			Assert.AreEqual(2, pm.Load<Person>().Count);

			Assert.IsTrue(pm.LoadWithId<Person>(e1.PersonId) is Employee);
			Assert.IsTrue(pm.LoadWithId<Person>(p1.PersonId) is Person);
		}

		[TestMethod]
		public void ShouldLoadAllSubClassesOfAnInterface()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "Michel";

			Country c1 = new Country();
			c1.Name = "France";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(c1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.AreEqual(2, pm.Load<IName>(typeof(IName)).Count);

			p1 = pm.LoadSingle<Person>(typeof(IName), "[id('" + p1.PersonId + "')]");
			c1 = pm.LoadSingle<Country>(typeof(IName), "[id('" + c1.Id + "')]");

			Assert.AreEqual("Michel", p1.Name);
			Assert.AreEqual("France", c1.Name);

			pm.Clear();

			IName pn = pm.LoadSingle<IName>(typeof(IName), "[id('" + p1.PersonId + "')]");
			IName cn = pm.LoadSingle<IName>(typeof(IName), "[id('" + c1.Id + "')]");

			Assert.AreEqual("Michel", p1.Name);
			Assert.AreEqual("France", c1.Name);

			pm.Clear();

			Assert.AreEqual(2, pm.Load<ITag>(typeof(ITag)).Count);

			p1 = pm.LoadSingle<Person>(typeof(ITag), "[id('" + p1.PersonId + "')]");
			c1 = pm.LoadSingle<Country>(typeof(ITag), "[id('" + c1.Id + "')]");

			Assert.AreEqual("Michel", p1.Name);
			Assert.AreEqual("France", c1.Name);

			pm.Clear();

			ITag pt = pm.LoadSingle<ITag>(typeof(ITag), "[id('" + p1.PersonId + "')]");
			ITag ct = pm.LoadSingle<ITag>(typeof(ITag), "[id('" + c1.Id + "')]");

			pm.Clear();

			Assert.AreEqual(2, pm.Load<IElement>(typeof(IElement)).Count);

		}

		[TestMethod]
		public void ShouldDeleteWithId()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "Michel";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			IList<Person> people = pm.Load<Person>();

			Assert.AreEqual(1, people.Count);

			pm.BeginTransaction();
			pm.DeleteWithId<Person>(people[0].PersonId);
			pm.CommitTransaction();

			Assert.AreEqual(0, pm.Load<Person>().Count);
		}

		[TestMethod]
		public void ShouldNotSerializeVirtualPropertiesAutomatically()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "Michel";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			IList<Person> people = pm.Load<Person>();

			Assert.AreEqual(1, people.Count);
		}

		[TestMethod]
		public void ShouldHandleGuidType()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "Michel";
			p1.Owner = new Guid();

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			IList<Person> people = pm.Load<Person>();

			Assert.AreEqual(1, people.Count);
		}

		[TestMethod]
		public void ShouldNotSerializeIfCancelled()
		{
			pm.EnableEvents = true;
			pm.InitializeRepository();
			pm.ObjectInserting += new EussCancelEventHandler(pm_ObjectInserting);

			Person p = new Person();
			p.Name = "Nicolas";

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.EnableEvents = false;

			Assert.AreEqual(0, pm.Load<Person>().Count);

			pm = os.CreateObjectContext();

		}

		void pm_ObjectInserting(object sender, System.ComponentModel.CancelEventArgs e)
		{
			e.Cancel = true;
		}

		[TestMethod]
		public void ShouldNotLoadReferenceWhenUpdating()
		{
			pm.InitializeRepository();

			Company c = new Company();
			c.Name = "Evaluant";

			Employee ros = new Employee();
			ros.Name = "Ros";
			c.Employees.Add(ros);

			Employee chabrier = new Employee();
			chabrier.Name = "Chabrier";
			c.Employees.Add(chabrier);

			Employee gross = new Employee();
			gross.Name = "Gross";
			chabrier.Employees.Add(gross);

			Employee e = new Employee();
			e.Name = "Cornet";
			gross.Employees.Add(e);

			e = new Employee();
			e.Name = "Penin";
			gross.Employees.Add(e);

			pm.BeginTransaction();
			pm.Serialize(c);
			pm.CommitTransaction();

			pm.Clear();

			c = pm.LoadSingle<Company>("[Name='Evaluant']");
			e = pm.LoadSingle<Employee>();

			pm.BeginTransaction();
			pm.Serialize(c);
			pm.Serialize(e);

			// TODO: Check that LoadReference is not called in the engine during commit
			pm.CommitTransaction();
		}

		[TestMethod]
		public void ShouldUpdateDetachedEntity()
		{
			pm.InitializeRepository();

			Company c = new Company();
			c.Name = "Evaluant";

			Employee ros = new Employee();
			ros.Name = "Ros";
			c.Employees.Add(ros);

			pm.BeginTransaction();
			pm.Serialize(c);
			pm.CommitTransaction();

			string id = pm.GetId(c);

			Company dc = pm.Detach(c);

			pm.Clear();

			dc.Name = "Evaluant RC";

			pm.BeginTransaction();
			pm.SerializeDetached(dc, id);
			pm.CommitTransaction();

			pm.Clear();

			dc = pm.LoadSingle<Company>("[Name = 'Evaluant RC']");

			Assert.IsNotNull(dc);
			Assert.AreEqual(1, pm.Load<Company>().Count);
		}

		[TestMethod]
		public void ShouldSerializeCircularReferencesOnNewEntites()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "John Smith";
			p1.Age = 50;
			p1.Birth = new DateTime(2005, 12, 28);

			Person p2 = new Person();
			p2.Name = "Tom Cruise";
			p2.Age = 50;
			p2.Birth = new DateTime(2005, 12, 28);

			p1.BestFriend = p2;
			p2.BestFriend = p1;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.CommitTransaction();
		}

		[TestMethod]
		public void ObjectContextShouldNotKeepStrongReferencesOnNewlySerializedEntities()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.Name = "John Smith";
			p1.Age = 50;
			p1.Birth = new DateTime(2005, 12, 28);

			Person p2 = new Person();
			p2.Name = "John Smith";
			p2.Age = 50;
			p2.Birth = new DateTime(2005, 12, 28);

			p1.BestFriend = p2;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			p1 = null;
			p2 = null;

			GC.Collect();
			GC.WaitForPendingFinalizers();

			p1 = pm.LoadSingle<Person>();

			Assert.IsTrue(p1 == null || p1 is IPersistable);
		}
	}
}
