using System;
using System.Collections;
using NUnit.Framework;
using MyNamespace;

using System.Runtime.Serialization.Formatters.Soap;
using System.Xml.Serialization;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;

using Evaluant.Uss;
using Evaluant.Uss.MetaData;
using Evaluant.Uss.ObjectContext;


namespace TestPersistenceManager
{
	/// <summary>
	/// Description rsume de Class1.
	/// </summary>
	[TestFixture]
	public class TestGenerated
	{
		public ObjectContext pm;
		
		public TestGenerated()
		{
			Configure();
		}

		public virtual void Configure()
		{
			ObjectService factory = new ObjectService("engines.config", "Prevalence");
			factory.RegisterMetaData(MetaDataFactory.FromModelFile(@"..\..\PersistenceManager\HardCodedAndExplicit\schema.xml"));
			
			this.pm = factory.CreateObjectContext();
		}

		[Test]
		public void Order()
		{
			pm.InitializeRepository();

			Person p4 = new Person();
			p4.FirstName = "John";
			p4.Age = 12;

			Person p1 = new Person();
			p1.FirstName = "Eric";

			Person p2 = new Person();
			p2.FirstName = "John";
			p2.Age = 50;

			Person p3 = new Person();
			p3.FirstName = "Ben";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.Serialize(p3);
			pm.Serialize(p4);
			pm.CommitTransaction();
			pm.Clear();

			Query query = new Query(typeof(Person));
			query.OrderBy.Add("FirstName");
			query.OrderBy.Add("Age");

			IList people = pm.Load(query);

			Assert.AreEqual("Ben", ((Person)people[0]).FirstName);
			Assert.AreEqual("Eric", ((Person)people[1]).FirstName);
			Assert.AreEqual("John", ((Person)people[2]).FirstName);
			Assert.AreEqual("John", ((Person)people[3]).FirstName);
			Assert.AreEqual(12, ((Person)people[2]).Age);
			Assert.AreEqual(50, ((Person)people[3]).Age);

		}

		[SetUp]
		public void Init()
		{
			pm.InitializeRepository();
		}

		[Test]
		public void TestCompositionDelete()
		{		
			Dog d1 = new Dog();
		
			pm.BeginTransaction();
			pm.Serialize(d1);
			pm.CommitTransaction();
			pm.Clear();
			
			Assert.AreEqual(1, pm.Load(typeof(Dog)).Count);
			
			Dog dog = pm.LoadSingle(typeof(Dog)) as Dog;
			
			pm.BeginTransaction();
			pm.Delete(dog);
			pm.CommitTransaction();
		}

		[Test]
		public void Update()
		{
			Person p1 = new Person();
			p1.FirstName = "FirstName";
			p1.Age = 12;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			p1 = (Person)pm.Load(typeof(Person))[0];
			p1.FirstName = "Modified";
			p1.Age = 13;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			p1 = (Person)pm.Load(typeof(Person))[0];

			Assert.AreEqual("Modified", p1.FirstName);
			Assert.AreEqual(13, p1.Age);
		}

		[Test]
		public void TestEmptyClass()
		{		
			Person p1 = new Person();
			p1.FirstName = "FirstName";

			Company c1 = new Company();
			c1.logo = "logo";

			c1.Employees.Add(p1);

			pm.BeginTransaction();
			pm.Serialize(c1);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(1, pm.Load(typeof(Company), ".Employees").Count);
			c1 = (Company)pm.Load(typeof(Company))[0];

			Assert.AreEqual(1, c1.Employees.Count);
			Assert.AreEqual("FirstName", c1.Employees[0].FirstName);
		}
		
		[Test]
		public void TestInterface()
		{		
			Person p1 = new Person();
			p1.FirstName = "lon";
			p1.Code = "the killer";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Person p = (Person)pm.Load(typeof(Person), "[FirstName = 'lon']")[0];
			Assert.AreEqual(p.Code, "the killer");
		}

		[Test]
		public void TestEnumeration()
		{			
			Person anonyme = new Person();
			anonyme.FirstName = "gerard";
			anonyme.Title = Title.Sir;

			pm.BeginTransaction();
			pm.Serialize(anonyme);
			pm.CommitTransaction();
			pm.Clear();

			Person p = (Person)pm.Load(typeof(Person), "[FirstName = 'gerard']")[0];
			Assert.AreEqual(Title.Sir, p.Title);
		}

		[Test]
		public void TestReferenceCyclique()
		{
			Person p1 = new Person();
			p1.FirstName = "p1";

			Person p2 = new Person();
			p2.FirstName = "p2";
 
			p1.Friends.Add(p2);
			p2.Friends.Add(p1);
	
			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Person sp1 = (Person)pm.Load(typeof(Person), "[FirstName = 'p1']")[0];
			Person sp2 = (Person)pm.Load(typeof(Person), "[FirstName = 'p2']")[0];
		
			Assert.AreEqual(sp1.Friends[0].FirstName, "p2");
			Assert.AreEqual(sp2.Friends[0].FirstName, "p1");
			Assert.IsTrue(sp1.Friends[0] == sp2);
			Assert.IsTrue(sp2.Friends[0] == sp1);
		}
		
		[Test]
		public void TestComposition()
		{
			Library l = new Library();
			l.name = "bookshop";
			
			Book book = new Book();
			book.name = "le petit poucet";
			
			l.MyBooks.Add(book);
			
			pm.BeginTransaction();
			pm.Serialize(l);
			pm.CommitTransaction();

			Assert.IsTrue(l.MyBooks.Contains(book));
		
			pm.BeginTransaction();
			pm.Delete(l);
			pm.CommitTransaction();

			pm.Clear();
			
			Assert.AreEqual(0, pm.Load(typeof(Library)).Count);
			Assert.AreEqual(0, pm.Load(typeof(Book)).Count);
		}

        [Test]
        public void TestComposition2()
        {
            Library l = new Library();
            l.name = "bookshop";

            Book book = new Book();
            book.name = "le petit poucet";

            Shelf shelf = new Shelf();
            shelf.name = "big one";

            l.MyBooks.Add(book);
            l.Shelf = shelf;

            pm.BeginTransaction();
            pm.Serialize(l);
            pm.CommitTransaction();

            pm.BeginTransaction();
            pm.Delete(book);
            pm.CommitTransaction();

            pm.BeginTransaction();
            pm.Delete(shelf);
            pm.CommitTransaction();

            pm.Clear();

            Assert.AreEqual(1, pm.Load(typeof(Library)).Count);
            Assert.AreEqual(0, pm.Load(typeof(Book)).Count);
            Assert.AreEqual(0, pm.Load(typeof(Shelf)).Count);
        }

		[Test]
		public void TestCreation()
		{			
			Person p1 = new Person();
			//Vrification que l'entit a t cre
			Assert.IsNotNull(((IPersistable)p1).Entity);
			
			p1.FirstName = "jean";
			Assert.AreEqual("jean",p1.FirstName);
		
			Company c1 = new Company();
			c1.logo = "evaluant";
			
			//Vrification que la collection d'employs appartenant  p1 est vide
			Assert.AreEqual(0,(p1.Friends).Count);
			
			//Ajout d'une compagnie dans la collection d'employs appartenant  p1
			p1.Friends.Add(c1);
			Assert.AreEqual(1,(p1.Friends).Count);
			Assert.IsTrue(p1.Friends.Contains(c1));
			
			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			//Test de requtes
			IPersistable persist1;
			persist1= (IPersistable)pm.Load(typeof(Person), "[FirstName = 'jean'and count(Friends) > 0]")[0];
			Assert.IsNotNull(persist1);
		}

		[Test]
		public void TestIdentityMap()
		{			
			pm.InitializeRepository();

			Person p = new Person();
			p.FirstName = "Eric";

			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			pm.Clear();

			Person p1 = (Person)pm.Load(typeof(Person))[0];
			Person p2 = (Person)pm.Load(typeof(Person))[0];
			
			Assert.IsTrue(p1 == p2);
		}

		[Test]
		public void TestCount()
		{
			pm.InitializeRepository();

			Person p = new Person();
			p.FirstName = "Eric";
			
			pm.BeginTransaction();
			pm.Serialize(p);
			pm.CommitTransaction();

			Assert.AreEqual(1, pm.Load(typeof(Person)).Count);	
		}

		[Test]
		public void GetParents()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.FirstName = "Eric";

			Person p2 = new Person();
			p2.FirstName = "John";

			p1.Friends.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Person john = (Person)pm.Load(typeof(Person), "[FirstName = 'John']")[0];
			Assert.AreEqual(1, pm.LoadParents(typeof(Person), john, "Friends").Count);
		}

		[Test]
		public void TestDelete()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.FirstName = "Eric";

			Person p2 = new Person();
			p2.FirstName = "John";

			p1.Friends.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(2, pm.Load(typeof(Person)).Count);

			pm.BeginTransaction();
			pm.Delete(p1);
			pm.CommitTransaction();
            pm.Clear();

			Assert.AreEqual(1, pm.Load(typeof(Person)).Count);
		}

		[Test]
		public void TestDispose()
		{
			Person p1 = new Person();
			Person p2 = new Person();
			p1.FirstName="sylvain";
			p2.FirstName="alex";

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.Serialize(p2);
			pm.CommitTransaction();
			
			IPersistable persist1 = (IPersistable)pm.Load(typeof(Person), "[FirstName = 'sylvain']")[0];

			pm.Clear();
			
			IPersistable persist2 = (IPersistable)pm.Load(typeof(Person), "[FirstName = 'sylvain']")[0];
			
			//Vrification que les rfrences sont diffrentes avant et aprs suppression
			Assert.IsTrue(persist1 != persist2);

			//Vrification que l'objet n'a plus de ObjectContext
			Assert.IsNull(((IPersistable)p1).ObjectContext);
		}

		[Test]
		public void TestInheritance()
		{
			pm.InitializeRepository();

			Company c = new Company();
			SARL s = new SARL();

			pm.BeginTransaction();
			pm.Serialize(c);
			pm.Serialize(s);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(2, pm.Load(typeof(Company)).Count);
			Assert.AreEqual(1, pm.Load(typeof(SARL)).Count);

			// Make the test again to check the types cache is not buggy

			Assert.AreEqual(2, pm.Load(typeof(Company)).Count);
			Assert.AreEqual(1, pm.Load(typeof(SARL)).Count);

		}

		[Test]
		public void TestCollection()
		{
			PersonCollection people = new PersonCollection();

			Person p = new Person();
			p.FirstName = "toto";

			people.Add(new Person());
			people.Add(new Person());
			people.Add(new Person());
			people.Add(p);

			Assert.AreEqual(4, people.Count);
			Assert.IsNotNull(people[3]);
			Assert.AreEqual("toto", people[3].FirstName);

			pm.BeginTransaction();
			pm.Serialize(people);
			pm.CommitTransaction();
		}

		[Test]
		public void TestReference2()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.FirstName = "John";

			p1.Friends.Add(new Person());

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(1, pm.Load(typeof(Person), ".Friends").Count);

			pm.Clear();

			p1 = (Person)pm.Load(typeof(Person), "[FirstName = 'John']")[0];

			p1.Friends.Add(new Person());

			pm.BeginTransaction();
			pm.Serialize(p1);
			
			pm.CommitTransaction();

			pm.Clear();

			p1 = (Person)pm.Load(typeof(Person), "[FirstName = 'John']")[0];

			Assert.AreEqual(2, p1.Friends.Count);
		}

		[Test]
		public void LazyLoading()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.FirstName = "John";

			Person p2 = new Person();
			p2.FirstName = "Eric";
			p2.Age = 24;

			p1.Friends.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = (Person)pm.Load(typeof(Person), "[FirstName = 'John']")[0];

			Assert.IsNotNull(p1.Friends);
			Assert.AreEqual(1, p1.Friends.Count);
			Assert.AreEqual("Eric", p1.Friends[0].FirstName);
			Assert.AreEqual(24, p1.Friends[0].Age);

		}

		[Test]
		public void EagerLoading()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.FirstName = "John";

			Person p2 = new Person();
			p2.FirstName = "Eric";
			p2.Age = 24;

			Person p3 = new Person();
			p3.FirstName = "Mark";
			p3.Age = 24;

			p1.Friends.Add(p2);
			p1.Deputy = p3;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			int c1 = p1.Friends.Count;
			int c2 = p2.Friends.Count;
			object obj1 = p1.Deputy;
			object obj2 = p2.Deputy;


			//pm.Clear();

			Query q = new Query(typeof(Person), "[FirstName = 'John']");
			q.InferredRelationships.Add("Friends");
			q.InferredRelationships.Add("Deputy");

			p1 = (Person)pm.Load(q)[0];

			// The code above should not execute a query (loadReference)

			Assert.IsNotNull(p1.Friends);
			Assert.AreEqual(1, p1.Friends.Count);
			Assert.AreEqual("Eric", p1.Friends[0].FirstName);
			Assert.AreEqual(24, p1.Friends[0].Age);
			Assert.AreEqual("Mark", p1.Deputy.FirstName);

			//------------------------------------------
			// test with empty references without clearing the cache (loadReference)

			q = new Query(typeof(Person), "[FirstName = 'Eric']");
			q.InferredRelationships.Add("Friends");
			q.InferredRelationships.Add("Deputy");

			p2 = (Person)pm.Load(q)[0];
			
			// The code above should not execute a query

			Assert.IsNotNull(p2.Friends);
			Assert.AreEqual(0, p2.Friends.Count);
			Assert.IsNull(p2.Deputy);

			//------------------------------------------
			// test with empty references after clearing the cache
			pm.Clear();

			q = new Query(typeof(Person), "[FirstName = 'Mark']");
			q.InferredRelationships.Add("Friends");
			q.InferredRelationships.Add("Deputy");

			p2 = (Person)pm.Load(q)[0];
			
			// The code above should not execute a query (loadReference)

			Assert.IsNotNull(p2.Friends);
			Assert.AreEqual(0, p2.Friends.Count);
			Assert.IsNull(p2.Deputy);

			//------------------------------------------
			// test with none empty references after clearing the cache
			pm.Clear();

			q = new Query(typeof(Person), "[FirstName = 'John']");
			q.InferredRelationships.Add("Friends");
			q.InferredRelationships.Add("Deputy");

			p1 = (Person)pm.Load(q)[0];

			// The code above should not execute a query

			Assert.IsNotNull(p1.Friends);
			Assert.AreEqual(1, p1.Friends.Count);
			Assert.AreEqual("Eric", p1.Friends[0].FirstName);
			Assert.AreEqual(24, p1.Friends[0].Age);
			Assert.AreEqual("Mark", p1.Deputy.FirstName);
		}

		[Test]
		public void LoadSingle()
		{
			pm.InitializeRepository();

			Assert.IsNull(pm.LoadSingle(typeof(Person)));

			Person p1 = new Person();
			p1.FirstName = "John";

			Person p2 = new Person();
			p2.FirstName = "Eric";
			p2.Age = 24;

			p1.Friends.Add(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			Assert.IsNotNull(pm.LoadSingle(typeof(Person)));
		}

		[Test]
		public void OneToOne()
		{
			pm.InitializeRepository();

			// Instanciation
			Person p1 = new Person();
			p1.FirstName = "p1";

			Person p2 = new Person();
			p2.FirstName = "p2";

			p1.Deputy = p2;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(2, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1']").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Deputy").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Deputy[FirstName = 'p2']").Count);
			
			// New Association
			Person p3 = new Person();
			p3.FirstName = "p3";

			p1.Deputy = p3;
			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(3, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1']").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Deputy").Count);
			Assert.AreEqual(0, pm.Load(typeof(Person), "[FirstName = 'p1'].Deputy[FirstName = 'p2']").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Deputy[FirstName = 'p3']").Count);

		}

		[Test]
		public void OneToMany()
		{
			pm.InitializeRepository();

			// Instanciation
			Person p1 = new Person();
			p1.FirstName = "p1";

			Person p2 = new Person();
			p2.FirstName = "p2";

			Person p3 = new Person();
			p3.FirstName = "p3";

			p1.Friends.Add(p2);
			p1.Friends.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(3, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1']").Count);
			Assert.AreEqual(2, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']").Count);
			
			// New Association
			Person p4 = new Person();
			p4.FirstName = "p4";

			p1.Friends[0] = p4;
			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(4, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1']").Count);
			Assert.AreEqual(2, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends").Count);
			Assert.AreEqual(0, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p4']").Count);

			p1.Friends.Clear();
			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(4, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1']").Count);
			Assert.AreEqual(0, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends").Count);
		}

		[Test]
		public void ManyToMany()
		{
			pm.InitializeRepository();

			// Instanciation
			Person p1 = new Person();
			p1.FirstName = "p1";

			Person p2 = new Person();
			p2.FirstName = "p2";

			Person p3 = new Person();
			p3.FirstName = "p3";

			p1.Friends.Add(p2);
			p1.Friends.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(3, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1']").Count);
			Assert.AreEqual(2, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p3']").Count);

			Person p4 = new Person();
			p4.FirstName = "p4";

			Person p5 = new Person();
			p5.FirstName = "p5";

			p1.Friends.Clear();
			p1.Friends.AddRange(new Person[] { p4, p5 });

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
			pm.Clear();

			Assert.AreEqual(5, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1']").Count);
			Assert.AreEqual(2, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends").Count);
			Assert.AreEqual(0, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']").Count);
			Assert.AreEqual(0, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p3']").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p4']").Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p5']").Count);

		}

		[Test]
		public void Reference()
		{
			// Instanciation
			Person p1 = new Person();
			p1.FirstName = "p1";

			Person p2 = new Person();
			p2.FirstName = "p2";

			Person p3 = new Person();
			p3.FirstName = "p3";

			p1.Friends.Add(p2);
			p1.Friends.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Person x1 = pm.Load(typeof(Person), "[FirstName = 'p1']")[0] as Person;

			Assert.AreEqual(p1, x1);
			Assert.AreEqual(((IPersistable)p1).Entity, ((IPersistable)x1).Entity);

			Person x2 = pm.Load(typeof(Person), "[FirstName = 'p1'].Friends[FirstName = 'p2']")[0] as Person;
			Assert.AreEqual(p2, x2);
			Assert.AreEqual(((IPersistable)p2).Entity, ((IPersistable)x2).Entity);

			if(x1.Friends[0].FirstName == "p2")
			{
				Assert.AreEqual(p2, x1.Friends[0]);
				Assert.AreEqual(((IPersistable)p2).Entity, ((IPersistable)x1.Friends[0]).Entity);
			}

			if(x1.Friends[0].FirstName == "p3")
			{
				Assert.AreEqual(p3, x1.Friends[0]);
				Assert.AreEqual(((IPersistable)p3).Entity, ((IPersistable)x1.Friends[0]).Entity);
			}
		}

		[Test]
		public void DefaultValue()
		{
			Person p = new Person();
			Assert.IsNull(p.FirstName);
			Assert.AreEqual(0, p.Age);
		}

		[Test] 
		public void ParentRelationShipOne()
		{
			Person p1 = new Person();
			p1.FirstName = "p1";

			Person p2 = new Person();
			p2.FirstName = "p2";

			Person p3 = new Person();
			p3.FirstName = "p3";

			p1.Deputy = p2;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = (Person)pm.LoadSingle(typeof(Person), "[FirstName = 'p1']");
			p2 = (Person)pm.LoadSingle(typeof(Person), "[FirstName = 'p2']");
			Assert.IsTrue(p1.Deputy == p2);

			p1.Deputy = p3;

			// p2 should no more be the deputy of p1

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			Assert.AreEqual(3, pm.Load(typeof(Person)).Count);
			Assert.AreEqual(1, pm.Load(typeof(Person), ".Deputy").Count);
			Assert.AreEqual("p3", ((Person)pm.Load(typeof(Person), ".Deputy")[0]).FirstName);
		}

		[Test] 
		public void ParentRelationShipMany()
		{
			Person p1 = new Person();
			p1.FirstName = "p1";

			Person p2 = new Person();
			p2.FirstName = "p2";

			Person p3 = new Person();
			p3.FirstName = "p3";

			p1.Friends.Add(p2);
			p1.Friends.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = (Person)pm.LoadSingle(typeof(Person), "[FirstName = 'p1']");
			p2 = (Person)pm.LoadSingle(typeof(Person), "[FirstName = 'p2']");

			p1.Friends.Remove(p2);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = (Person)pm.LoadSingle(typeof(Person), "[FirstName = 'p1']");

			Assert.AreEqual(1, p1.Friends.Count);
			Assert.AreEqual("p3", ((Person)p1.Friends[0]).FirstName);

		}

		[Test]
		public void CascadeUpdate()
		{
			Person p1 = new Person();
			p1.FirstName = "p1";

			Person p2 = new Person();
			p2.FirstName = "p2";

			Person p3 = new Person();
			p3.FirstName = "p3";

			p1.Friends.Add(p2);
			p1.Friends.Add(p3);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();

			p1 = (Person)pm.LoadSingle(typeof(Person), "[FirstName = 'p1']");
			p2 = (Person)pm.LoadSingle(typeof(Person), "[FirstName = 'p2']");

			// Explicitly calls the Friend relationship to create the reference to p2
			Assert.AreEqual(2, p1.Friends.Count);
			p2.Age = 50;

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();

			pm.Clear();
			
			p2 = (Person)pm.LoadSingle(typeof(Person), "[FirstName = 'p2']");

			Assert.AreEqual(50, p2.Age);
		}

		[Ignore("Don't pass for the moment")]
		public void Detach()
		{
			pm.InitializeRepository();

			Person p1 = new Person();
			p1.FirstName = "p1";

			Person p2 = new Person();
			p2.FirstName = "p2";

			Person p3 = new Person();
			p3.FirstName = "p3";

			p1.Friends.Add(p2);
			
			p2.Friends.Add(p3);

			p3.Friends.Add(p1);

			pm.BeginTransaction();
			pm.Serialize(p1);
			pm.CommitTransaction();
	
			pm.Clear();

			p1 = pm.LoadSingle(typeof(Person), "[FirstName = 'p1']") as Person;
			Assert.AreEqual(1, p1.Friends.Count); // Lazyload partners for p1

			p1 = pm.Detach(p1) as Person;

			Assert.AreEqual(typeof(Person), p1.GetType());
			Assert.AreEqual("p1", p1.FirstName);
			Assert.AreEqual(1, p1.Friends.Count); // Check lazyloaded items are there
			Assert.AreEqual(0, ((Person)p1.Friends[0]).Friends.Count); // Check none lazyloaded items are not there
		}

		[TearDown]
		public void Final()
		{
		}
			
	}


	
}
