using System;
using System.IO;
using System.Xml;
using System.Xml.Xsl;

using System.Reflection;
using System.Collections;
using System.Text;
using Evaluant.Uss.Xmi;

namespace Evaluant.Uss.PersistenceLayerGen
{
    /// <summary>
    /// Description rsume de ScriptGenerator.
    /// </summary>
    public class PersistenceLayerGen
    {
        #region Members

        private string _ModelFileName;
        private string _DALFileName;
        private ProgrammingLangEnum _ProgLang;

        public enum ProgrammingLangEnum { CSharp, VBnet };

        private string CSBeginTag = "#region ";
        private string CSEndTag = "#endregion";
        private string VBBeginTag = "#Region ";
        private string VBEndTag = "#EndRegion";

        #endregion

        #region Properties

        public string DALFileName
        {
            get { return _DALFileName; }
            set { _DALFileName = value; }
        }

        public string ModelFileName
        {
            get { return _ModelFileName; }
            set { _ModelFileName = value; }
        }

        public ProgrammingLangEnum OutputLanguage
        {
            get { return _ProgLang; }
            set { _ProgLang = value; }
        }

        #endregion

        #region Ctor

        public PersistenceLayerGen(string modelFileName, ProgrammingLangEnum outputLanguage)
        {
            _ModelFileName = modelFileName;
            _ProgLang = outputLanguage;

            _DALFileName = Path.GetFileNameWithoutExtension(modelFileName) + GetDefaultExtension();

        }

        public PersistenceLayerGen(string modelFileName, string outputFileName, ProgrammingLangEnum outputLanguage)
        {
            _ModelFileName = modelFileName;
            _DALFileName = outputFileName;
            _ProgLang = outputLanguage;

        }

        #endregion

        #region Generate

        public void GenerateFile()
        {
            string resultCode = null;

            using (StreamReader modelReader = new StreamReader(_ModelFileName))
                resultCode = GenerateCode(_ModelFileName, modelReader.ReadToEnd());

            using (StreamWriter dalWriter = new StreamWriter(_DALFileName, false))
                dalWriter.Write(resultCode);
        }

        public string GenerateCode(string file, string contents)
        {
            try
            {
                string beginTag = String.Empty;
                string endTag = String.Empty;

                //if (this.GetDefaultExtension().EndsWith("cs"))
                if (_ProgLang == ProgrammingLangEnum.CSharp)
                {
                    beginTag = CSBeginTag;
                    endTag = CSEndTag;
                }
                else
                {
                    beginTag = VBBeginTag;
                    endTag = VBEndTag;
                }

                string sourceFileName = Path.Combine(Path.GetDirectoryName(file), Path.GetFileNameWithoutExtension(file) + GetDefaultExtension());
                Hashtable customCode = ExtractCustomRegions(sourceFileName, beginTag, endTag);

                //Does not handle anything else but a meta file
                string simple = contents;
                // simple = Converter.DeleteSchema(simple);

                XmlDocument xmlDoc = new XmlDocument();
                xmlDoc.LoadXml(simple);

                XmlDocument xslDoc = new XmlDocument();
                xslDoc.Load(Assembly.GetExecutingAssembly().GetManifestResourceStream("PersistenceLayerGen.xsl.AssociationClassConverter.xslt"));

                StringBuilder sb = new StringBuilder();
                XslTransform xslt = new XslTransform();
                xslt.Load(xslDoc, null, null);
                xslt.Transform(xmlDoc, null, new StringWriter(sb), null);
                File.WriteAllText("model2.xml", sb.ToString(), Encoding.Unicode);
                XmlDocument xmlResult = new XmlDocument();
                xmlResult.LoadXml(sb.ToString());

                //if (this.GetDefaultExtension().EndsWith("cs"))
                if (_ProgLang == ProgrammingLangEnum.CSharp)
                {
                    if (System.Environment.Version.ToString().StartsWith("1.1"))
                        xslDoc.Load(Assembly.GetExecutingAssembly().GetManifestResourceStream("PersistenceLayerGen.xsl.CSPersistenceManagerv11.xslt"));
                    else
                        xslDoc.Load(Assembly.GetExecutingAssembly().GetManifestResourceStream("PersistenceLayerGen.xsl.CSPersistenceManagerv20.xslt"));
                }
                else
                {
                    if (System.Environment.Version.ToString().StartsWith("1.1"))
                        xslDoc.Load(Assembly.GetExecutingAssembly().GetManifestResourceStream("PersistenceLayerGen.xsl.VBPersistenceManagerv11.xslt"));
                    else
                        xslDoc.Load(Assembly.GetExecutingAssembly().GetManifestResourceStream("PersistenceLayerGen.xsl.VBPersistenceManagerv20.xslt"));
                }

                sb = new StringBuilder();
                xslt = new XslTransform();
                xslt.Load(xslDoc, null, null);
                xslt.Transform(xmlResult, null, new StringWriter(sb), null);

                string result = InjectCustomRegions(sb.ToString(), customCode, beginTag, endTag);
                return result;
            }
            catch (Exception e)
            {
                return "";
            }

        }

        public virtual string GetDefaultExtension()
        {
            switch (_ProgLang)
            {
                case ProgrammingLangEnum.CSharp:
                    return ".cs";
                case ProgrammingLangEnum.VBnet:
                    return ".vb";
            }
            return ".cs";
        }
        #endregion

        #region Custom regions

        /// <summary>
        /// Reads all user's editable sections
        /// </summary>
        /// <param name="filename"></param>
        /// <returns></returns>
        protected Hashtable ExtractCustomRegions(string filename, string beginTag, string endTag)
        {
            Hashtable customCode = new Hashtable();

            if (!File.Exists(filename))
                return customCode;

            using (StreamReader sr = new StreamReader(filename))
            {

                string line = null;
                while ((line = sr.ReadLine()) != null)
                {
                    if (line.Trim().StartsWith(beginTag))
                    {
                        string tag = line.Trim();
                        string index = tag.Substring(beginTag.Length, tag.Length - beginTag.Length);

                        // Reads the content and stores it
                        StringBuilder customRegion = new StringBuilder();

                        int deep = 1;
                        while ((line = sr.ReadLine()) != null)
                        {
                            if (line.Trim().StartsWith(beginTag))
                                deep++;

                            if (line.Trim().StartsWith(endTag))
                                deep--;

                            if (deep == 0)
                                break;

                            customRegion.Append(line).Append(System.Environment.NewLine);
                        }

                        customCode.Add(index, customRegion.ToString());
                    }
                }
            }

            return customCode;
        }


        /// <summary>
        /// Put back a set of custom code regions
        /// </summary>
        /// <param name="content"></param>
        /// <param name="customRegions"></param>
        /// <returns></returns>
        protected string InjectCustomRegions(string content, Hashtable customRegions, string beginTag, string endTag)
        {
            foreach (DictionaryEntry de in customRegions)
            {
                string tag = beginTag + de.Key;
                int index = content.IndexOf(tag);
                if (index != -1)
                {
                    content = content.Insert(index + tag.Length + 1, (string)de.Value);
                }
            }

            return content;
        }

        #endregion
    }
}
